/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Common api routines
//
// FmaApi.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 03/Oct/1999
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Generic message area object structure

  typedef struct _XAREA {       /* area */
	ULONG   magic;                      // Magic number (mesage base specific)
	PAPI    papi;                       // Message base api pointer
  } XAREA, FAR * PXAREA;

  // Fts message object structure

  typedef struct _XMSG {        /* msg */
	struct _XMSG * pmsgPrev;            // Pointer to prev open msg in area
	struct _XMSG * pmsgNext;            // Pointer to next open msg in area
	ULONG   magic;                      // Magic number (mesage base specific)
	PXAREA  parea;                      // Message area handle
  } XMSG, FAR * PXMSG;

  // Macros to get an api pointer from harea and hmsg

#define PAPIFROMHAREA(harea) (((PXAREA)(harea))->papi)
#define PAPIFROMHMSG(hmsg)   (((PXMSG)(hmsg))->parea->papi)

  // Registered api table structure

  typedef struct _APIREF { /* apiref */
	UINT    type;                         // Message base type
	PAPI    papi;                         // Pointer to the api structure
  } APIREF, FAR * PAPIREF;

  // Registered api table

  static APIREF aApi[MBTYPE_MAXNO];       // Registered ip table

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine registers message base api
//

  FMAAPI_(BOOL) FmaRegisterApi(UINT type, PAPI (FMAAPI * pfnGetApi)(VOID))
  {
	UINT iApi;
	PAPI papi;

	// Valtypeate parameters

	if (!type || !pfnGetApi) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Scan registered api table looking for the specified type and 
	// if found, fail registration

	for (iApi = 0; iApi < numbof(aApi) && aApi[iApi].type; iApi++) {
	  if (aApi[iApi].type == type) {
		FMAERROR(ALREADY_EXISTS);
		return FALSE;
	  }
	}

	// Check if we exhaused registered ip table. Well, the error code
	// is a bit weird but i'm not about to create another one just for this

	if (iApi == numbof(aApi)) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	// Get the api pointer and check if ok. Note that the api initialization 
	// function is supposed to set error code if it fails

	if ((papi = pfnGetApi()) == NULL)
	  return FALSE;

	// Check if the api structure size is correct and fail if not. Note that
	// message base api providers may be in dlls built with the old headers.

	if (papi->cb != sizeof(API)) {
	  FMAERROR(NOT_SUPPORTED);
	  return FALSE;
	}

	// Otehrwise we assume that all function pointers area valid so,
	// set up the registered api table entry

	aApi[iApi].type = type;
	aApi[iApi].papi = papi;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens an area
//

  FMAAPI_(HAREA) FmaOpenArea(PSZ pszPath, UINT type, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr)
  {
	UINT iApi;

	// Valtypeate parameters

	if (!pszPath || !pszPath[0] || !type) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Scan registered api table looking for the specified type and 
	// if found, call the api open area function to do the rest

	for (iApi = 0; iApi < numbof(aApi) && aApi[iApi].type; iApi++) {
	  if (aApi[iApi].type == type) {
		ASSERT(aApi[iApi].papi != NULL);
		return aApi[iApi].papi->pfnOpenArea(pszPath, fOpenMode, fl, pnetAddr);
	  }
	}

	// If we got here there is no api with the specified type

	FMAERROR(NOT_REGISTERED);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes an area
//

  FMAAPI_(BOOL) FmaCloseArea(HAREA harea)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnCloseArea(harea);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine locks and unlocks an area
//

  FMAAPI_(BOOL) FmaLockArea(HAREA harea, BOOL bLock)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnLockArea(harea, bLock);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message number
//

  FMAAPI_(BOOL) FmaValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnValidateMsg(harea, msgn, pimsg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message map
//

  FMAAPI_(BOOL) FmaCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnCreateMsgMap(harea, paumsg, pcumsg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates messages
//

  FMAAPI_(BOOL) FmaEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes a message
//

  FMAAPI_(BOOL) FmaDeleteMsg(HAREA harea, MSGN msgn)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnDeleteMsg(harea, msgn);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets current message
//

  FMAAPI_(BOOL) FmaSetCurMsg(HAREA harea, MSGN msgn)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnSetCurMsg(harea, msgn);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets current message
//

  FMAAPI_(MSGN) FmaGetCurMsg(HAREA harea)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnGetCurMsg(harea);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets first message
//

  FMAAPI_(MSGN) FmaGetFirstMsg(HAREA harea)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnGetFirstMsg(harea);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets highest message
//

  FMAAPI_(MSGN) FmaGetHighMsg(HAREA harea)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnGetHighMsg(harea);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message count
//

  FMAAPI_(LONG) FmaGetMsgCount(HAREA harea)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnGetMsgCount(harea);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts message number to unique message id
//

  FMAAPI_(UMSG) FmaMsgnToUid(HAREA harea, MSGN msgn)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnMsgnToUid(harea, msgn);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts unique message id to message number
//

  FMAAPI_(MSGN) FmaUidToMsgn(HAREA harea, UMSG umsg, UINT fMode)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnUidToMsgn(harea, umsg, fMode);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens a message
//

  FMAAPI_(HMSG) FmaOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnOpenMsg(harea, mode, msgn, fl);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes a message
//

  FMAAPI_(BOOL) FmaCloseMsg(HMSG hmsg)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnCloseMsg(hmsg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads an abstracted message
//

  FMAAPI_(BOOL) FmaLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnLoadMsg(hmsg, ppfmm, fl);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves an abstracted message
//

  FMAAPI_(BOOL) FmaSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnSaveMsg(hmsg, pfmm, fl, cchTextTotal);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads a native message
//

  FMAAPI_(BOOL) FmaReadMsg(HMSG hmsg, PVOID pmsgh, PVOID * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnReadMsg(hmsg, pmsgh, ppchCtrl, pcchCtrl,
										  ppchText, pcchText, poffText);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes a native message
//

  FMAAPI_(BOOL) FmaWriteMsg(HMSG hmsg, PVOID pmsgh, PVOID pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnWriteMsg(hmsg, pmsgh, pchCtrl, cchCtrl,
							pchText, cchText, poffText, cchTextTotal);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message control info length
//

  FMAAPI_(BOOL) FmaGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnGetMsgCtrlLen(hmsg, pcchCtrl);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message text length
//

  FMAAPI_(BOOL) FmaGetMsgTextLen(HMSG hmsg, ULONG * pcchText)
  {
	ASSERT(hmsg != NULL);

	return PAPIFROMHMSG(hmsg)->pfnGetMsgTextLen(hmsg, pcchText);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lastread message info
//

  FMAAPI_(BOOL) FmaGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnGetLRInfo(harea, uid, plri);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets lastread message info
//

  FMAAPI_(BOOL) FmaSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, PSZ pszUserName)
  {
	ASSERT(harea != NULL);

	return PAPIFROMHAREA(harea)->pfnSetLRInfo(harea, uid, plri, pszUserName);
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaApi.c
/////////////////////////////////////////////////////////////////////////////
