/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app -- FMA Copy message base utility
//
// FmaCopy.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 01/Oct/1999
// Updated: 11/Feb/2000
//
// History: 01/Oct/1999 kvitekp - Initially created.
//
//          11/Feb/1999 kvitekp - Added .Pkt support
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
#include "FtsApi.h"
#include "SquApi.h"
#include "JamApi.h"
#include "PktApi.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Command line info

  static CHAR  achSrcBase[MAX_PATH];    // Source message base path
  static CHAR  achTgtBase[MAX_PATH];    // Target message base path

  static UINT  nSrcBaseType;            // Source message base type
  static UINT  nTgtBaseType;            // Target message base type

  static NETADDR netAddr;				// Default network address
  static NETADDR netOrig;				// Originatinon network address
  static NETADDR netDest;				// Destination network address

  static PSZ   pszArea;					// Area kludge

  static BOOL  bClearTargetBase;        // Clear target base before copy
  static BOOL  bVerbose;                // Verbose mode

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine displays usage help
//

  static VOID DoShowHelp(VOID)
  {
	printf( "\n"
			"FMA Message base copy utility built "__DATE__", "__TIME__"\n"
			"Written by Pete Kvitek of JV Dialogue 1st BBS, 2:5020/6@fidonet\n"
			"\n"
			"Usage: FmaCopy <source base> <target base> [options]\n"
			"\n"
			"       /A[areatag]   - add ^aAREA: kludge\n"
			"       /N[z:n/n.p]   - default network address (from Fts, Pkt)\n"
			"       /O[z:n/n.p]   - origination network address (to Pkt)\n"
			"       /D[z:n/n.p]   - destination network address (to Pkt)\n"
			"       /C            - clear target base before copy\n"
			"       /V            - verbose mode\n"
			"       /G[file|*|#]  - log to file, * to debugger, # to console\n"
			"\n"
			"Message base type is specified using the base path extension.\n"
			"Supported bases: *.msg (add backslash), Squish, JAM and Pkt.\n"
			"\n"
		  );
  }

/////////////////////////////////////////////////////////////////////////////
// This routine displays usage help
//

  static BOOL DoProcCmdLine(int argc, char * argv[])
  {
	int iarg;

	// Check if there are no command line options

	if (argc <= 1) {
	  DoShowHelp();
	  return FALSE;
	}

	// Scan the command line 

	for (iarg = 1; iarg < argc; iarg++) {
	  if (argv[iarg][0] == '/' || argv[iarg][0] == '-') {

		// Process the options

		switch (TOUPPER(argv[iarg][1])) {

		  case 'A': // Add area kludge
			pszArea = &argv[iarg][2];
			break;

		  case 'N': // Default network address
			if (!FmaScanNetAddr(&netAddr, &argv[iarg][2])) {
			  printf("Invalid network address: %s\n", argv[iarg]);
			  return FALSE;
			}
			break;

		  case 'O': // Origination network address
			if (!FmaScanNetAddr(&netOrig, &argv[iarg][2])) {
			  printf("Invalid network address: %s\n", argv[iarg]);
			  return FALSE;
			}
			break;

		  case 'D': // Destination network address
			if (!FmaScanNetAddr(&netDest, &argv[iarg][2])) {
			  printf("Invalid network address: %s\n", argv[iarg]);
			  return FALSE;
			}
			break;

		  case 'G': // Log options
			{ CHAR ach[MAX_PATH];
			  Format(ach, lengof(ach), "FMA-LOG=%s", &argv[iarg][2]);
			  putenv(ach);
			  Log("Logging to: %s", &argv[iarg][2]);
			}
			break;

		  case 'C': // Clear target base before copy
			bClearTargetBase = TRUE;
			break;

		  case 'V': // Verbose mode
			bVerbose = TRUE;
			break;

		  default:
			printf("Invalid option: %s\n", argv[iarg]);
			return FALSE;

		}

	  } else {

		// Process message base specification

		if (!achSrcBase[0]) {
		  strncpy(achSrcBase, argv[iarg], lengof(achSrcBase));
		} else 
		if (!achTgtBase[0]) {
		  strncpy(achTgtBase, argv[iarg], lengof(achTgtBase));
		} else {
		  printf("Extra message base specification: %s\n", argv[iarg]);
		  return FALSE;
		}
	  }
	}

	// Check if there is no source message base specified and fail

	if (!achSrcBase[0]) {
	  printf("Source message base is not specified\n");
	  return FALSE;
	}

	// Check if there is no target message base specified and fail

	if (!achTgtBase[0]) {
	  printf("Target message base is not specified\n");
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine finds out message base type
//

  static UINT DoGetBaseType(CHAR achMsgBase[])
  {
	CHAR achDrive[MAX_PATH], achDir[MAX_PATH];
	CHAR achFile[MAX_PATH], achExt[MAX_PATH];
	UINT cch, iExt;

	struct {
	  PSZ psz; UINT type; 
	} aExt[] = {
	  { ".squ",           MBTYPE_SQU },
	  { SQU_SQD_FILEEXT,  MBTYPE_SQU },
	  { SQU_SQI_FILEEXT,  MBTYPE_SQU },
	  { SQU_SQL_FILEEXT,  MBTYPE_SQU },
	  { ".jam",           MBTYPE_JAM },
	  { JAM_JHR_FILEEXT,  MBTYPE_JAM },
	  { JAM_JDT_FILEEXT,  MBTYPE_JAM },
	  { JAM_JDX_FILEEXT,  MBTYPE_JAM },
	  { JAM_JLR_FILEEXT,  MBTYPE_JAM },
	  { PKT_PKT_FILEEXT,  MBTYPE_PKT },
	};

	// Split the message base file name into parts

	_splitpath(achMsgBase, achDrive, achDir, achFile, achExt);

	// Check if there is no file and extension and assume it's
	// fts message base provided we have valid directory name

	if (!achFile[0] && !achExt[0]) {
	  if ((cch = strlen(achDir)) == 0 || achDir[cch - 1] != '\\')
		goto Fail;
	  return MBTYPE_FTS;
	}

	// Check if we have both file and extension and scan extension table

	if (achFile[0] && achExt[0]) {
	  for (iExt = 0; iExt < numbof(aExt); iExt++) {
		if (FmaEqPsz(aExt[iExt].psz, achExt)) {
		  _makepath(achMsgBase, achDrive, achDir, achFile, NULL);
		  return aExt[iExt].type;
		}
	  }
	}

	// Report an error    

Fail: printf("Can't determine message base type: %s\n", achMsgBase);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns message base types string
//

  static PSZ DoGetBaseTypeString(UINT nType)
  {
	switch (nType) {
	  case MBTYPE_FTS: return "*.msg";
	  case MBTYPE_SQU: return "Squish";
	  case MBTYPE_JAM: return "JAM";
	  case MBTYPE_PKT: return "PKT";
	}

	return "<Unknown>";
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for FmaEnumMsgs()
//

  static BOOL FMACALLBACK DoCopyMsg(PENUMMSGINFO pemi, HAREA hareaTgt)
  {
	HAREA hareaSrc = pemi->harea;
	HMSG hmsg;
	PFMM pfmm;
	MSGN msgn;

	// Display message number

	printf("Copying msg %lu of %lu\r", 1 + pemi->imsg, FmaGetMsgCount(hareaSrc));

	// Retrieve source message

	msgn = pemi->msgn;
	if ((hmsg = FmaOpenMsg(hareaSrc, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("\nCan't open source message %ld, error=%d\n", msgn, FmaGetLastError());
	  return FALSE;
	} else {

	  // Load source message and check if ok

	  if (!FmaLoadMsg(hmsg, &pfmm, 0)) {
		printf("\nCan't load source message %ld, error=%d\n", msgn, FmaGetLastError());
		FmaCloseMsg(hmsg);
		return FALSE;
	  } else
		FmaCloseMsg(hmsg);
	}

	// Add area kludge if requested

	if (pszArea) {
	  FmaAddFmp(pfmm, FMP_AREA, pszArea, (ULONG)-1);
	}

	// Store target message

	msgn = 0;
	if ((hmsg = FmaOpenMsg(hareaTgt, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("\nCan't open target message %ld, error=%d\n", msgn, FmaGetLastError());
	  return FALSE;
	} else {

	  // Save target message and check if ok

	  if (!FmaSaveMsg(hmsg, pfmm, 0, (ULONG)-1)) {
		printf("\nCan't load target message %ld, error=%d\n", msgn, FmaGetLastError());
		FmaDestroyMsg(pfmm);
		FmaCloseMsg(hmsg);
		return FALSE;
	  } else
		FmaCloseMsg(hmsg);
	}

	// Destroy copied message abstraction

	FmaDestroyMsg(pfmm);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine copies source message base into the target one
//

  static BOOL DoCopyBase(VOID)
  {
	ULONG cmsg, fl = FOA_LOCKAREA;
	HAREA hareaSrc = NULL;
	HAREA hareaTgt = NULL;
	UINT  nOpenType;

	// Open the source message base and check if ok

	hareaSrc = FmaOpenArea(achSrcBase, nSrcBaseType, OPENAREA_EXISTING, fl, &netAddr);
	if (!hareaSrc) {
	  printf("Can't open source message base %s, error: %d\n", achSrcBase, FmaGetLastError());
Fail: if (hareaSrc) FmaCloseArea(hareaSrc);
	  if (hareaTgt) FmaCloseArea(hareaTgt);
	  return FALSE;
	}

	// Open the target message base and check if ok

	nOpenType = bClearTargetBase ? OPENAREA_CREATE_ALWAYS : OPENAREA_ALWAYS;
	hareaTgt = FmaOpenArea(achTgtBase, nTgtBaseType, nOpenType, fl, &netAddr);
	if (!hareaTgt) {
	  printf("Can't open target message base %s, error: %d\n", achTgtBase, FmaGetLastError());
	  goto Fail;
	}

	// Inform user of what we're going to do

	cmsg = FmaGetMsgCount(hareaSrc);
	printf("Copying %lu msg%s in %s (%s) to %s (%s)...\n",
			cmsg, cmsg == 1 ? "" : "s",
			achSrcBase, DoGetBaseTypeString(nSrcBaseType),
			achTgtBase, DoGetBaseTypeString(nTgtBaseType));

	// Enumerate source base messages

	FmaEnumMsgs(hareaSrc, DoCopyMsg, hareaTgt, 0);

	// Inform user we're done

	cmsg = FmaGetMsgCount(hareaTgt);
	printf("Done... %lu msg%s in %s (%s)\n",
			cmsg, cmsg == 1 ? "" : "s",
			achTgtBase, DoGetBaseTypeString(nTgtBaseType));

	// Close both message bases

	FmaCloseArea(hareaSrc);
	FmaCloseArea(hareaTgt);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	BOOL bInitialized = FALSE;
	UINT type;

	// Process commad line

	if (!DoProcCmdLine(argc, argv)) {
Fail: if (bInitialized) FmaTerminate();
//    while (kbhit()) getch(); getch();
	  exit(EXIT_FAILURE);
	}

	// Initialize and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("Can't initialize Fidonet Message Access, error=%u\n", FmaGetLastError());
	  goto Fail;
	} else
	  bInitialized = TRUE;

	// Register all core message base apis

	if (!FmaRegisterApi(type = MBTYPE_FTS, FtsGetApi) ||
		!FmaRegisterApi(type = MBTYPE_SQU, SquGetApi) ||
		!FmaRegisterApi(type = MBTYPE_JAM, JamGetApi) ||
		!FmaRegisterApi(type = MBTYPE_PKT, PktGetApi)) {
	  printf("FmaRegisterApi() failed, type=%d, error=%u\n", type, FmaGetLastError());
	  goto Fail;
	}

	// Find out message base types and check if ok

	if ((nSrcBaseType = DoGetBaseType(achSrcBase)) == 0) goto Fail;
	if ((nTgtBaseType = DoGetBaseType(achTgtBase)) == 0) goto Fail;

	// Check if source message base is *.msg and warn user if no
	// default network address specified

	if ((nSrcBaseType == MBTYPE_FTS || nSrcBaseType == MBTYPE_PKT) && 
	   !netAddr.zone) {
	  printf("Missing default network address\n");
	  goto Fail;
	}

	// Check if target message base is pkt and warn user if no
	// origination or destination network address specified

	if (nTgtBaseType == MBTYPE_PKT && 
	   (!netAddr.zone || !netOrig.zone || !netDest.zone)) {
	  printf("Missing default, origination or destination network address\n");
	  goto Fail;
	}

	// Copy base and check if ok

	if (!DoCopyBase()) goto Fail;

	// Cleanup and exit

	if (!FmaTerminate()) {
	  printf("Can't terminte Fidonet Message Access, error=%u\n", FmaGetLastError());
	  goto Fail;
	}

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaCopy.c
/////////////////////////////////////////////////////////////////////////////
