/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Initialization and last error routines
//
// FmaInit.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 27/Oct/1999
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Include standard headers 

#if defined _DEBUG && _MSC_VER >= 1000
#include <crtdbg.h>
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Include auto-generated version header 

#include "FmaVer.h"

  // Define common version numbers

#define VER_MAJOR       rmj
#define VER_MINOR       rmm
#define VER_BUILD       rup

  // Define build name string

#ifndef VER_BUILD_NAME
#define VER_BUILD_NAME  szVerName
#endif

  // Define build user string

#ifndef VER_BUILD_USER
#define VER_BUILD_USER  szVerUser
#endif

  // Define build type string

#ifdef _DEBUG
#define VER_BUILD_TYPE  "(debugging build)"
#endif

#ifdef NDEBUG
#define VER_BUILD_TYPE  ""
#endif

#ifndef VER_BUILD_TYPE
#define VER_BUILD_TYPE  ""
#endif

  // Define build platform name

#ifndef VER_BUILD_PLATFORM
#define VER_BUILD_PLATFORM ""
#endif

  // Some preprocessor magic to convert version numbers into a string literal

#define VER_STRING1(n) #n
#define VER_STRING2(n)  VER_STRING1(n)
#define VERSION         VER_STRING2(VER_MAJOR##.##VER_MINOR##.##VER_BUILD)

  // Define version name string

#define VER_NAME "Fidonet Mail Access toolkit v" VERSION VER_BUILD_PLATFORM \
				 " " VER_BUILD_TYPE " " VER_BUILD_NAME

  // Define version stamp string

  static CHAR achVersionName[] = 
  "\n" VER_NAME;

  static CHAR achVersionNameEx[] = 
  "\nBuilt " __TIME__ " "__DATE__ " by " VER_BUILD_USER 
  "\nCopyright (c) 1999-2000 by Pete Kvitek"
  "\n";

  // Moudle local data

  static BOOL bInitialized;               // FMA has been initialized flag

#ifndef _MT  
  static UINT nLastErrorCode;             // Last error code
#else
#if defined __W32__
  static UINT nLastErrorTlsIndex;         // Last error code tls index
#else
#endif
#endif

  // Private runtime info

  FMARTM fmaRtm;

  // Global initialization structure

  FMAGLOBAL_(FMAINIT) fmaInit = { FmaMemAlloc, FmaMemRealloc, FmaMemFree };

#ifdef _DEBUG
  extern ULONG cMemAlloc;                 // Allocation counter
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine initializes Fidonet Mail Access
//

  FMAAPI_(BOOL) FmaInitialize(PFMAINIT pfmaInit)
  {
	// Turn on debugging options

#if defined _DEBUG && _MSC_VER >= 1000 && !defined _FMA_DLL_BUILD
//  _crtDbgFlag|= _CRTDBG_CHECK_ALWAYS_DF 
	_crtDbgFlag|= _CRTDBG_LEAK_CHECK_DF;
//  _crtDbgFlag|= _CRTDBG_DELAY_FREE_MEM_DF;
#endif

	// Check if already initialized and leave with error condition

	if (bInitialized) {
	  FMAERROR(ALREADY_EXISTS);
	  return FALSE;
	} else {

	  // Initialize per thread lasterror code storage 

#ifdef _MT

#ifdef __W32__
	  if ((nLastErrorTlsIndex = TlsAlloc()) == -1) {
		return FALSE;
	  }
#else
#error FmaInitialize: Thread Local Storage
#endif

#endif

	  // Set initialized flag and reset last error code

	  bInitialized = TRUE;
	  FMAERROR(NONE);
	}

	// Check if caller provided initialization structure and 
	// copy it over initialized global one

	if (pfmaInit) memcpy(&fmaInit, pfmaInit, sizeof(fmaInit));

	// Set in default memory management routines if necessary

	if (!fmaInit.pfnMemAlloc)   fmaInit.pfnMemAlloc   = FmaMemAlloc;
	if (!fmaInit.pfnMemRealloc) fmaInit.pfnMemRealloc = FmaMemRealloc;
	if (!fmaInit.pfnMemFree)    fmaInit.pfnMemFree    = FmaMemFree;

	// Reset open area count

	fmaRtm.nOpenAreas = FALSE;

	// Create ^aMSGID stamp generation protection mechanism

#ifdef _MT
#ifdef __W32__
	InitializeCriticalSection(&fmaRtm.csMsgId);
#endif
#endif

	// Seed unique message id stamp generator

	FmaGetMsgIdStamp();

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine initializes Fidonet Mail Access
//

  FMAAPI_(BOOL) FmaTerminate(VOID)
  {
	// Check if not initialized and leave

	if (!bInitialized) return TRUE;

	// Check if there are open areas

	if (fmaRtm.nOpenAreas) {
	  FMAERROR(STILL_OPEN);
#ifdef FMA_LOG
Log("FmaTerminate: %ld area(s) still open", fmaRtm.nOpenAreas); 
#endif
	  return FALSE;
	}

	// Destroy ^aMSGID stamp generation protection mechanism

#ifdef _MT
#ifdef __W32__
	DeleteCriticalSection(&fmaRtm.csMsgId);
#endif
#endif

	// Reset initialized flag

	bInitialized = FALSE;

	// Check allocation leaks

#ifdef _DEBUG
if (cMemAlloc) Log("FmaTerminate: Memory allocation leak, cMemAlloc = %ld", cMemAlloc); 
#endif

	// Free last error tls slot

#ifdef _MT

#ifdef __W32__
	TlsFree(nLastErrorTlsIndex);
#endif

#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns version information
//

  FMAAPI_(BOOL) FmaGetVersion(PFMAVER pfmaVer)
  {
	if (!pfmaVer) return FALSE;

	// Make up the version numbers

	pfmaVer->verMajor = VER_MAJOR;
	pfmaVer->verMinor = VER_MINOR;
	pfmaVer->verBuild = VER_BUILD;

	// Make up version name string

	Format(pfmaVer->achName, sizeof(pfmaVer->achName), 
		  "%.*s", lengof(pfmaVer->achName), 
		   achVersionName + 1);

	// Shut up compiler warning

	(void)achVersionNameEx[0];

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// Last error management routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine sets last error
//

  FMAAPI_(UINT) FmaSetLastError(UINT code)
  {
	// Set global last error code

#ifndef _MT
	return nLastErrorCode = code;
#else

	// Get thread specific last error code

#if   defined __W32__
	TlsSetValue(nLastErrorTlsIndex, (LPVOID) code);
	return code;
#else
#error FmaSetLastError
#endif

#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets last error
//

  FMAAPI_(UINT) FmaGetLastError(VOID)
  {
	// Get global last error code

#ifndef _MT    
	return nLastErrorCode;
#else

	// Get thread specific last error code

#if   defined __W32__
	return (UINT)TlsGetValue(nLastErrorTlsIndex);
#else
#error FmaGetLastError
#endif

#endif
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaInit.c
/////////////////////////////////////////////////////////////////////////////
