/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Logging routines module
//
// FmaLog.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 10/Feb/2000
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          10/Feb/2000 kvitekp - Added FmaAssertV()
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Logging support declarations

#ifndef FMA_NO_LOG

#define LF_FILE       0x0001      // Log to the file
#define LF_SYSTEM     0x0002      // Log to the system debug output
#define LF_CONSOLE    0x0004      // Log to the console window

  static USHORT fsLog;            // Logging flags

#ifndef FMA_NO_STDIO
  static FILE * pfileLog;         // Log file
#endif

#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine initializes the debugging log
//
#ifndef FMA_NO_LOG
  static BOOL FMASUB DoInitLog(VOID)
  {
#ifndef FMA_NO_STDIO
	PSZ pszLog;
#endif

	// Check if already initialized

	if (fsLog) return TRUE;

#ifdef FMA_NO_STDIO

	fsLog|= LF_SYSTEM;

#else

	// Check if debugging log requested

	if ((pszLog = getenv("FMA-LOG")) == NULL)
	  return FALSE;

	// Check logging options

	switch (pszLog[0]) {

	  case '*': fsLog|= LF_SYSTEM;
				break;

	  case '#': fsLog|= LF_CONSOLE;
				break;

	  default : if ((pfileLog = fopen(pszLog, "wt")) != NULL)
				  fsLog|= LF_FILE;
				break;
	}
#endif

	return fsLog;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine terminates the debugging log
//
#ifndef FMA_NO_LOG
  static VOID FMASUB DoQuitLog(VOID)
  {
#ifndef FMA_NO_STDIO

	// Check if log goes to a file and close it

	if (pfileLog != NULL) {
	  fclose(pfileLog);
	  pfileLog = NULL;
	}
#endif

	// Terminate logging

	fsLog = FALSE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine writes debug string 
//
#ifndef FMA_NO_LOG
  static VOID FMASUBV DoPutLog(PSZ pszFormat, va_list parg)
  {
	CHAR ach[1024];
	PSZ psz = NULL;
	INT cch, cb;
	PBYTE pb;

	// Check if logging is in effect

	if (!DoInitLog()) return;

	// Format the output string and check if it fits into the
	// buffer allocated on the stack. Otherwise allocate buffer
	// large enough to hold the formatted string

	cch = FormatV(ach, sizeof(ach) - sizeof("\n"), pszFormat, parg);
	if (cch >= 0) {
	  strcpy(ach + cch, "\n");
	  psz = ach;
	} else {
	  for (cb = sizeof(ach) * 2; ; cb*= 2) {

		// Reallocate the string buffer and check if ok

		if ((pb = MemRealloc(psz, cb)) == NULL) {
		  if (psz) MemFree(psz); psz = ach;
		  strcpy(ach, "<<< TOO LONG STRING >>>\n");
		  break;
		} else
		  psz = (PSZ)pb;

		// Format the string and check if it fits in

		cch = FormatV(psz, cb - sizeof("\n"), pszFormat, parg);
		if (cch >= 0) {
		  strcpy(psz + cch, "\n");
		  break;
		}
	  } 
	}

	// Check if logging to a system debug output

#ifdef __W32__
	if (fsLog & LF_SYSTEM) {
	  OutputDebugString(psz);
	} 
#endif

#ifndef FMA_NO_STDIO

	// Check if logging to a console

	if (fsLog & LF_CONSOLE) {
	  fputs(psz, stdout);
	  fflush(stdout);
	} 

	// Otherwise logging to a file

	if (pfileLog != NULL) {
	  fputs(psz, pfileLog);
	  fflush(pfileLog);
	} 

#endif

	// Free the buffer if it was allocated

	if (psz != ach) MemFree(psz);
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine formats and writes the debug string 
//

  FMAAPIV_(VOID) FmaLog(PSZ pszFormat, ...)
  {
#ifdef FMA_NO_LOG
#pragma message (__FILE__ ": Logging routines excluded")
#else
	va_list parg;
	va_start(parg, pszFormat);

	// If called with a null format string quit logging

	if (pszFormat)
	  DoPutLog(pszFormat, parg);
	else
	  DoQuitLog();

	va_end(parg);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine formats and writes the debug string given an arglist pointer 
//

  FMAAPI_(VOID) FmaLogV(PSZ pszFormat, va_list parg)
  {
#ifndef FMA_NO_LOG

	// If called with a null format string quit logging

	if (pszFormat)
	  DoPutLog(pszFormat, parg);
	else
	  DoQuitLog();
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is called by an ASSERT macro
//

  FMAAPI_(VOID) FmaAssertV(PSZ pszFormat, va_list parg)
  {
#ifndef FMA_NO_LOG
	
	DoPutLog(pszFormat, parg);

#ifdef FMA_BREAK_ON_FAILED_ASSERT
#if !defined(__BORLANDC__) || (__BORLANDC__ < 0x500)
	{
	  _asm int 3
	}
#else
	  __emit__(0xCC);
#endif
#endif
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is called by an ASSERT macro
//

  FMAAPIV_(VOID) FmaAssert(PSZ pszFormat, ...)
  {
#ifndef FMA_NO_LOG
	va_list parg;
	va_start(parg, pszFormat);

	FmaAssertV(pszFormat, parg);

	va_end(parg);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaLog.c
/////////////////////////////////////////////////////////////////////////////
