/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Utility routines module
//
// FmaUti.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 17/Feb/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          30/Jan/2000 kvitekp - Fixed zero node check in FmaScanNetAddr
//
//          17/Feb/2000 kvitekp - Renamed FormatNetAddr to FmaFormatNetAddr
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This subroutine checks if the space terminated string has the given char
//

  static BOOL FMASUB DoHasAddrChar(PCH pch, CHAR ch)
  {
	ASSERT(pch != NULL);

	while (*pch && !ISLWSP(*pch))
	  if (*pch == ch) return TRUE;
	  else pch++;

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This subroutine checks if the given char is a valid network address char
///

  static BOOL FMASUB DoIsAddrChar(CHAR ch)
  {
	return (ch && strchr("0123456789:/.", ch));
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// String management declarations
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine skips over the leading spaces and returns first nonspace
//

  FMAAPI_(CHAR) FmaSkipSpaces(PCH * ppch)
  {
	ASSERT(ppch != NULL);
	ASSERT(*ppch != NULL);

	while (ISLWSP(**ppch)) 
	  (*ppch)++;

	return **ppch;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine skips over the leading spaces and kills trailing spaces
//

  FMAAPI_(CHAR) FmaTrimSpaces(PCH * ppsz)
  {
	UINT ich;
	PCH psz;

	ASSERT(ppsz != NULL);
	ASSERT(*ppsz != NULL);

	// Skip over the leading spaces and check if there is more

	if (!SkipSpaces(ppsz))
	  return FALSE;

	// Trim trailing spaces

	for (ich = strlen(*ppsz) - 1, psz = *ppsz + ich;
		 ich > 0 && ISLWSP(*psz); 
		 ich--)
	  *psz-- = '\0';

	return **ppsz;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine allocates and copies the given string
//

  FMAAPI_(PSZ) FmaAllocString(PSZ * ppsz, PCH pch, UINT cch)
  {
	PSZ psz;

	ASSERT(ppsz != NULL);

	// Check if we just need to remove the existing string

	if (pch == NULL) {

	  if (*ppsz != NULL) {
		MemFree(*ppsz);
		*ppsz = NULL;
	  }

	  return NULL;
	} 

	// Check if this is an ascii string and calc its length

	if (cch == (UINT)-1) cch = strlen(pch);

	// Allocate memory to hold the new string and check if ok

	if ((psz = MemAlloc(cch + 1, 0)) == NULL)
	  return NULL;

	// Make up the new string before we free the old one to
	// allow case when pch == *ppsz

	memcpy(psz, pch, cch);
	psz[cch] = '\0';

	// Free the existing string if it exist

	if (*ppsz != NULL)
	  MemFree(*ppsz);

	// Set in the new string

	*ppsz = psz;

	return psz;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if two strings are equal
//

  FMAAPI_(BOOL) FmaEqPsz(PSZ psz1, PSZ psz2)
  {
	ASSERT(psz1 != NULL);
	ASSERT(psz2 != NULL);

	do {
	  if (*psz1 == *psz2 || TOLOWER(*psz1) == TOLOWER(*psz2)) {
		psz1++; psz2++;
	  } else 
		return FALSE;
	} while (*psz1);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if two character arrays are equal
//

  FMAAPI_(BOOL) FmaEqPch(PCH pch1, PCH pch2, UINT cch)
  {
	ASSERT(pch1 != NULL);
	ASSERT(pch2 != NULL);

	while (cch--) {
	  if (*pch1 == *pch2 || TOLOWER(*pch1) == TOLOWER(*pch2)) {
		pch1++; pch2++;
	  } else 
		return FALSE;
	} 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine copies one string into another up to max characters
//

  FMAAPI_(UINT) FmaCpyStrMax(PCH pchTo, PCH pchFrom, UINT cchMax)
  {
	UINT cch = min(strlen(pchFrom), cchMax - 1);

	memcpy(pchTo, pchFrom, cch);

	pchTo[cch] = '\0';

	return cch;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine calculates hash value for the given string
//

  FMAAPI_(ULONG) FmaGetHash(PSZ psz)
  {
	ULONG hash = 0L;
	PCH pch;

	ASSERT(psz != NULL);

	// Calculate hash value for the given string

	for (pch = psz; *pch; pch++) {
	 hash+= (hash << 3) + (ULONG) TOLOWER(*pch);
	}

	return hash;
  }

/////////////////////////////////////////////////////////////////////////////
// Network address management routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine scans the network address in form z:net/node.point
//

  FMAAPI_(PCH) FmaScanNetAddr(NETADDR * pnetAddr, PSZ psz)
  {
	PCH pch, pchEnd;

	// Skip through the leading spaces

	for (pch = psz; ISLWSP(*pch); pch++);
	pchEnd = pch;

	// Scan in the zone if any

	if (DoIsAddrChar(*pch) && DoHasAddrChar(pch, ':')) {
	  while (ISDIGIT(*pchEnd)) pchEnd++;
	  if (*pchEnd != ':' || pch == pchEnd) return NULL;
	  pnetAddr->zone = (USHORT)atoi(pch);
	  pch = ++pchEnd;
	  if (!ISDIGIT(*pch) || !DoHasAddrChar(pch, '/')) return NULL;
	}

	// Scan in the net if any

	if (DoIsAddrChar(*pch) && DoHasAddrChar(pch, '/')) {
	  while (ISDIGIT(*pchEnd)) pchEnd++;
	  if (*pchEnd != '/' || pch == pchEnd) return NULL;
	  pnetAddr->net = (USHORT)atoi(pch);
	  pch = ++pchEnd;
	  if (!ISDIGIT(*pch)) return NULL;
	}

	// Scan in the node if any

	if (DoIsAddrChar(*pch) && *pch != '.') {
	  while (ISDIGIT(*pchEnd)) pchEnd++;
	  pnetAddr->node = (USHORT)atoi(pch);
	  pch = pchEnd;
	}

	// Scan in the point if any

	if (*pch != '.') {
	  pnetAddr->point = 0;
	} else {
	  for (pchEnd = ++pch; ISDIGIT(*pchEnd); pchEnd++);
	  pnetAddr->point = (USHORT)atoi(pch);
	  pch = pchEnd;
	}

	// Check if zone or net is zero and return

	return (pnetAddr->zone && pnetAddr->net) ? pch : NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine formats network address string
//

  FMAAPI_(UINT) FmaFormatNetAddr(PCH pch, UINT cchMax, NETADDR * pnetAddr, PSZ pszDomain)
  {
	NETADDR netAddr = *pnetAddr;
	UINT cch;

	// Initialize the target string

	pch[0] = '\0'; cch = 0;

	// Check if we have something to format

	if (!netAddr.zone && !netAddr.net && !netAddr.node && !netAddr.point)
	  return FALSE;

	// Check if the zone is not zero and format it

	if (netAddr.zone) {
	  cch+= Format(pch + cch, cchMax - cch, "%u:", netAddr.zone);
	}

	// Always format the net and node numbers

	cch+= Format(pch + cch, cchMax - cch, "%u/%u", netAddr.net, netAddr.node);

	// Check if the point is not zero and format it

	if (netAddr.point) {
	  cch+= Format(pch + cch, cchMax - cch, ".%u", netAddr.point);
	}

	// Check if domain is specified and format it taking care
	// of possible leading @ character

	if (pszDomain && pszDomain[0]) {
	  if (pszDomain[0] == '@') pszDomain++;
	  if (pszDomain[0]) {
		cch+= Format(pch + cch, cchMax - cch, "@%s", pszDomain);
	  }
	}

	// Return number of characters formatted

	return cch;
  }

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// End of FmaUti.c
/////////////////////////////////////////////////////////////////////////////
