/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts area routines module
//
// FtsArea.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 17/Feb/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          08/Feb/2000 kvitekp - Moved area path validation to DoCreateArea
//
//          17/Feb/2000 kvitekp - Fixed FtsUidToMsgn range check
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Message map allocation 

#define MSGMAP_GROW      32             // Allocation grow granularity
#define MSGMAP_GROWTWEAK 4              // Allocation grow tweak factor

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates area object
//

  static PFAREA FMASUB DoCreateArea(PSZ pszPath, NETADDR * pnetAddr)
  {
	CHAR achPath[MAX_PATH];
	PFAREA parea;
	int cch;

	// Get area path length and check if ok

	if (!pszPath || (cch = strlen(pszPath)) == 0) {
BadP: FMAERROR(INVALID_PARAM);
	  return NULL;
	}

	// Check the area path length

	if (cch >= MAX_PATH - lengof("12345678.msg")) 
	  goto BadP;

	// Copy path to the local storage and append trailing backslash if 
	// it's not there

	memcpy(achPath, pszPath, cch + 1);
	if (achPath[cch - 1] != PATH_SEP) {
	  achPath[cch++] = PATH_SEP; 
	  achPath[cch] = '\0';
	}

	// Create message area object and check if ok

	cch = strlen(achPath);
	if ((parea = MemAlloc(sizeof(FAREA) + cch, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize new area object

	parea->magic = FAREA_MAGIC;
	parea->papi = FtsGetApi();
	if (pnetAddr) parea->netAddr = *pnetAddr;
	memcpy(parea->achPath, achPath, cch + 1);

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys area object
//

  static BOOL FMASUB DoDestroyArea(PFAREA parea)
  {
	// Check if we have area object and leave if not

	if (!parea) return FALSE;

	// Free message map if any

	if (parea->aumsg) MemFree(parea->aumsg);

	// Reset message area object magic number

	parea->magic = ~FAREA_MAGIC;

	// Free message area object

	MemFree(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates area
//

  static BOOL FMASUB DoCreateNewArea(PFAREA parea)
  {
	ASSERT(parea != NULL);

	// Create area message base and check if ok

	if (!FmaMakeDir(parea->achPath)) {
	  FMAERROR(CANNOT_CREATE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for FmaEnumFiles to delete message files
//

  static BOOL FMACALLBACK DoDelMsg(PSZ pszFile, ULONG cbFile, ULONG attr, 
								   PVOID pvCallbackData)
  {
	ASSERT(pszFile != NULL);

	// Delete message file

	FmaDeleteFile(pszFile);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine clears existing area
//

  static BOOL FMASUB DoClearExistingArea(PFAREA parea)
  {
	// Delete all the message files

	FmaEnumFiles(parea->achPath, "*"FTS_MSG_FILEEXT, DoDelMsg, 
				 NULL, FENUMFILES_FULLPATH);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for FmaEnumFiles to add message files
//

  static BOOL FMACALLBACK DoGrabMsg(PSZ pszFile, ULONG cbFile, ULONG attr, 
									PVOID parea)
  {
	UINT msgn;

	ASSERT(pszFile != NULL);

	// Check the size if the message file and leave if too small

	if (cbFile < sizeof(FMH)) 
	  return TRUE;

	// Convert message number and check if ok

	if ((msgn = atoi(pszFile)) == 0) 
	  return TRUE;

	// Add the new message to the message map and check if ok

	if (!FtsAddMsg(parea, msgn))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine shrinks area message map allocation
//

  static BOOL FMASUB DoFixMsgMap(PFAREA parea)
  {
	UINT * aumsg;
	UINT cumsg;

	ASSERT(parea->cmsg < UINT_MAX);

	// Shrink message map allocation to store exact number of messages

	if (parea->cumsg > (UINT)parea->cmsg) {
	  cumsg = (int)parea->cmsg;
	  if ((aumsg = MemRealloc(parea->aumsg, cumsg * sizeof(UINT))) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		parea->aumsg = aumsg;
		parea->cumsg = cumsg;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine compares message numbers for qsort()
//

  static int DoCmpMsgn(const void * pmsgn1, const void * pmsgn2)
  {
	return *((PUINT)pmsgn1) - *((PUINT)pmsgn2);
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens area
//

  FMAAPI_(HAREA) FtsOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr)
  {
	BOOL bAreaExists;
	PFAREA parea;

	// Assert known structure sizes to ensure correct structure packing

	ASSERT(sizeof(FMH) == 190);
	ASSERT(sizeof(FMH2) == sizeof(FMH));

#ifdef FMA_LOG_FTS_AREA
Log("\nFtsOpenArea      >: %s", pszPath);
#endif

	// Create new area object and check if ok

	if ((parea = DoCreateArea(pszPath, pnetAddr)) == NULL) {
Fail: DoDestroyArea(parea);

#ifdef FMA_LOG_FTS_AREA
Log("FtsOpenArea      <: %s FAILED, error=%d", pszPath, FmaGetLastError());
#endif

	  return NULL;
	}

	// Check if area message base exists

	bAreaExists = FmaDirExists(parea->achPath);

	// Act upon an opening mode

	switch (fOpenMode) {

	  case OPENAREA_EXISTING:

		// Fail if does not exist

		if (!bAreaExists) {
		  FMAERROR(NOT_FOUND);
		  goto Fail;
		}

		break;

	  case OPENAREA_ALWAYS:

		// Create if does not exist

		if (!bAreaExists && !DoCreateNewArea(parea)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_CREATE_NEW:

		// Fail if exists, otherwise create new

		if (bAreaExists) {
		  FMAERROR(ALREADY_EXISTS);
		  goto Fail;
		} else
		if (!DoCreateNewArea(parea)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_CREATE_ALWAYS:

		// Clear if exists, otherwise create new

		if (bAreaExists) {
		  if (!DoClearExistingArea(parea)) {
			goto Fail;
		  }
		} else 
		if (!DoCreateNewArea(parea)) {
		  goto Fail;
		}

		break;
	}

	// Rescan message area and check if ok

	if (!FtsRescanArea(parea)) {
	  goto Fail;
	}

	// Increment open area count in a thread safe manner

	FmaInterlockedIncrement(&fmaRtm.nOpenAreas);

	ASSERT(FtsCheckArea(parea));

	// Check if caller wants to lock the area upon opening

	if (parea->fl & FOA_LOCKAREA) {
	  if (!FtsLockArea(parea, TRUE)) {
		FtsCloseArea(parea);
		parea = NULL;
		goto Fail;
	  }
	}

#ifdef FMA_LOG_FTS_AREA
Log("FtsOpenArea      <: [%09Fp] cmsg=%ld msgnHigh=%ld in %s", parea, parea->cmsg, parea->msgnHigh, parea->achPath);
#endif

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes area
//

  FMAAPI_(BOOL) FtsCloseArea(HAREA harea)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_FTS_AREA
Log("\nFtsOpenArea      >: [%09Fp] cmsg=%ld msgnHigh=%ld in %s", parea, parea->cmsg, parea->msgnHigh, parea->achPath);
#endif

	// Check if caller locked the area upon opening and
	// area is still locked

	if (parea->fl & FOA_LOCKAREA && parea->bAreaLocked) {
	  FtsLockArea(parea, FALSE);
	}

	// Check if there are any open messages

	if (parea->pmsgList) {
	  FMAERROR(STILL_OPEN);
#ifdef FMA_LOG_FTS_AREA
Log("FtsCloseArea      : [%09Fp] %ld msg(s) still open", parea, ListCount(parea->pmsgList)); 
#endif

	  // Close all open messages

	  while (parea->pmsgList) {
		FtsCloseMsg(parea->pmsgList);
	  }
	}

	// Check if we have area locked

	while (parea->bAreaLocked) {
#ifdef FMA_LOG_FTS_AREA
Log("FtsCloseArea      : [%09Fp] bAreaLocked=%d", parea, parea->bAreaLocked); 
#endif
	  FtsLockArea(parea, FALSE);
	}

	// Destroy message area object

	DoDestroyArea(parea);

	// Decrement open area count in a thread safe manner

	FmaInterlockedDecrement(&fmaRtm.nOpenAreas);

#ifdef FMA_LOG_FTS_AREA
Log("FtsCloseArea     <: [%09Fp] CLOSED", parea); 
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine locks area
//

  FMAAPI_(BOOL) FtsLockArea(HAREA harea, BOOL bLock)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if locking or unlocking

	if (bLock) {

	  // Check if area is not yet locked

	  if (!parea->bAreaLocked) {

#ifdef FMA_LOG_FTS_AREA
Log("FtsLockArea       : [%09Fp] %s -- LOCKED", parea, parea->achPath);
#endif
	  }

	  // Increment area lock count

	  parea->bAreaLocked++;

	} else {

	  // Check if area is not locked and leave

	  if (!parea->bAreaLocked) {
		FMAERROR(NOT_LOCKED);
		return FALSE;
	  }

	  // Check if area has to be unlocked

	  if (parea->bAreaLocked == 1) {

#ifdef FMA_LOG_FTS_AREA
Log("FtsLockArea       : [%09Fp] %s -- UNLOCKED", parea, parea->achPath);
#endif
	  }

	  // Decrement area lock count

	  parea->bAreaLocked--;

	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message
//

  FMAAPI_(BOOL) FtsValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg)
  {
	PFAREA parea;
	UINT * pmsgn;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if (!parea->cmsg) return FALSE;

	ASSERT(parea->aumsg != NULL);
	ASSERT(parea->cumsg != 0);
	ASSERT(parea->cmsg < UINT_MAX);

	// Search the message map for the given message number and check if ok

	if ((pmsgn = bsearch(&msgn, parea->aumsg, (int)parea->cmsg, sizeof(UINT), DoCmpMsgn)) == NULL) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Check if caller needs message index and calculate it

	if (pimsg) *pimsg = (ULONG)(pmsgn - parea->aumsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message map which is array of unique message ids
//

  FMAAPI_(BOOL) FtsCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg)
  {
	ULONG cb, cumsg, iumsg;
	PFAREA parea;
	UMSG * aumsg;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(paumsg != NULL);
	ASSERT(pcumsg != NULL);

	// Check if no messages in this area and calculate message map size

	if ((cumsg = parea->cmsg) == 0 || !parea->aumsg) {
	  *paumsg = NULL; *pcumsg = 0;
	  return TRUE;
	} else
	  cb = cumsg * sizeof(UMSG);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Allocate memory to hold message map and check if ok

	if ((aumsg = MemAlloc((int)cb, 0)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	// Copy message map

	for (iumsg = 0; iumsg < cumsg; iumsg++) {
	  aumsg[iumsg] = (UMSG)parea->aumsg[iumsg];
	}

	// Set caller's pointers

	*paumsg = aumsg; *pcumsg = cumsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates messages
//

  FMAAPI_(BOOL) FtsEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl)
  {
	BOOL bSafeEnum = (fl & FENUMMSGS_SAFEENUM) ? TRUE : FALSE;
	ULONG iumsg, cumsg;
	BOOL bOk = TRUE;
	ENUMMSGINFO emi;
	UMSG * pumsg;
	UMSG * aumsg;
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea) || !pfnCallback) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if (!parea->cmsg) return FALSE;

	ASSERT(parea->aumsg != NULL);
	ASSERT(parea->cumsg > 0);

	// Since internal fts message map is an array of ints we always create
	// message map even if not enumerating safely

	if (!FtsCreateMsgMap(parea, &aumsg, &cumsg) || !cumsg)
	  return FALSE;

	// Prepare enum message info structure

	memset(&emi, 0, sizeof(emi));
	emi.harea = harea;

	// Check if need to enumerate in reversed order

	if (fl & FENUMMSGS_REVERSE) {

	  // Enumerate in reversed order

	  pumsg = aumsg + cumsg - 1;
	  for (iumsg = cumsg; iumsg > 0 ; iumsg--, pumsg--) {
		emi.msgn = emi.umsg = *pumsg; emi.imsg = iumsg - 1;
		if (bSafeEnum && !FtsValidateMsg(harea, (MSGN)*pumsg, &emi.imsg)) continue;
		bOk = pfnCallback(&emi, pvCallbackData);
		if (!bOk) break;
	  }
	} else {

	  // Enumerate in normal order

	  pumsg = aumsg;
	  for (iumsg = 0; iumsg < cumsg; iumsg++, pumsg++) {
		emi.msgn = emi.umsg = *pumsg; emi.imsg = iumsg;
		if (bSafeEnum && !FtsValidateMsg(harea, (MSGN)*pumsg, &emi.imsg)) continue;
		bOk = pfnCallback(&emi, pvCallbackData);
		if (!bOk) break;
	  }
	}

	// Free temporary copy of the message map

	MemFree(aumsg);

	// Return whatever last callback returned

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message
//

  FMAAPI_(BOOL) FtsDeleteMsg(HAREA harea, MSGN msgn)
  {
	static CHAR achMsgFmt[] = "%s%lu"FTS_MSG_FILEEXT;
	CHAR achMsg[MAX_PATH];
	PFAREA parea;
	ULONG imsg;
	int cb;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Validate message number and check if ok

	if (!FtsValidateMsg(harea, msgn, &imsg)) {
	  return FALSE;
	}

	// Make up the message file name

	Format(achMsg, lengof(achMsg), achMsgFmt, parea->achPath, msgn);

	// Delete the file and check if ok

	if (!FmaDeleteFile(achMsg))
	  return FALSE;

	ASSERT(parea->cmsg > 0);
	ASSERT(parea->aumsg != NULL);
	ASSERT(!FmaFileExists(achMsg));

	// Remove entry from the message map

	cb = (int)(parea->cmsg - (imsg + 1)) * sizeof(UINT);
	memmove(&parea->aumsg[imsg], &parea->aumsg[imsg + 1], cb);
	parea->cmsg--;

#ifdef _DEBUG
	ASSERT((ULONG)parea->cumsg >= parea->cmsg);
	cb = (int)(parea->cumsg - parea->cmsg) * sizeof(UINT);
	memset(&parea->aumsg[parea->cmsg], -1, cb);
#endif

	// Check if we just deleted current message and reset it

	if (parea->msgnCur == msgn) {
	  parea->msgnCur = 0;
	}

	// Check if we just deleted highest message and update it

	if (parea->msgnHigh == msgn) {
	  parea->msgnHigh = parea->cmsg ? parea->aumsg[parea->cmsg - 1] : 0;
	}

#ifdef FMA_LOG_FTS_AREA
Log("FtsDeleteMsg: %s msgn=%lu", parea->achPath, msgn);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets current message
//

  FMAAPI_(BOOL) FtsSetCurMsg(HAREA harea, MSGN msgn)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(msgn <= parea->msgnHigh);

	// Set current message

	parea->msgnCur = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets current message
//

  FMAAPI_(MSGN) FtsGetCurMsg(HAREA harea)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(parea->msgnCur <= parea->msgnHigh);

	// Return current message

	return parea->msgnCur;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lowest message number
//

  FMAAPI_(MSGN) FtsGetFirstMsg(HAREA harea)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we have no messages

	if (!parea->cmsg) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	ASSERT(parea->cumsg > 0);
	ASSERT(parea->aumsg != NULL);

	// Return lowest message number

	return parea->aumsg[0];
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets highest message number
//

  FMAAPI_(MSGN) FtsGetHighMsg(HAREA harea)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return highest message number

	return parea->msgnHigh;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message count
//

  FMAAPI_(LONG) FtsGetMsgCount(HAREA harea)
  {
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return message count

	return parea->cmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts message number to unique message id
//

  FMAAPI_(UMSG) FtsMsgnToUid(HAREA harea, MSGN msgn)
  {
	int imsg, imsgLo, imsgHi, imsgLast; // these have to be signed for bsearch
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!msgn || msgn > parea->msgnHigh) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Calculate very last message index

	imsgLast = (int)parea->cmsg - 1;

	// Run binary search looking for umsgid match

	imsgLo = 0; imsgHi = imsgLast;

	while (imsgLo <= imsgHi) {

	  // Split the current range in the middle

	  imsg = (imsgLo + imsgHi) / 2;

	  ASSERT(imsg >= 0);
	  ASSERT(imsg <= imsgLast);

	  // Check if we got a match and break

	  if (msgn == parea->aumsg[imsg])
		return msgn;

	  // Select a new range

	  if (msgn > parea->aumsg[imsg])
		imsgLo = imsg + 1;
	  else
		imsgHi = imsg - 1;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts unique message id to message number
//

  FMAAPI_(MSGN) FtsUidToMsgn(HAREA harea, UMSG umsg, UINT fMode)
  {
	int imsg = 0, imsgLo, imsgHi, imsgLast; // these have to be signed for bsearch
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message unique id

	if (!umsg || !parea->cmsg) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Calculate very last message index

	imsgLast = (int)parea->cmsg - 1;

	// Run binary search looking for umsgid match

	imsgLo = 0; imsgHi = imsgLast;

	while (imsgLo <= imsgHi) {

	  // Split the current range in the middle

	  imsg = (imsgLo + imsgHi) / 2;

	  ASSERT(imsg >= 0);
	  ASSERT(imsg <= imsgLast);

	  // Check if we got a match and break

	  if (umsg == parea->aumsg[imsg])
		break;

	  // Select a new range

	  if (umsg > parea->aumsg[imsg])
		imsgLo = imsg + 1;
	  else
		imsgHi = imsg - 1;
	}

	// Check if we got an exact match and return message number
	// if caller wants an exact one

	if ((UMSG)parea->aumsg[imsg] == umsg) {
	  if (fMode & UMSG_EXACT || !fMode)
		return umsg;
	}

	// Check if the caller wants next or previous message

	if (fMode & UMSG_NEXT) {
	  if ((UMSG)parea->aumsg[imsg] > umsg)
		return parea->aumsg[imsg];
	  else
		return imsg < imsgLast ? parea->aumsg[imsg + 1] : 0;
	} else
	if (fMode & UMSG_PREV) {
	  if ((UMSG)parea->aumsg[imsg] < umsg)
		return parea->aumsg[imsg];
	  else
		return imsg > 0 ? parea->aumsg[imsg - 1] : 0;
	} 

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message area object
//

  FMAAPI_(BOOL) FtsCheckArea(PFAREA parea)
  {
	ASSERT(parea != NULL);

	// Check the message area object magic number

	if (parea->magic != FAREA_MAGIC)
	  return FALSE;

	// Debug build only checks

#ifdef _DEBUG
	if (parea->cmsg) {
	  ASSERT(parea->cumsg > 0 && parea->aumsg != NULL);
#ifndef FMA_STRESS_TEST
	  ASSERT(parea->msgnHigh == parea->aumsg[parea->cmsg - 1]);
#endif
	} else {
	  ASSERT(parea->msgnHigh == 0);
	}
#endif

	// Check message map sorting

#ifdef _DEBUG
	if (parea->cmsg) {
	  MSGN msgn = 0; UINT imsg; BOOL bBadMessageMapSorting = FALSE;
	  for (imsg = 0; imsg < (UINT)parea->cmsg; imsg++) {
		if (msgn >= parea->aumsg[imsg]) {
		  bBadMessageMapSorting = TRUE;
		  break;
		} else
		  msgn = parea->aumsg[imsg];
	  }
	  ASSERT(bBadMessageMapSorting == FALSE);
	}
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sorts area message map allocation
//

  FMAAPI_(VOID) FtsSortMsgMap(PFAREA parea)
  {
	// Quick sort the message map and update the highest message number

	if (parea->cmsg) {
	  ASSERT(parea->aumsg != NULL);
	  ASSERT(parea->cmsg < UINT_MAX);
	  qsort(parea->aumsg, (int)parea->cmsg, sizeof(UINT), DoCmpMsgn);
	  parea->msgnHigh = parea->aumsg[parea->cmsg - 1];
	}

#ifdef FMA_LOG_FTS_AREA
//Log("FtsSortMsgMap: cmsg=%lu cumsg=%u %s", parea->cmsg, parea->cumsg, parea->achPath);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine rescans area
//

  FMAAPI_(BOOL) FtsRescanArea(PFAREA parea)
  {
	ASSERT(parea != NULL);

#ifdef FMA_LOG_FTS_AREA
//Log("FtsRescanArea: %s", parea->achPath);
#endif

	// Reset existing message data if any

	parea->cmsg = 0; parea->cumsg = 0; parea->msgnHigh = 0;
	if (parea->aumsg) MemFree(parea->aumsg); parea->aumsg = NULL;

	// Rescan all the message files and if failed, check if there are
	// no messages in this area and don't set error condition if so

	if (!FmaEnumFiles(parea->achPath, "*"FTS_MSG_FILEEXT, DoGrabMsg, parea, 0)) {
	  return (!parea->cmsg && FmaGetLastError() == FMAERROR_NOT_FOUND) ? TRUE : FALSE;
	}

	// Shrink message map allocation and sort it

	DoFixMsgMap(parea);
	FtsSortMsgMap(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds a message to the end of area message map
//

  FMAAPI_(BOOL) FtsAddMsg(PFAREA parea, MSGN msgn)
  {
	UINT cumsg, cumsgGrow = MSGMAP_GROW;
	UINT * aumsg;

	// Check if there is no room for a new message and reallocate message map

	if (!(parea->cmsg < parea->cumsg)) {

	  // Tweak allocation grow factor to prevent excessive heap fragmentation

	  if (parea->cumsg > MSGMAP_GROW * MSGMAP_GROWTWEAK) {
		cumsgGrow+= parea->cumsg / MSGMAP_GROWTWEAK;
	  }

	  // Reallocate message map buffer and check if ok

	  cumsg = parea->cumsg + cumsgGrow;
	  if ((aumsg = MemRealloc(parea->aumsg, cumsg * sizeof(UINT))) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		parea->cumsg = cumsg;
		parea->aumsg = aumsg;
	  }
	}

	// Add the new message to the message map and increment message count

	parea->aumsg[parea->cmsg++] = (UINT) msgn;

#ifdef FMA_LOG_FTS_AREA
//Log("FtsAddMsg: msgn=%lu cmsg=%lu cumsg=%u", msgn, parea->cmsg, parea->cumsg);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsArea.c
/////////////////////////////////////////////////////////////////////////////
