/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts read routines module
//
// FtsRead.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 07/Mar/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          09/Feb/2000 kvitekp - Handle AREA: with no ^a in DoReadCtrl()
//
//          07/Mar/2000 kvitekp - Ignore newlines in DoReadCtrl()
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message header
//

  static BOOL FMASUB DoReadFmh(PFMSG pmsg)
  {
	ULONG cbRead;

	ASSERT(pmsg != NULL);
	ASSERT(pmsg->hfile != -1);

	// Read the message header and check if ok

	if (!FmaReadFile(pmsg->hfile, 0, &pmsg->fmh, sizeof(FMH), &cbRead))
	  return FALSE;
	else
	  if (cbRead != sizeof(FMH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Fixup message header strings

	pmsg->fmh.achTo[sizeof(pmsg->fmh.achTo) - 1] = '\0';
	pmsg->fmh.achBy[sizeof(pmsg->fmh.achBy) - 1] = '\0';
	pmsg->fmh.achRe[sizeof(pmsg->fmh.achRe) - 1] = '\0';
	pmsg->fmh.achDate[sizeof(pmsg->fmh.achDate) - 1] = '\0';

	// Set flags to prevent further header reading

	SETFLAG(pmsg->fs, FMF_GOTMSGH);

#ifdef FMA_LOG_FTS_READ
Log("DoReadFmh         : [%09Fp] %s to %s \"%s\"", pmsg, pmsg->fmh.achBy, pmsg->fmh.achTo, pmsg->fmh.achRe);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads message control info
//

  static BOOL FMASUB DoReadCtrl(PFMSG pmsg)
  {
	static CHAR achArea[] = "AREA:";
	ULONG cch, ich, cchCtrl = 0;
	BOOL bDone = FALSE;
	BOOL bCtrl = FALSE;
	BOOL bArea = FALSE;
	CHAR achCtrl[512];
	UINT cchRead = 0;
	PSZ pszRead;
	PCH pch;

	ASSERT(pmsg != NULL);
	ASSERT(pmsg->hfile != -1);

	// Reset associated message control info if any

	if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl); 
	pmsg->pszCtrl = NULL; pmsg->cchCtrl = 0;

	// Loop reading the message control info in order to
	// determine control info length

	while (!bDone) {

	  // Read portion of the message control info and check if ok

	  if (!FmaReadFile(pmsg->hfile, sizeof(FMH) + cchCtrl, 
					   achCtrl, sizeof(achCtrl), &cch))
		return FALSE;

	  // Check if read nothing and leave loop

	  if (!cch) break;

	  // Scan portion of the message control info according to FTS-0001.016 
	  // specifications, that is looking for ^a at the beginning of the line
	  // or the message and ignoring \n's. Note special case for AREA: which
	  // often does not have ^a but goes first...

	  for (pch = achCtrl, ich = 0; ich < cch; pch++, ich++, cchCtrl++) {
		if (*pch == '\n') continue;
		if (*pch == 0x01) {
		  bCtrl = TRUE;
		} else 
		if (*pch == '\r') {
		  if (!bCtrl) {
			bDone = TRUE;
			break;
		  } else
			bCtrl = FALSE;
		} else 
		if (!bCtrl) {
		  if (!cchCtrl && !memcmp(pch, achArea, lengof(achArea))) {
			bCtrl = TRUE; bArea = TRUE;
		  }	else {
			bDone = TRUE;
			break;
		  }
		}
	  }
	}

	// Check if there is control info 

	if (cchCtrl) {

	  // Check if we've seen AREA: kludge without leading ^a and 
	  // add space for it

	  if (bArea) cchCtrl++;

	  // Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	  if (!CHECK_DATA_SIZE_LIMIT(cchCtrl)) {
		FMAERROR(DATA_SIZE_LIMIT);
		return FALSE;
	  }
#endif
	  // Allocate memory to hold control info and check if ok

	  if ((pmsg->pszCtrl = MemAlloc((int)cchCtrl + 1, 0)) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		pmsg->pszCtrl[cchCtrl] = '\0';
		pmsg->cchCtrl = (UINT)cchCtrl;
	  }

	  // Take care of the AREA: kludge without leading ^a

	  if (bArea) {
		pszRead = pmsg->pszCtrl + 1; pmsg->pszCtrl[0] = '\x01';
		cchRead = pmsg->cchCtrl - 1;
	  } else {
		pszRead = pmsg->pszCtrl;
		cchRead = pmsg->cchCtrl;
	  }

	  // Read the message control info and check if ok

	  if (!FmaReadFile(pmsg->hfile, sizeof(FMH), pszRead, cchRead, &cch)) {
		MemFree(pmsg->pszCtrl); pmsg->pszCtrl = NULL; pmsg->cchCtrl = 0;
		return FALSE;
	  } else 
		if (cch != cchRead) {
		  FMAERROR(INVALID_DATA);
		  return FALSE;
		}

	  // According to FTS-0001.016 newlines have to be ignored so remove
	  // them from the message control info

	  for (pch = pmsg->pszCtrl; (pch = strchr(pch, '\n')) != NULL; ) {
		strcpy(pch, pch + 1); pmsg->cchCtrl--;
	  }

	  // Check if messsage control info normalized to nothing and remove ot

	  if (!pmsg->cchCtrl) {
		if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl); pmsg->pszCtrl = NULL;
		cchRead = 0;
	  }
	}

	// Save message text offset

	pmsg->lmt = sizeof(FMH) + cchRead;

	// Set flags to prevent further control info reading

	SETFLAG(pmsg->fs, FMF_GOTCTRL);

	// Set flag to denote AREA: w/o ^a
	
	SETFLAGTO(pmsg->fs, FMF_CTRLHASAREA, bArea);

#ifdef FMA_LOG_FTS_READ
Log("DoReadCtrl        : [%09Fp] cchCtrl=%d \"%s\"", pmsg, pmsg->cchCtrl, pmsg->pszCtrl);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads message text
//

  static BOOL FMASUB DoReadText(PFMSG pmsg, PCH * ppchText, ULONG * pcchText, 
								ULONG * poffText)
  {
	BOOL bAlloc = FALSE;
	LONG offs, offsEnd;
	ULONG cch, cchAlloc;
	ULONG cbRead;

	ASSERT(pmsg != NULL);
	ASSERT(pmsg->hfile != -1);

	// Calculate the message text to read and message end offsets

	offs = pmsg->lmt;
	if (poffText) offs+= *poffText;
	offsEnd = FmaGetFileSize(pmsg->hfile);

	// Check if offset is past the end of the message text and fail,
	// otherwise calculate size of the message text we can read 

	if (offs > offsEnd) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	} else 
	  cch = offsEnd - offs;

	// Check if caller does not need all the message text and
	// adjust the read count if it's greater than requested 

	if (*pcchText != FRM_GETALLTEXT) {
	  if (cch > *pcchText) cch = *pcchText;
	}

	// Check if there is anything to read 

	if (cch > 0) {

	  // Check if we need to allocate text buffer for caller

	  if (*ppchText == NULL) {
		cchAlloc = (*pcchText != FRM_GETALLTEXT) ? *pcchText : cch;

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchAlloc)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate text buffer and check if ok 

		if ((*ppchText = MemAlloc((int)cchAlloc + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  (*ppchText)[cchAlloc] = '\0';
		  bAlloc = TRUE;
		}
	  } 

	  // Read the message text and check if ok

	  if (!FmaReadFile(pmsg->hfile, offs, *ppchText, cch, &cbRead)) {
Fail:   if (bAlloc) { MemFree(*ppchText); *ppchText = NULL; }
		return FALSE;
	  } else 
		if (cbRead != cch) {
		  FMAERROR(INVALID_DATA);
		  goto Fail;
		}
	}

	// Update caller's pointers

	*pcchText = cch;
	if (poffText) *poffText+= cch;

#ifdef FMA_LOG_FTS_READ
Log("DoReadText        : [%09Fp] cchText=%ld", pmsg, cch);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message
//

  FMAAPI_(BOOL) FtsReadMsg(HMSG hmsg, PFMH pfmh, PCH * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, 
						   ULONG * poffText)
  {
	PFAREA parea;
	PFMSG pmsg;
	PCH pszCtrl;

	// Validate parameters

	if ((pmsg = (PFMSG)hmsg) == NULL || !FtsCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_FTS_READ
Log("FtsReadMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_FTS_READ
Log("\nFtsReadMsg       >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	ASSERT(pmsg->hfile != -1);

	// Read the message header unless it's cached and check if ok

	if (!(pmsg->fs & FMF_GOTMSGH)) {
	  if (!DoReadFmh(pmsg))
		goto Fail;
	}

	// Read the message control info unless it's cached and check if ok

	if (!(pmsg->fs & FMF_GOTCTRL)) {
	  if (!DoReadCtrl(pmsg))
		goto Fail;
	}

	// Check if caller needs message header

	if (pfmh) {
	  *pfmh = pmsg->fmh;
	}

	// Check if caller needs message control info 

	if (ppchCtrl) {
	  if (!pmsg->pszCtrl || !pmsg->cchCtrl) {
		*ppchCtrl = NULL; *pcchCtrl = 0;
	  } else {

		// Check if caller needs a readonly pointer and set it,
		// otherwise allocate a copy of message control info

		if (*pcchCtrl == FRM_GETCTRLPTR) {
		  *ppchCtrl = pmsg->pszCtrl; *pcchCtrl = pmsg->cchCtrl;
		} else 
		if ((pszCtrl = MemAlloc(pmsg->cchCtrl + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  goto Fail;
		} else {
		  memcpy(pszCtrl, pmsg->pszCtrl, pmsg->cchCtrl + 1);
		  *ppchCtrl = pszCtrl; *pcchCtrl = pmsg->cchCtrl;
		}
	  }
	}

	// Check if caller needs message text.

	if (ppchText) {
	  if (!DoReadText(pmsg, ppchText, pcchText, poffText))
		goto Fail;
	}

#ifdef FMA_LOG_FTS_READ
Log("FtsReadMsg       <: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message control info length
//

  FMAAPI_(BOOL) FtsGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl)
  {
	PFMSG pmsg;

	// Validate parameters

	if ((pmsg = (PFMSG)hmsg) == NULL || !FtsCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	ASSERT(pmsg->hfile != -1);

	// Read the message control info unless it's cached and check if ok

	if (!(pmsg->fs & FMF_GOTCTRL)) {
	  if (!DoReadCtrl(pmsg))
		goto Fail;
	}

	// Set message control info length for caller

	if (pcchCtrl) *pcchCtrl = pmsg->cchCtrl;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message text length
//

  FMAAPI_(BOOL) FtsGetMsgTextLen(HMSG hmsg, ULONG * pcchText)
  {
	ULONG cbMsg;
	PFMSG pmsg;

	// Validate parameters

	if ((pmsg = (PFMSG)hmsg) == NULL || !FtsCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	ASSERT(pmsg->hfile != -1);

	// Read the message control info unless it's cached and check if ok

	if (!(pmsg->fs & FMF_GOTCTRL)) {
	  if (!DoReadCtrl(pmsg))
		goto Fail;
	}

	// Get the message file length and check if ok

	if ((cbMsg = FmaGetFileSize(pmsg->hfile)) == (ULONG)-1) {
	  FMAERROR(CANNOT_READ);
	  goto Fail;
	}

	ASSERT((LONG)cbMsg >= pmsg->lmt);

	// Set message text length for caller

	if (pcchText) *pcchText = cbMsg - pmsg->lmt;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsRead.c
/////////////////////////////////////////////////////////////////////////////
