/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts write routines module
//
// FtsWrite.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 28/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Following is used for stress testing only

#ifdef FMA_STRESS_TEST
  FMAAPI_(BOOL) FmaQueryDiskAllocFailure(HFILE hfile, ULONG cb);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds space to the specified file
//

  static BOOL FMASUB DoAddSpace(HFILE hfile, ULONG cb)
  {
	LONG offs, offsEof;
	CHAR ch = 0;

	ASSERT(hfile != -1);

	// Following is used for stress testing only

#ifdef FMA_STRESS_TEST
	if (FmaQueryDiskAllocFailure(hfile, cb)) return FALSE;
#endif

	// Truncate the file and check if ok

	if (!FmaSetFileSize(hfile, offsEof = 0)) {
Fail: FMAERROR(CANNOT_WRITE);
	  return FALSE;
	}

	// Check if need more space

	if (cb > 0) {

	  // Calculate offset of the very last byte of this space

	  offs = offsEof + cb - 1;

	  // Write a byte at this offset to make sure we have enough disk space

	  if (!FmaWriteFile(hfile, offs, &ch, 1, NULL))
		goto Fail;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the frame to write the message to
//

  static BOOL FMASUB DoGetWriteFrame(PFMSG pmsg, ULONG cchCtrl, ULONG cchTextTotal)
  {
	ULONG cbMsg;

	PFAREA parea = pmsg->parea;

	ASSERT(parea != NULL);

	// Check if creating a message 

	if (pmsg->mode == OPENMSG_CREATE) {

	  // Allocate space for the new message

	  cbMsg = sizeof(FMH) + cchCtrl + cchTextTotal;
	  if (!DoAddSpace(pmsg->hfile, cbMsg))
		return FALSE;

	  // Set message text writing controls

	  pmsg->lmt = sizeof(FMH) + cchCtrl;
	  pmsg->lmtEnd = pmsg->lmt;
	  pmsg->lmtMax = pmsg->lmt + cchTextTotal;

	} else 

	// Check if writing an existing message

	if (pmsg->mode == OPENMSG_WRITE || pmsg->mode == OPENMSG_READWRITE) {

	  ASSERT(pmsg->fs & FMF_GOTCTRL);

	  // Set message text writing controls

	  pmsg->lmtEnd = pmsg->lmt;
	  pmsg->lmtMax = FmaGetFileSize(pmsg->hfile);

	} else {

	  // Don't allow writing if message has been opened in read mode

	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Set flag that we've got the writing frame

	SETFLAG(pmsg->fs, FMF_GOTFRAME);

	// Update current message cache

	parea->msgnCur = pmsg->msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message header
//

  static BOOL FMASUB DoWriteFmh(PFMSG pmsg, PFMH pfmh)
  {
	ULONG cbWritten;

	ASSERT(pmsg != NULL);
	ASSERT(pmsg->hfile != -1);

	// Check if message header is not specified and 
	// assume the cached one

	if (!pfmh) pfmh = &pmsg->fmh;

	// Write the message header and check if ok

	if (!FmaWriteFile(pmsg->hfile, 0, pfmh, sizeof(FMH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(FMH));

	// Save message text offset

	pmsg->lmt = sizeof(FMH);

	// Check if wrote a new message header and 
	// update the cached one

	if (pfmh != &pmsg->fmh) {
	  SETFLAG(pmsg->fs, FMF_GOTMSGH);
	  pmsg->fmh = *pfmh;
	}

#ifdef FMA_LOG_FTS_WRITE
Log("DoWriteFmh        : [%09Fp] %s to %s \"%s\"", pmsg, pmsg->fmh.achBy, pmsg->fmh.achTo, pmsg->fmh.achRe);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message control info
//

  static BOOL FMASUB DoWriteCtrl(PFMSG pmsg, PCH pchCtrl, ULONG cchCtrl)
  {
	BOOL bNew = FALSE;
	ULONG cbWritten;
	PCH pch;

	ASSERT(pmsg != NULL);
	ASSERT(pmsg->hfile != -1);

	// Check if message contol info is not specified and 
	// assume the cached one, otherwise allocate a copy of
	// the new message control info unless it's empty

	 if (!pchCtrl) {
	  pchCtrl = pmsg->pszCtrl;
	  cchCtrl = pmsg->cchCtrl;
	} else {
	  bNew = TRUE;
	  if (!cchCtrl) {
		pchCtrl = NULL;
	  } else {

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchCtrl)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate memory to hold message control info

		if ((pch = MemAlloc((int)cchCtrl + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  memcpy(pch, pchCtrl, (int)cchCtrl); pchCtrl = pch; 
		  pchCtrl[cchCtrl] = '\0';
		}
	  }
	}

	// Check if we have message control info

	if (pchCtrl && cchCtrl) {

	  // Seek past the message header and write the message control info

	  if (!FmaWriteFile(pmsg->hfile, sizeof(FMH), pchCtrl, cchCtrl, &cbWritten)) {
		if (bNew) MemFree(pchCtrl);
		return FALSE;
	  }

	  ASSERT(cbWritten == cchCtrl);
	}

	// Save message text offset

	pmsg->lmt = sizeof(FMH) + cchCtrl;

	// Check if wrote a new message control info and update the cached one

	if (bNew) {
	  if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);
	  SETFLAG(pmsg->fs, FMF_GOTCTRL);
	  pmsg->pszCtrl = pchCtrl;
	  pmsg->cchCtrl = (UINT)cchCtrl;
	}

	// Reset flag to denote AREA: w/o ^a since we always
	// write it with ^a
	
	RESETFLAG(pmsg->fs, FMF_CTRLHASAREA);

#ifdef FMA_LOG_FTS_WRITE
Log("DoWriteCtrl       : [%09Fp] cchCtrl=%d \"%s\"", pmsg, pmsg->cchCtrl, pmsg->pszCtrl);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message text
//

  static BOOL FMASUB DoWriteText(PFMSG pmsg, PCH pchText, ULONG cchText, 
								 ULONG * poffText)
  {
//  PFAREA parea = pmsg->parea;
	ULONG cbWritten;
	LONG offs;

	ASSERT(pmsg->hfile != -1);

	// Calculate the message text offset

	offs = pmsg->lmt; if (poffText) offs+= *poffText;

	// Check if message text we're going to write will 
	// run past the end of the write frame

	if (offs + (LONG)cchText > pmsg->lmtMax) {
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message text and check if ok

	if (pchText && cchText) {

	  if (!FmaWriteFile(pmsg->hfile, offs, pchText, cchText, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cchText);
	}

	// Update the message text end and update flags

	pmsg->lmtEnd = offs  + cchText;
	SETFLAG(pmsg->fs, FMF_TEXTWRITTEN);

	// Update caller's text offset

	if (poffText) *poffText+= cchText;

	// Check if we're updating message and chop the message file past
	// the last text written. Note: for MSGOPEN_CREATE message file
	// is truncated when opened

	if (pmsg->mode == OPENMSG_WRITE || pmsg->mode == OPENMSG_READWRITE) {
	  if (!FmaSetFileSize(pmsg->hfile, offs + cchText))
		return FALSE;
	}

#ifdef FMA_LOG_FTS_WRITE
Log("DoWriteText       : [%09Fp] cchText=%d", pmsg, cchText);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine writes message
//

  FMAAPI_(BOOL) FtsWriteMsg(HMSG hmsg, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal)
  {
	PFAREA parea;
	PFMSG pmsg;

	// Validate parameters

	if ((pmsg = (PFMSG)hmsg) == NULL || !FtsCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_FTS_WRITE
Log("FtsWriteMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_FTS_WRITE
Log("\nFtsWriteMsg      >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	ASSERT(pmsg->hfile != -1);

	// Check if there is no frame for this messsage

	if (!(pmsg->fs & FMF_GOTFRAME)) {

	  // We need a message header and control info 
	  // on the first write to any message

	  if (!pfmh || (cchCtrl && !pchCtrl)) {
		FMAERROR(INVALID_PARAM);
		goto Fail;
	  }

	  // Get the frame to write and save result

	  if (!DoGetWriteFrame(pmsg, cchCtrl, cchTextTotal))
		goto Fail;
	}

	// Check if caller wants to write message header.

	if (pfmh) {
	  if (DoWriteFmh(pmsg, pfmh)) {
		SETFLAG(pmsg->fs, FMF_MSGHWRITTEN);
	  } else 
		goto Fail;
	}

	// Check if caller wants to write message control info

	if (pchCtrl) {
	  if (DoWriteCtrl(pmsg, pchCtrl, cchCtrl)) {
		SETFLAG(pmsg->fs, FMF_CTRLWRITTEN);
	  } else 
		goto Fail;
	}

	// Check if caller wants to write message text

	if (pchText) {
	  if (DoWriteText(pmsg, pchText, cchText, poffText)) {
		SETFLAG(pmsg->fs, FMF_TEXTWRITTEN);
	  } else 
		goto Fail;
	}

	// Check if the total message text length is specified and
	// set the message file size

	if (cchTextTotal) {
	  if (!FmaSetFileSize(pmsg->hfile, pmsg->lmt + cchTextTotal)) {
		FMAERROR(CANNOT_WRITE);
		goto Fail;
	  }
	}

#ifdef FMA_LOG_FTS_WRITE
Log("FtsWriteMsg      <: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsWrite.c
/////////////////////////////////////////////////////////////////////////////
