/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM index routines module
//
// JamIndex.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 14/Sep/1999
// Updated: 22/Sep/1999
//
// History: 14/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////
//
// Remarks: This code implements virtual JAM index which can be file based
//          or memory based (i.e. cached). The caching is controlled by calls
//          to JamBeginJdxCache()/JamEndJdxCache() which can be nested. When 
//          not caching, all index manipulation apis update the index file.
//          When caching, the same apis update cached index in memory and
//          the index file is updated during a call to JamEndJdxCache()...
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Macro to check if message number is valid

#define GOODMSGN(msgn)  ((msgn) >= parea->msgnBase && (msgn) <= parea->msgnHigh)

  // Macro to get lowest msgn value

#define MSGNBASE        (parea->msgnBase)

  // Index cache allocation constants

#define JIC_JDX_MAX  ((1*1024)/sizeof(JDX)) // Max number of index entries on page
#define JIC_JDX_GROW JIC_JDX_MAX            // Index cache grow factor

  // Macro to check for empty index entry

#define JDX_EMPTY(pjdx) ((pjdx)->lmh == 0)

  // Function prototypes

#ifndef NDEBUG  
  static BOOL FMASUB DoDumpJdxCache(PJAREA parea);
  static BOOL FMASUB DoCheckJdxCache(PJAREA parea);
  static BOOL FMASUB DoCompareJdxCache(PJAREA parea);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds JAM index cache page
//

  static PJIC FMASUB DoAddJicPage(PJAREA parea, UINT cjdx)
  {
	PJIC pjic;
	int cb;

	ASSERT(parea != NULL);
	ASSERT(cjdx > 0);

	// Allocate new cache page and check if ok

	cb = sizeof(JIC) + (cjdx - 1) * sizeof(JDX);
	if ((pjic = MemAlloc(cb, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} else
	  pjic->cjdx = cjdx;

	// Link up the new cache page

	ListAdd(&parea->pjicList, pjic, LIST_END);

#ifdef FMA_LOG_JAM_INDEX
//Log("DoAddJicPage      : [%09Fp] cjdx=%d", parea, pjic->cjdx);
#endif

	return pjic;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes JAM index cache page
//

  static BOOL FMASUB DoDelJicPage(PJAREA parea, PJIC pjicDel)
  {
	PJIC pjic;
	UINT ijic;

	ASSERT(parea != NULL);

	// Scan index cache pages looking for the specified page and if found,
	// unlink and destroy it

	for (ijic = 0, pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext, ijic++) {
	  if (pjic == pjicDel) {
		pjic = ListDel(&parea->pjicList, ijic);
		ASSERT(pjic != NULL);
		MemFree(pjic);

#ifdef FMA_LOG_JAM_INDEX
//Log("DoDelJicPage      : [%09Fp]", parea);
#endif
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes JAM index cache
//

  static VOID FMASUB DoDelJic(PJAREA parea)
  {
	PJIC pjic;

	ASSERT(parea != NULL);

	// Delete all index cache pages

	while (parea->pjicList) {
	  pjic = ListDel(&parea->pjicList, 0);
	  ASSERT(pjic != NULL);
	  MemFree(pjic);
	}

	// Reset index cache dirty pointers

	parea->msgnLoDirty = 0;
	parea->msgnHiDirty = 0;

#ifdef FMA_LOG_JAM_INDEX
//Log("DoDelJic          : [%09Fp]", parea);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns number of entries in JAM index cache
//

  static ULONG FMASUB DoGetJicEntryCount(PJAREA parea)
  {
	ULONG cjdx = 0;
	PJIC pjic;

	// Scan index cache pages adding up the used entries counts

	for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
	  cjdx+= pjic->ijdx;
	}

	return cjdx;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads specified number of JAM index entries
//

  static LONG FMASUB DoReadJdx(PJAREA parea, ULONG ijdx, JDX ajdx[], ULONG cjdx)
  {
	ULONG cb, cbRead;
	LONG offs;

	ASSERT(ajdx != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Find out specified index entry offset and number of bytes to read

	offs = ijdx * sizeof(JDX);
	cb   = cjdx * sizeof(JDX);

	// Read index entries and check if ok

	if (!FmaReadFile(parea->hfileJdx, offs, ajdx, cb, &cbRead))
	  return -1;

	// Return number of index entries read

	return cbRead / sizeof(JDX);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes specified number of JAM index entries
//

  static LONG FMASUB DoWriteJdx(PJAREA parea, ULONG ijdx, JDX ajdx[], ULONG cjdx)
  {
	ULONG cb, cbWritten;
	LONG offs;

	ASSERT(ajdx != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Find out specified index entry offset and number of bytes to write

	offs = ijdx * sizeof(JDX);
	cb   = cjdx * sizeof(JDX);

	// Write index entries and check if ok

	if (!FmaWriteFile(parea->hfileJdx, offs, ajdx, cb, &cbWritten))
	  return -1;

	ASSERT(cbWritten == cb);

	// Return number of index entries written

	return cbWritten / sizeof(JDX);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine truncates jdx file to be exactly the needed size
//

  static BOOL FMASUB DoChopJdx(PJAREA parea, ULONG cjdx)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->bBaseLocked);

	// Truncate index file to the desired size if necessary

	if ((ULONG)FmaGetFileSize(parea->hfileJdx) / sizeof(JDX) > cjdx) {
	  if (!FmaSetFileSize(parea->hfileJdx, cjdx * sizeof(JDX))) {
		FMAERROR(CANNOT_WRITE);
		return FALSE;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads jdx cache
//

  static BOOL FMASUB DoLoadJic(PJAREA parea)
  {
	int ijdx, cjdx;
	ULONG msgi;
	PJIC pjic;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Delete existing JAM index cache if any

	DoDelJic(parea);

	// Find out total number of jdx entries to load and
	// loop reading index entries

	for (cjdx = (int)(FmaGetFileSize(parea->hfileJdx) / sizeof(JDX)), msgi = 0; 
		 cjdx > 0; cjdx-= ijdx, msgi+= ijdx) {

	  // Calculate number if index entries to load into cache page

	  ijdx = min(cjdx, JIC_JDX_MAX);

	  // Allocate cache page  and check if ok

	  if ((pjic = DoAddJicPage(parea, ijdx)) == NULL) {
Fail:   DoDelJic(parea);
		return FALSE;
	  }

	  // Read cache page worth portion of the index file

	  if (DoReadJdx(parea, msgi, pjic->ajdx, ijdx) != ijdx) {
		goto Fail;
	  }

	  // Set number of used index entries on this cache page

	  pjic->ijdx = ijdx;
	}

#ifdef FMA_LOG_JAM_INDEX
Log("DoLoadJic         : [%09Fp] cjdx=%ld", parea, FmaGetFileSize(parea->hfileJdx) / sizeof(JDX));
#endif

	ASSERT(DoCompareJdxCache(parea));
	ASSERT(DoCheckJdxCache(parea));
//  ASSERT(DoDumpJdxCache(parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves jdx cache
//

  static BOOL FMASUB DoSaveJic(PJAREA parea)
  {
	MSGN msgnLoDirty, msgnHiDirty;
	MSGN msgnFrom, msgnTo, msgn;
	PJDX pjdx;
	PJIC pjic;
	int cjdx;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Pick up dirty boundary on stack

	msgnLoDirty = parea->msgnLoDirty;
	msgnHiDirty = parea->msgnHiDirty;

	ASSERT(msgnLoDirty <= msgnHiDirty);

	// Check if index cache is dirty

	if (msgnLoDirty || msgnHiDirty) {

	  // Loop through all the index cache pages

	  for (msgn = MSGNBASE, pjic = parea->pjicList; pjic != NULL; 
		   msgn+= pjic->ijdx, pjic = pjic->pjicNext) {

		// Check if this cache page does not contain index entries

		if (!pjic->ijdx) continue;

		// Check if this cache page index entries are entirely 
		// below the low dirty boundary and skip it

		if (msgn + pjic->ijdx - 1 < msgnLoDirty) continue;

		// Check if this cache page index entries are entirely 
		// above the hight dirty boundary and leave loop

		if (msgn > msgnHiDirty) break;

		// Find out the region of index entries to save on this page

		msgnFrom = max(msgn, msgnLoDirty);
		msgnTo = min(msgn + pjic->ijdx - 1, msgnHiDirty);

		ASSERT(msgnFrom <= msgnTo);

		// Calulate pointer to index entry and number of index entries to save

		pjdx = &pjic->ajdx[msgnFrom - msgn];
		cjdx = (int)(msgnTo - msgnFrom + 1);

		ASSERT(parea->bBaseLocked);

		// Write out index entries and check if ok

		if (DoWriteJdx(parea, msgnFrom - MSGNBASE, pjdx, cjdx) != cjdx) {
		  return FALSE;
		} 

#ifdef FMA_LOG_JAM_INDEX
Log("DoSaveJic         : [%09Fp] msgnLoDirty=%ld msgnHiDirty=%ld msgnFrom=%ld msgnTo=%ld", parea, parea->msgnLoDirty, parea->msgnHiDirty, msgnFrom, msgnTo);
#endif
	  }
	}

	ASSERT(DoCheckJdxCache(parea));
	ASSERT(DoCompareJdxCache(parea));

	// Delete JAM index cache

	DoDelJic(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds specified range to index cache dirty range
//

  static BOOL FMASUB DoAddDirty(PJAREA parea, MSGN msgnFrom, MSGN msgnTo)
  {
	BOOL bChanged = FALSE;

	ASSERT(parea != NULL);
	ASSERT(msgnFrom != 0);
	ASSERT(msgnTo != 0);
	ASSERT(msgnFrom <= msgnTo);

	// Adjust low dirty boundary

	if (!parea->msgnLoDirty || msgnFrom < parea->msgnLoDirty) {
	  parea->msgnLoDirty = msgnFrom;
	  bChanged = TRUE;
	}

	// Adjust high dirty boundary

	if (!parea->msgnHiDirty || msgnTo > parea->msgnHiDirty) {
	  parea->msgnHiDirty = msgnTo;
	  bChanged = TRUE;
	}

#ifdef FMA_LOG_JAM_INDEX
//Log("DoAddDirty        : [%09Fp] msgnLoDirty=%ld msgnHiDirty=%ld msgnFrom=%ld msgnTo=%ld", parea, parea->msgnLoDirty, parea->msgnHiDirty, msgnFrom, msgnTo);
#endif

	return bChanged;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine dumps JAM index cache
//
#ifndef NDEBUG
  static BOOL FMASUB DoDumpJdxCache(PJAREA parea)
  {
	MSGN msgn = MSGNBASE;
	PJIC pjic;
	UINT ijdx;

	ASSERT(parea != NULL);

	Log("DoDumpJdxCache   >:");

	for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
	  Log(" pjic=[%09Fp] cjdx=%u, ijdx=%u", pjic, pjic->cjdx, pjic->ijdx);
	  for (ijdx = 0; ijdx < pjic->ijdx; ijdx++, msgn++) {
		Log("  %4u: %08lx %08lx", msgn, pjic->ajdx[ijdx].crc, pjic->ajdx[ijdx].lmh);
	  }
	}

	Log("DoDumpJdxCache   <:");

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine checks JAM index cache
//
#ifndef NDEBUG
  static BOOL FMASUB DoCheckJdxCache(PJAREA parea)
  {
	MSGN msgn = MSGNBASE;
	PJIC pjic;
	UINT ijdx;

	ASSERT(parea != NULL);

	for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {

	  // Check if page index entry counters are valid

	  if (!pjic->ijdx || !pjic->cjdx || pjic->ijdx > pjic->cjdx) {
		Log("DoCheckJdxCache   : cjdx=%d ijdx=%d (BAD PAGE)", pjic->cjdx, pjic->ijdx);
Fail:   DoDumpJdxCache(parea);
		return FALSE;
	  }

	  // Check this page index entries 

	  for (ijdx = 0; ijdx < pjic->ijdx; ijdx++, msgn++) {

		// Check if it has valid message header offset

		if (pjic->ajdx[ijdx].lmh != (ULONG)-1) {
		  if (pjic->ajdx[ijdx].lmh < (ULONG)sizeof(JBH) || 
			  pjic->ajdx[ijdx].lmh > (ULONG)FmaGetFileSize(parea->hfileJhr)) {
			Log("DoCheckJdxCache   : msgn=%ld -- %08lx %08lx (BAD MESSAGE HEADER OFFSET)", msgn,
				 pjic->ajdx[ijdx].lmh, pjic->ajdx[ijdx].crc);
			goto Fail;
		  }
		}
	  }
	}

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine compares JAM index cache to an index file
//
#ifndef NDEBUG
  static BOOL FMASUB DoCompareJdxCache(PJAREA parea)
  {
	MSGN msgn = MSGNBASE;
	PJIC pjic;
	UINT ijdx;
	JDX jdx;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Scan through all cached index entries

	for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
	  for (ijdx = 0; ijdx < pjic->ijdx; ijdx++, msgn++) {

		// Read index entry and check if ok

		if (DoReadJdx(parea, msgn - MSGNBASE, &jdx, 1) != 1) {
		  Log("DoCompareJdxCache : can't read index file");
		  return FALSE;
		}

		// Compare cached and file index entry

		if (memcmp(&pjic->ajdx[ijdx], &jdx, sizeof(JDX))) {
		  Log("DoCompareJdxCache : msgn=%ld, cache/disk: %08lx/%08lx %08lx/%08lx",
			   pjic->ajdx[ijdx].crc, jdx.crc,
			   pjic->ajdx[ijdx].lmh, jdx.lmh);
		  return FALSE;
		}
	  }
	}

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine begins JAM message base index caching
//

  FMAAPI_(BOOL) JamBeginJdxCache(PJAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

#ifdef FMA_LOG_JAM_INDEX
Log("JamBeginJdxCache >: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif

	// Check if already caching and leave incrementing cache count

	if (parea->bJdxCached) {
	  parea->bJdxCached++;
#ifdef FMA_LOG_JAM_INDEX
Log("JamBeginJdxCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d ALREADY CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif
	  return TRUE;
	}

	// Load JAM index cache and check if ok

	if (!DoLoadJic(parea)) {
#ifdef FMA_LOG_JAM_INDEX
Log("JamBeginJdxCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d CANNOT LOAD CACHE", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif
	  return FALSE;
	}

	// Increment cache count 

	parea->bJdxCached++;

#ifdef FMA_LOG_JAM_INDEX
Log("JamBeginJdxCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d NOW CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine ends JAM message base index caching
//

  FMAAPI_(BOOL) JamEndJdxCache(PJAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

#ifdef FMA_LOG_JAM_INDEX
Log("JamEndJdxCache   >: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif

	// Check if not caching and fail

	if (!parea->bJdxCached) {
	  FMAERROR(NOT_CACHING);
#ifdef FMA_LOG_JAM_INDEX
Log("JamEndJdxCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d NOT CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif
	  return FALSE;
	}

	// Check if caching more than once and leave decrementing cache count

	if (parea->bJdxCached > 1) {
	  parea->bJdxCached--;
#ifdef FMA_LOG_JAM_INDEX
Log("JamEndJdxCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d STILL CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif
	  return TRUE;
	}

	ASSERT(parea->bJdxCached == 1);

	// Save JAM index cache and check if ok

	if (!DoSaveJic(parea)) {
#ifdef FMA_LOG_JAM_INDEX
Log("JamEndJdxCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d CANNOT SAVE CACHE", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif
	  return FALSE;
	}

	// Decrement cache count

	parea->bJdxCached--;

#ifdef FMA_LOG_JAM_INDEX
Log("JamEndJdxCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bJdxCached=%d NOW NOT CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bJdxCached);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns number of entries in JAM index
//

  FMAAPI_(ULONG) JamGetJdxCount(PJAREA parea)
  {
	ULONG cjdx;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Check if JAM index is cached and get number of cahced entries

	if (parea->bJdxCached) {

	  cjdx = DoGetJicEntryCount(parea);

	} else {

	  // If JAM index is not cached calculate the number 
	  // of index entries using the current idnex file size

	  cjdx = FmaGetFileSize(parea->hfileJdx) / sizeof(JDX);
	}

	return cjdx;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets JAM index entry
//

  FMAAPI_(BOOL) JamGetJdx(PJAREA parea, PJDX pjdx, MSGN msgn)
  {
	ULONG ijdx;
	PJIC pjic;

	ASSERT(pjdx != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);

	// Check for valid message number

	if (!GOODMSGN(msgn)) return FALSE;

	// Calculate requested message index entry index

	ijdx = msgn - MSGNBASE;

	// Check if JAM index is cached

	if (parea->bJdxCached) {

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, copy index entry for the caller

	  for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
		if (pjic->ijdx && ijdx < pjic->ijdx) {
		  *pjdx = pjic->ajdx[ijdx];
		  return TRUE;
		} else 
		  ijdx-= pjic->ijdx;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // read the index entry and check if ok

	  if (ijdx < JamGetJdxCount(parea)) {
		if (DoReadJdx(parea, ijdx, pjdx, 1) == 1) 
		  return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets JAM index entry
//

  FMAAPI_(BOOL) JamSetJdx(PJAREA parea, PJDX pjdx, MSGN msgn)
  {
	ULONG ijdx;
	PJIC pjic;

	ASSERT(pjdx != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->bBaseLocked);

	// Check for valid message number

	if (!GOODMSGN(msgn)) return FALSE;

	// Calculate requested message index entry index

	ijdx = msgn - MSGNBASE;

	// Check if JAM index is cached

	if (parea->bJdxCached) {

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, update index entry and dirty range

	  for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
		if (pjic->ijdx && ijdx < pjic->ijdx) {
		  DoAddDirty(parea, msgn, msgn);
		  pjic->ajdx[ijdx] = *pjdx;
		  return TRUE;
		} else 
		  ijdx-= pjic->ijdx;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // write the index entry and check if ok

	  if (ijdx < JamGetJdxCount(parea)) {
		if (DoWriteJdx(parea, ijdx, pjdx, 1) == 1) 
		  return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes JAM index entry
//

  FMAAPI_(BOOL) JamDelJdx(PJAREA parea, PJDX pjdx, MSGN msgn)
  {
	ULONG ijdx, cjdx;
	PJIC pjic;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->bBaseLocked);

	// Check if deleting the very last entry and leave if not

	if (msgn != parea->msgnHigh) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Calculate requested message index entry index

	ijdx = msgn - MSGNBASE;

	// Check if JAM index is cached

	if (parea->bJdxCached) {

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, delete index entry

	  for (pjic = parea->pjicList; pjic != NULL; pjic = pjic->pjicNext) {
		if (pjic->ijdx && ijdx < pjic->ijdx) {
		  if (pjdx) *pjdx = pjic->ajdx[ijdx];

		  ASSERT(ijdx == pjic->ijdx - 1);

		  // Check if this is the only used entry on this page and delete it, 
		  // otherwise decrement count of index entries on this page

		  if (pjic->ijdx == 1) {
			DoDelJicPage(parea, pjic);
		  } else {
			pjic->ijdx--;
		  }

		  // Chop index file to have exact amount of msgs

		  DoChopJdx(parea, DoGetJicEntryCount(parea));

		  return TRUE;
		} else 
		  ijdx-= pjic->ijdx;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // move the idnex entries over the deleted one

	  if (ijdx < (cjdx = JamGetJdxCount(parea))) {

		// Check if caller needs the deleted index entry and read it

		if (pjdx) {
		  if (DoReadJdx(parea, ijdx, pjdx, 1) != 1) 
			return FALSE;
		}

		// Chop index file to have exact amount of msgs

		DoChopJdx(parea, cjdx - 1);

		return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds specified JAM index entry to the end of index file
//

  FMAAPI_(BOOL) JamAddJdx(PJAREA parea, PJDX pjdx, MSGN * pmsgn)
  {
	ULONG ijdx;
	PJIC pjic;
	MSGN msgn;

	ASSERT(pjdx != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->bBaseLocked);

	// Check if JAM index is cached

	if (parea->bJdxCached) {

	  // Scan index cache pages looking for the last one

	  for (msgn = MSGNBASE, pjic = parea->pjicList; pjic != NULL; 
		   msgn+= pjic->ijdx, pjic = pjic->pjicNext) {
		if (!pjic->pjicNext) break;
	  }

	  // Check if there is no last cache page or it's all used and
	  // add another cache page if so

	  if (!pjic || pjic->ijdx == pjic->cjdx) {
		if (pjic) msgn+= pjic->ijdx;
		if ((pjic = DoAddJicPage(parea, JIC_JDX_GROW)) == NULL) 
		  return FALSE;
	  } 

	  ASSERT(pjic->ijdx < pjic->cjdx);

	  // Add  the specified index entry on the last page

	  msgn+= pjic->ijdx;
	  pjic->ajdx[pjic->ijdx++] = *pjdx;
	  DoAddDirty(parea, msgn, msgn);
	  if (pmsgn) *pmsgn = msgn;

	} else {

	  // Write index entry at the end of index file and check if ok

	  ijdx = JamGetJdxCount(parea);
	  if (DoWriteJdx(parea, ijdx, pjdx, 1) != 1) 
		return FALSE;

	  // Check if caller wants message number and set it in

	  if (pmsgn) *pmsgn = MSGNBASE + ijdx;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamIndex.c
/////////////////////////////////////////////////////////////////////////////
