/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM write routines module
//
// JamWrite.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 28/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Following is used for stress testing only

#ifdef FMA_STRESS_TEST
  FMAAPI_(BOOL) FmaQueryDiskAllocFailure(HFILE hfile, ULONG cb);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets crc of the specified subfield
// 

  static ULONG FMASUB DoGetSubfCrc(PJSF pjsf, ULONG cbjsf, USHORT id)
  {
	ULONG crc = (ULONG)-1;

	// Get the specified subfield crc

	if ((pjsf = JamGetJsf(pjsf, cbjsf, id, 0)) != NULL) {
	  crc = JamGetCrc32((PCH)PJSFDATA(pjsf), pjsf->cb, crc);
	}

	return crc;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds space at the end of the specified file
//

  static BOOL FMASUB DoAddSpace(HFILE hfile, LONG * poffs, ULONG cb)
  {
	long offs, offsEnd;
	char ch = 0;

	ASSERT(hfile != -1);
	ASSERT(poffs != NULL);

	// Following is used for stress testing only

#ifdef FMA_STRESS_TEST
	if (FmaQueryDiskAllocFailure(hfile, cb)) return FALSE;
#endif

	// Find out the space offset

	if (*poffs == -1)
	  offs = FmaGetFileSize(hfile);
	else
	  offs = *poffs;

	// Calculate end offset of this space

	offsEnd = offs + cb;

	// Check if need more space

	if (cb > 0) {

	  // Write a very last byte of the space

	  if (!FmaWriteFile(hfile, offsEnd - 1, &ch, 1, NULL))
		return FALSE;
	}

	// Chop the file if necessary

	if (FmaGetFileSize(hfile) > offsEnd) {
	  if (!FmaSetFileSize(hfile, offsEnd)) 
		return FALSE;
	}

	// Set the space offset for the caller

	*poffs = offs;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the frame to write the message to
//

  static BOOL FMASUB DoGetWriteFrame(PJMSG pmsg, ULONG cbjsf, ULONG cchTextTotal)
  {
	LONG lmh = -1, lmt = -1, offEnd;
	PJAREA parea = pmsg->parea;
	BOOL bAddJhr = FALSE;
	BOOL bAddJdt = FALSE;
	JMH jmh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	// Check if creating a message 

	if (pmsg->mode == OPENMSG_CREATE) {

	  // Check if creating a message with new msgn

	  if (!pmsg->lmh) {

		// Add space to the header file and check if ok

		if (!DoAddSpace(parea->hfileJhr, &lmh, sizeof(JMH) + cbjsf))
		  return FALSE;

		// Add space to the text file and if failed, 
		// roll back header file space allocation

		if (!DoAddSpace(parea->hfileJdt, &lmt, cchTextTotal)) {
		  FmaSetFileSize(parea->hfileJhr, lmh);
		  return FALSE;
		}

		// Set message header writing controls

		pmsg->lmh = lmh;
		pmsg->cbjsf = 0;
		pmsg->cbjsfMax = cbjsf;

		// Set message text writing controls

		pmsg->lmt = lmt;
		pmsg->lmtEnd = lmt;
		pmsg->lmtMax = lmt + cchTextTotal;

		// Set read deleted flag so that active message count would be 
		// incremented when writing message header

		SETFLAG(pmsg->fs, JMF_READDELETED);

	  } else {

		ASSERT(JamCheckLmh(parea, pmsg->lmh));

		// If creating a message with existing msgn,
		// load old message header and check if ok

		if (!JamLoadJmh(parea, &jmh, NULL, pmsg->lmh))
		  return FALSE;

		// Check if new subfields have different length and 
		// allocate space at the end of header file

		if (cbjsf != jmh.cbjsf) {

		  // Check if this message header is already at eof and 
		  // set allocation start offset

		  offEnd = pmsg->lmh + sizeof(JMH) + jmh.cbjsf;
		  if (offEnd == FmaGetFileSize(parea->hfileJhr))
			lmh = pmsg->lmh;

		  // Allocate space and check if ok

		  if (!DoAddSpace(parea->hfileJhr, &lmh, sizeof(JMH) + cbjsf))
			return FALSE;
		  else
			bAddJhr = TRUE;
		}

		// Check if new text will not fit into current place and 
		// allocate space at the end of text file

		if (cchTextTotal > jmh.cchText) {

		  // Check if this message text is already at eof and 
		  // set allocation start offset

		  offEnd = jmh.offText + jmh.cchText;
		  if (offEnd == FmaGetFileSize(parea->hfileJdt)) {
			lmt = jmh.offText;
		  }

		  // Allocate space and check if ok

		  if (!DoAddSpace(parea->hfileJdt, &lmt, cchTextTotal)) {
			if (bAddJhr) FmaSetFileSize(parea->hfileJhr, lmh);
			return FALSE;
		  } else
			bAddJdt = TRUE;
		}

		// Check if allocated a new header and if so, fix up the old one

		if (bAddJhr) {
		  JMH jmhSave = jmh;
		  SETFLAG(jmhSave.attr, JMA_DELETED);
		  jmhSave.offText = jmhSave.cchText = 0;
		  if (!JamSaveJmh(parea, &jmhSave, NULL, pmsg->lmh)) {
			if (bAddJhr) FmaSetFileSize(parea->hfileJhr, lmh);
			if (bAddJdt) FmaSetFileSize(parea->hfileJdt, lmt);
			return FALSE;
		  }
		}

		// Set message header writing controls

		if (bAddJhr) { 
		  pmsg->lmh = lmh;
		  pmsg->cbjsf = 0;
		  pmsg->cbjsfMax = cbjsf;
		} else {
		  pmsg->cbjsf = 0;
		  pmsg->cbjsfMax = jmh.cbjsf;
		}

		// Set message text writing controls

		if (bAddJdt) {
		  pmsg->lmt = lmt;
		  pmsg->lmtEnd = lmt;
		  pmsg->lmtMax = lmt + cchTextTotal;
		} else {
		  pmsg->lmt = jmh.offText;
		  pmsg->lmtEnd = jmh.offText;
		  pmsg->lmtMax = jmh.offText + jmh.cchText;
		}

		// Preserve message deleted flag so that we can
		// adjust cmsgActive when writing

		SETFLAGTO(pmsg->fs, JMF_READDELETED, jmh.attr & JMA_DELETED);
	  }
	} else 

	// Check if writing an existing message

	if (pmsg->mode == OPENMSG_WRITE || pmsg->mode == OPENMSG_READWRITE) {

	  ASSERT(JamCheckLmh(parea, pmsg->lmh));

	  // Load message header and check if ok

	  if (!JamLoadJmh(parea, &jmh, NULL, pmsg->lmh))
		return FALSE;

	  // Check if new subfields or text will not fit into current place

	  if (cbjsf != jmh.cbjsf || cchTextTotal > jmh.cchText) {
		FMAERROR(INVALID_SIZE);
		return FALSE;
	  }

	  // Set message header writing controls

	  pmsg->cbjsf = 0;
	  pmsg->cbjsfMax = jmh.cbjsf;

	  // Set message text writing controls

	  pmsg->lmt = jmh.offText;
	  pmsg->lmtEnd = jmh.offText;
	  pmsg->lmtMax = jmh.offText + jmh.cchText;

	  // Preserve message deleted flag so that we can
	  // adjust cmsgActive when writing

	  SETFLAGTO(pmsg->fs, JMF_READDELETED, jmh.attr & JMA_DELETED);

	} else {

	  // Don't allow writing if message has been opened in read mode

	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Set flag that we've got the writing frame

	SETFLAG(pmsg->fs, JMF_GOTFRAME);

	// Update current message cache

	JamSetCurMsgCache(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the frame to write the message to
//

  static BOOL FMASUB DoWriteJmh(PJMSG pmsg, PJMH pjmh, PJSF pjsf, ULONG cbjsf)
  {
	PJAREA parea = pmsg->parea;
	BOOL bNewJsf = FALSE;
	PJSF pjsfNew = NULL;

	ASSERT(parea->hfileJhr != -1);
	ASSERT(parea->bExclusive);

	ASSERT(JamCheckLmh(parea, pmsg->lmh));

	// Check if message header subfields are not specified and 
	// assume the cached ones, otherwise allocate a copy of
	// the new message header subfields unless they're empty

	if (!pjsf) {
	  pjsf = pmsg->pjsf;
	  cbjsf = pmsg->cbjsf;
	} else {
	  bNewJsf = TRUE;
	  if (!cbjsf) {
		pjsf = NULL;
	  } else {

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cbjsf)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate memory to hold subfield data and check if ok

		if ((pjsfNew = MemAlloc((int)cbjsf, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
Fail:     if (pjsfNew) MemFree(pjsfNew);
		  return FALSE;
		} else {
		  memcpy(pjsfNew, pjsf, (int)cbjsf);
		}
	  }
	}

	// Check if message header is not specified and 
	// assume the cached one, otherwise fix up the 
	// caller's message header

	if (!pjmh) {
	  pjmh = &pmsg->jmh;
	} else {

	  // These are the fixed ones, so force them now

	  pjmh->magic = JAM_MAGIC;
	  pjmh->revLevel = JAM_REVLEVEL;

	  // These depend on the subfields info so update only if we have it

	  if (pjsf) {
		pjmh->crcMsgId = DoGetSubfCrc(pjsf, cbjsf, JSF_MSGID);
		pjmh->crcReply = DoGetSubfCrc(pjsf, cbjsf, JSF_REPLYID);
	  }

	  // Fix up the null password crc

	  if (!pjmh->crcp) pjmh->crcp = (ULONG)-1;

	  // These are controlled by the api so copy from the message handle

	  pjmh->msgn = pmsg->msgn;
	  pjmh->offText = pmsg->lmt;
	  pjmh->cchText = pmsg->lmtEnd - pmsg->lmt;

	  // Check if caller does not specify subfields and use the subfields size 
	  // stored in message handle, otherwise make sure subfields info does not 
	  // exceed maximum size

	  if (!pjsf) {
		pjmh->cbjsf = pmsg->cbjsf;
	  } else {
		if (cbjsf > pmsg->cbjsfMax) {
		  FMAERROR(INVALID_SIZE);
		  goto Fail;
		} else 
		  pjmh->cbjsf = cbjsf;
	  }
	}

	// Save the message header and subfields if caller specified them

	if (JamSaveJmh(parea, pjmh, pjsf, pmsg->lmh))
	  RESETFLAG(pmsg->fs, JMF_SAVEJMH);
	else
	  goto Fail;

	// Set the writing flags

	SETFLAG(pmsg->fs, JMF_MSGHWRITTEN); 
	if (pjsf) SETFLAG(pmsg->fs, JMF_SUBFWRITTEN);

	// Check if message deleted status changes and update active 
	// message count then save the new message status

	if ((pmsg->fs & JMF_READDELETED) && !(pjmh->attr & JMA_DELETED)) {
	  ASSERT((LONG)parea->cmsgActive < JamGetMsgCount(parea));
	  parea->cmsgActive++; RESETFLAG(pmsg->fs, JMF_READDELETED);
	} else 
	if (!(pmsg->fs & JMF_READDELETED) && (pjmh->attr & JMA_DELETED)) {
	  ASSERT((LONG)parea->cmsgActive > 0);
	  parea->cmsgActive--; SETFLAG(pmsg->fs, JMF_READDELETED);
	}

	// Check if wrote a new message header and update the cached one

	if (pjmh != &pmsg->jmh) {
	  SETFLAG(pmsg->fs, JMF_GOTMSGH);
	  pmsg->jmh = *pjmh;
	}

	// Check if we wrote new subfields and update the cached ones

	if (bNewJsf) {
	  if (pmsg->pjsf) MemFree(pmsg->pjsf);
	  pmsg->pjsf = pjsfNew;
	  pmsg->cbjsf = cbjsf;
	  SETFLAG(pmsg->fs, JMF_GOTSUBF);
	}

#ifdef FMA_LOG_JAM_WRITE
Log("DoWriteJmh        : [%09Fp] msgn=%ld pjsf=%08lx cbjsf=%ld", pmsg, pmsg->msgn, pjsf, pmsg->jmh.cbjsf);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message text
//

  static BOOL FMASUB DoWriteText(PJMSG pmsg, PCH pchText, ULONG cchText, ULONG * poffText)
  {
	PJAREA parea = pmsg->parea;
	ULONG cbWritten;
	LONG offs;

	ASSERT(parea->hfileJdt != -1);
	ASSERT(parea->bExclusive);

	// Calculate the message text offset

	offs = pmsg->lmt; if (poffText) offs+= *poffText;

	// Check if message text we're going to write will 
	// run past the end of the frame

	if (offs + (LONG)cchText > pmsg->lmtMax) {
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message text and check if ok

	if (pchText && cchText) {

	  if (!FmaWriteFile(parea->hfileJdt, offs, pchText, cchText, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cchText);
	}

	// Update the message text length and update flags

	pmsg->lmtEnd = offs  + cchText;
	pmsg->jmh.cchText = pmsg->lmtEnd - pmsg->lmt;
	SETFLAG(pmsg->fs, JMF_SAVEJMH | JMF_TEXTWRITTEN);

	// Update caller's text offset

	if (poffText) *poffText+= cchText;

#ifdef FMA_LOG_JAM_WRITE
Log("DoWriteText       : [%09Fp] cchText=%ld", pmsg, cchText);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates the index entry associated with the message
//

  static BOOL FMASUB DoUpdateJdx(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	JDX jdx;

	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->bExclusive);

	ASSERT(pmsg->lmh != 0);

	// Set frame offset and hash values

	jdx.lmh = pmsg->lmh;
	jdx.crc = DoGetSubfCrc(pmsg->pjsf, pmsg->jmh.cbjsf, JSF_RECVRNAME);

	// Save index entry and check if ok

	if (!JamSetJdx(parea, &jdx, pmsg->msgn))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine writes message
//

  FMAAPI_(BOOL) JamWriteMsg(HMSG hmsg, PJMH pjmh, PJSF pjsf, ULONG cbjsf,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal)
  {
	PJAREA parea;
	PJMSG pmsg;

	// Validate parameters

	if ((pmsg = (PJMSG)hmsg) == NULL || !JamCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_JAM_WRITE
Log("JamWriteMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_JAM_WRITE
Log("\nJamWriteMsg      >: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	// Check if message is allowed to be written

	if (!JamCheckMsgWriteMode(pmsg))
	  goto Fail;

	// Check if there is no writing frame for this messsage

	if (!(pmsg->fs & JMF_GOTFRAME)) {

	  // We need a message header and subfields on 
	  // the first write to a new message

	  if (pmsg->mode == OPENMSG_CREATE)
		if (!pjmh || (cbjsf && !pjsf)) {
		  FMAERROR(INVALID_PARAM);
		  goto Fail;
		}

	  // Get the frame to write and save result

	  if (!DoGetWriteFrame(pmsg, cbjsf, cchTextTotal))
		goto Fail;
	}

	ASSERT(pmsg->lmh <= FmaGetFileSize(parea->hfileJhr));
	ASSERT(pmsg->lmt <= FmaGetFileSize(parea->hfileJdt));

	// Check if caller wants to write message header or subfields

	if (pjmh || pjsf) {
	  if (!DoWriteJmh(pmsg, pjmh, pjsf, cbjsf))
		goto Fail;

	  // Update index entry so that the message will be valid as soon
	  // as the message header and subfields are written

	  if (!DoUpdateJdx(pmsg))
		goto Fail;
	}

	// Check if caller wants to write message text

	if (pchText) {
	  if (!DoWriteText(pmsg, pchText, cchText, poffText))
		goto Fail;
	}

	// Update message header if necessary

	if (pmsg->fs & JMF_SAVEJMH) {
	  if (JamSaveJmh(parea, &pmsg->jmh, NULL, pmsg->lmh))
		RESETFLAG(pmsg->fs, JMF_SAVEJMH);
	  else
		return FALSE;
	}

#ifdef FMA_LOG_JAM_WRITE
Log("JamWriteMsg      <: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamWrite.c
/////////////////////////////////////////////////////////////////////////////
