/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt area routines module
//
// PktArea.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 17/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
//          17/Feb/2000 kvitekp - Added better umsgid support
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Macro to check if message number is valid

#define GOODMSGN(msgn) ((msgn) >= parea->msgnBase && (msgn) <= parea->msgnHigh)

  // Message map allocation 

#define MSGMAP_GROW      32             // Allocation grow granularity
#define MSGMAP_GROWTWEAK 4              // Allocation grow tweak factor

  // Forward declarations

  static VOID FMASUB DoCloseAreaFiles(PPAREA parea);

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates area object
//

  static PPAREA FMASUB DoCreateArea(PSZ pszPath, NETADDR * pnetAddr)
  {
	CHAR achPath[MAX_PATH];
	PPAREA parea;
	int cch;

	// Get area path length and check if ok

	if (!pszPath || (cch = strlen(pszPath)) == 0) {
BadP: FMAERROR(INVALID_PARAM);
	  return NULL;
	}

	// Get area path length and check if ok, then copy it to the local storage

	if (cch >= MAX_PATH - lengof(PKT_PKT_FILEEXT))
	  goto BadP;
	else
	  memcpy(achPath, pszPath, cch + 1);

	// Check if there is no extension specified and assume ".pkt"

	if (cch < lengof(PKT_PKT_FILEEXT) || 
	    achPath[cch - lengof(PKT_PKT_FILEEXT)] != '.') {
	  strcat(achPath, PKT_PKT_FILEEXT); 
	  cch+= lengof(PKT_PKT_FILEEXT);
	}

	// Create message area object and check if ok

	if ((parea = MemAlloc(sizeof(PAREA) + cch, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize new area object

	parea->magic = PAREA_MAGIC;
	parea->papi = PktGetApi();
	if (pnetAddr) parea->netAddr = *pnetAddr;
	memcpy(parea->achPath, achPath, cch + 1);
	parea->hfilePkt = -1;

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys area object
//

  static BOOL FMASUB DoDestroyArea(PPAREA parea)
  {
	// Check if we have area object and leave if not

	if (!parea) return FALSE;

	// Close associated files

	DoCloseAreaFiles(parea);

	// Free message map if any

	if (parea->apmi) MemFree(parea->apmi);

	// Reset message area object magic number

	parea->magic = ~PAREA_MAGIC;

	// Free message area object

	MemFree(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens the area files
//

  static BOOL FMASUB DoOpenAreaFiles(PPAREA parea, ULONG fl, UINT fOpen)
  {
	ASSERT(parea != NULL);

	// Make up the file open mode flags

	if (fl & FOA_READONLY)
	  fOpen|= FOPENFILE_READONLY;
	else
	  fOpen|= FOPENFILE_READWRITE;

	// Make up the file sharing flags

	if (fl & FOA_DENYALL)
	  fOpen|= FOPENFILE_DENYALL;
	else
	  fOpen|= FOPENFILE_DENYNONE;

	// Open the packet file and check if ok

	if (!FmaOpenFile(parea->achPath, fOpen, &parea->hfilePkt)) {
	  DoCloseAreaFiles(parea);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes the area files
//

  static VOID FMASUB DoCloseAreaFiles(PPAREA parea)
  {
	ASSERT(parea != NULL);

	// Close packet file

	FmaCloseFile(&parea->hfilePkt);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes area files
//

  static BOOL FMASUB DoDeleteAreaFiles(PPAREA parea)
  {
	BOOL bOk = TRUE;

	ASSERT(parea != NULL);

	// Make sure the files area closed

	DoCloseAreaFiles(parea);

	// Delete area files and check if ok

	if (!FmaDeleteFile(parea->achPath)) bOk = FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area directory
//

  static BOOL FMASUB DoMakeAreaDir(PPAREA parea)
  {
	CHAR achPath[MAX_PATH];
	PCH pch;

	ASSERT(parea != NULL);

	// Copy path to the local storage and chop the base name

	strcpy(achPath, parea->achPath);
	if ((pch = strrchr(achPath, PATH_SEP)) != NULL) {
	  *pch = '\0';
	} else
	  return TRUE;

	// Check if at root or directory exists and leave if so

	if (!achPath[0] || FmaDirExists(achPath))
	  return TRUE;

	// Create directory and check if ok

	if (!FmaMakeDir(achPath)) {
	  FMAERROR(CANNOT_CREATE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area
//

  static BOOL FMASUB DoCreateNewArea(PPAREA parea, ULONG fl)
  {
	ASSERT(parea != NULL);

	ASSERT(parea->hfilePkt == -1);

	// Create area directory if necessary

	DoMakeAreaDir(parea);

	// Create area files and check if ok

	if (!DoOpenAreaFiles(parea, fl, FOPENFILE_CREATE | FOPENFILE_TRUNCATE))
	  return FALSE;

	// Initialize packet header

	memset(&parea->pbh, 0, sizeof(PBH));

	// Set packet creation time

	FmaGetLocalDttm(&parea->dttm);

	// Save empty base header and packed end zeros. Note: order
	// of calls is important for asserts 

	if (!PktSetEnd(parea, sizeof(PBH)) || !PktSavePbh(parea)) {
	  DoDeleteAreaFiles(parea);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine clears existing area
//

  static BOOL FMASUB DoClearExistingArea(PPAREA parea, ULONG fl)
  {
	ASSERT(parea != NULL);

	ASSERT(parea->hfilePkt == -1);

	// Open and truncate area files and check if ok

	if (!DoOpenAreaFiles(parea, fl, FOPENFILE_TRUNCATE))
	  return FALSE;

	// Save empty base header and packed end zeros

	if (!PktSavePbh(parea) || !PktSetEnd(parea, -1)) {
	  DoCloseAreaFiles(parea);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine shrinks area message map allocation
//

  static BOOL FMASUB DoFixMsgMap(PPAREA parea)
  {
	UINT cpmi;
	PPMI apmi;

	ASSERT(parea->cmsg < UINT_MAX);

	// Shrink message map allocation to store exact number of messages

	if (parea->cpmi > (UINT)parea->cmsg) {
	  cpmi = (int)parea->cmsg;
	  if ((apmi = MemRealloc(parea->apmi, cpmi * sizeof(PMI))) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		parea->apmi = apmi;
		parea->cpmi = cpmi;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens area
//

  FMAAPI_(HAREA) PktOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr)
  {
	BOOL bAreaExists;
	PPAREA parea;

	// Assert known structure sizes to ensure correct structure packing

	ASSERT(sizeof(PBH) == 58);
	ASSERT(sizeof(PMH) == 34);

#ifdef FMA_LOG_PKT_AREA
Log("\nPktOpenArea      >: %s", pszPath);
#endif

	// Create new area object and check if ok

	if ((parea = DoCreateArea(pszPath, pnetAddr)) == NULL) {
Fail: DoDestroyArea(parea);

#ifdef FMA_LOG_PKT_AREA
Log("PktOpenArea      <: %s FAILED, error=%d", pszPath, FmaGetLastError());
#endif

	  return NULL;
	}

	// Check if area message base exists

	bAreaExists = FmaFileExists(parea->achPath);

	// Act upon an opening mode

	switch (fOpenMode) {

	  case OPENAREA_EXISTING:

		// Fail if does not exist, otherwise open

		if (!bAreaExists) {
		  FMAERROR(NOT_FOUND);
		  goto Fail;
		} else 
Open:   if (!DoOpenAreaFiles(parea, fl, 0)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_ALWAYS:

		// Create if does not exist, otherwise open

		if (!bAreaExists) {
		  if (!DoCreateNewArea(parea, fl))
			goto Fail;
		} else
		  goto Open;

		break;

	  case OPENAREA_CREATE_NEW:

		// Fail if exists, otherwise create new

		if (bAreaExists) {
		  FMAERROR(ALREADY_EXISTS);
		  goto Fail;
		} else
		if (!DoCreateNewArea(parea, fl)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_CREATE_ALWAYS:

		// Clear if exists, otherwise create new

		if (bAreaExists) {
		  if (!DoClearExistingArea(parea, fl)) {
			goto Fail;
		  }
		} else 
		if (!DoCreateNewArea(parea, fl)) {
		  goto Fail;
		}

		break;
	}

	ASSERT(parea->hfilePkt != -1);

	// Load base header and if failed, remove area files if
	// they were just created for the new area

	if (!PktLoadPbh(parea)) {
	  if (!bAreaExists) DoDeleteAreaFiles(parea);
	  goto Fail;
	}

	// Rescan message area and check if ok

	if (!PktRescanArea(parea)) {
	  goto Fail;
	}

	// Set up message area numbers

	parea->msgnBase = 1;
	parea->msgnHigh = parea->msgnBase + parea->cmsg - 1;

	// Increment open area count in a thread safe manner

	FmaInterlockedIncrement(&fmaRtm.nOpenAreas);

	ASSERT(PktCheckArea(parea));

	// Check if caller wants to lock the area upon opening

	if (parea->fl & FOA_LOCKAREA) {
	  if (!PktLockArea(parea, TRUE)) {
		PktCloseArea(parea);
		parea = NULL;
		goto Fail;
	  }
	}

#ifdef FMA_LOG_PKT_AREA
Log("PktOpenArea      <: [%09Fp] cmsg=%ld msgnHigh=%ld in %s", parea, parea->cmsg, parea->msgnHigh, parea->achPath);
#endif

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes area
//

  FMAAPI_(BOOL) PktCloseArea(HAREA harea)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_PKT_AREA
Log("\nPktOpenArea      >: [%09Fp] cmsg=%ld msgnHigh=%ld in %s", parea, parea->cmsg, parea->msgnHigh, parea->achPath);
#endif

	// Check if caller locked the area upon opening and
	// area is still locked

	if (parea->fl & FOA_LOCKAREA && parea->bAreaLocked) {
	  PktLockArea(parea, FALSE);
	}

	// Check if there are any open messages

	if (parea->pmsgList) {
	  FMAERROR(STILL_OPEN);
#ifdef FMA_LOG_PKT_AREA
Log("PktCloseArea      : [%09Fp] %ld msg(s) still open", parea, ListCount(parea->pmsgList)); 
#endif

	  // Close all open messages

	  while (parea->pmsgList) {
		PktCloseMsg(parea->pmsgList);
	  }
	}

	// Check if we have exclusive access

	while (parea->bExclusive) {
#ifdef FMA_LOG_PKT_AREA
Log("PktCloseArea      : [%09Fp] bExclusive=%d", parea, parea->bExclusive); 
#endif
	  PktLeaveExclusive(parea);
	}

	// Check if we have area locked

	while (parea->bAreaLocked) {
#ifdef FMA_LOG_PKT_AREA
Log("PktCloseArea      : [%09Fp] bAreaLocked=%d", parea, parea->bAreaLocked); 
#endif
	  PktLockArea(parea, FALSE);
	}

	// Destroy message area object

	DoDestroyArea(parea);

	// Decrement open area count in a thread safe manner

	FmaInterlockedDecrement(&fmaRtm.nOpenAreas);

#ifdef FMA_LOG_PKT_AREA
Log("PktCloseArea     <: [%09Fp] CLOSED", parea); 
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message
//

  FMAAPI_(BOOL) PktValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if (!parea->cmsg) return FALSE;

	ASSERT(parea->apmi != NULL);
	ASSERT(parea->cpmi != 0);
	ASSERT(parea->cmsg < UINT_MAX);

	// Check for the valid message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Check if caller needs message index and set it

	if (pimsg) *pimsg = (ULONG)(msgn - parea->msgnBase);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message map which is array of unique message ids
//

  FMAAPI_(BOOL) PktCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg)
  {
	ULONG cb, cumsg, iumsg;
	PPAREA parea;
	UMSG * aumsg;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(paumsg != NULL);
	ASSERT(pcumsg != NULL);

	// Check if no messages in this area and calculate message map size

	if ((cumsg = parea->cmsg) == 0 || !parea->apmi) {
	  *paumsg = NULL; *pcumsg = 0;
	  return TRUE;
	} else
	  cb = cumsg * sizeof(UMSG);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Allocate memory to hold message map and check if ok

	if ((aumsg = MemAlloc((int)cb, 0)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	ASSERT(parea->cmsg > 0);
	ASSERT(parea->apmi != NULL);

	// Copy message map

	for (iumsg = 0; iumsg < cumsg; iumsg++) {
	  aumsg[iumsg] = parea->apmi[iumsg].umsg;
	}

	// Set caller's pointers

	*paumsg = aumsg; *pcumsg = cumsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates messages
//

  FMAAPI_(BOOL) PktEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl)
  {
	BOOL bSafeEnum = (fl & FENUMMSGS_SAFEENUM) ? TRUE : FALSE;
	ULONG iumsg, cumsg;
	BOOL bOk = TRUE;
	ENUMMSGINFO emi;
	UMSG * pumsg;
	UMSG * aumsg;
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea) || !pfnCallback) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if (!parea->cmsg) return FALSE;

	ASSERT(parea->apmi != NULL);
	ASSERT(parea->cpmi > 0);

	// Since internal pkt message map is an array of ints we always create
	// message map even if not enumerating safely

	if (!PktCreateMsgMap(parea, &aumsg, &cumsg) || !cumsg)
	  return FALSE;

	// Prepare enum message info structure

	memset(&emi, 0, sizeof(emi));
	emi.harea = harea;

	// Check if need to enumerate in reversed order

	if (fl & FENUMMSGS_REVERSE) {

	  // Enumerate in reversed order

	  pumsg = aumsg + cumsg - 1;
	  for (iumsg = cumsg; iumsg > 0 ; iumsg--, pumsg--) {
		emi.msgn = parea->msgnBase + iumsg - 1; emi.imsg = iumsg - 1;
		if (bSafeEnum && !PktValidateMsg(harea, (MSGN)*pumsg, &emi.imsg)) continue;
		emi.umsg = parea->apmi[emi.imsg].umsg;
		emi.offs = parea->apmi[emi.imsg].lmh;
		bOk = pfnCallback(&emi, pvCallbackData);
		if (!bOk) break;
	  }
	} else {

	  // Enumerate in normal order

	  pumsg = aumsg;
	  for (iumsg = 0; iumsg < cumsg; iumsg++, pumsg++) {
		emi.msgn = parea->msgnBase + iumsg; emi.imsg = iumsg;
		if (bSafeEnum && !PktValidateMsg(harea, (MSGN)*pumsg, &emi.imsg)) continue;
		emi.umsg = parea->apmi[emi.imsg].umsg;
		emi.offs = parea->apmi[emi.imsg].lmh;
		bOk = pfnCallback(&emi, pvCallbackData);
		if (!bOk) break;
	  }
	}

	// Free temporary copy of the message map

	MemFree(aumsg);

	// Return whatever last callback returned

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message
//

  FMAAPI_(BOOL) PktDeleteMsg(HAREA harea, MSGN msgn)
  {
	PPAREA parea;
	BOOL bOk = FALSE;
	ULONG imsg;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Validate message number and check if ok

	if (!PktValidateMsg(harea, msgn, &imsg)) {
	  return FALSE;
	}

	// Check if this message is open and leave if so

	if (PktGetOpenMsg(parea, msgn)) {
	  FMAERROR(STILL_OPEN);
	  return FALSE;
	}

	// Acquire exclusive access and check if ok

	if (!PktEnterExclusive(parea))
	  return FALSE;

	// Delete message and check if ok

	if (!PktDelMsg(parea, imsg))
	  goto Out;

	// Adjust highest message number

	parea->msgnHigh--;

	// Adjust current message number

	if (parea->msgnCur == msgn) {
	  parea->msgnCur = 0;
	} else
	if (parea->msgnCur > msgn) {
	  parea->msgnCur--;
	}

	// If we got here everything is fine

	bOk = TRUE;

	// Relinquish exclusive access and check if ok

Out:if (!PktLeaveExclusive(parea))
	  bOk = FALSE;

#ifdef FMA_LOG_PKT_AREA
Log("PktDeleteMsg: %s msgn=%lu", parea->achPath, msgn);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets current message
//

  FMAAPI_(BOOL) PktSetCurMsg(HAREA harea, MSGN msgn)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(msgn <= parea->msgnHigh);

	// Set current message

	parea->msgnCur = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets current message
//

  FMAAPI_(MSGN) PktGetCurMsg(HAREA harea)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(parea->msgnCur <= parea->msgnHigh);

	// Return current message

	return parea->msgnCur;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lowest message number
//

  FMAAPI_(MSGN) PktGetFirstMsg(HAREA harea)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we have no messages

	if (!parea->cmsg) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	ASSERT(parea->cpmi > 0);
	ASSERT(parea->apmi != NULL);

	// Return lowest message number

	return parea->msgnBase;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets highest message number
//

  FMAAPI_(MSGN) PktGetHighMsg(HAREA harea)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return highest message number

	return parea->msgnHigh;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message count
//

  FMAAPI_(LONG) PktGetMsgCount(HAREA harea)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return message count

	return parea->cmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts message number to unique message id
//

  FMAAPI_(UMSG) PktMsgnToUid(HAREA harea, MSGN msgn)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	ASSERT(parea->cmsg > 0);
	ASSERT(parea->apmi != NULL);

	// Return message uid

	return parea->apmi[msgn - parea->msgnBase].umsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts unique message id to message number
//

  FMAAPI_(MSGN) PktUidToMsgn(HAREA harea, UMSG umsg, UINT fMode)
  {
	int imsg, imsgLo, imsgHi, imsgLast; // these have to be signed for bsearch
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message unique id

	if (!umsg || !parea->cmsg) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Calculate very last message index

	imsgLast = (int)parea->cmsg - 1;

	// Run binary search looking for umsgid match

	imsg = 0; imsgLo = 0; imsgHi = imsgLast;

	while (imsgLo <= imsgHi) {

	  // Split the current range in the middle

	  imsg = (imsgLo + imsgHi) / 2;

	  ASSERT(imsg >= 0);
	  ASSERT(imsg <= imsgLast);

	  // Check if we got a match and break

	  if (umsg == parea->apmi[imsg].umsg)
		break;

	  // Select a new range

	  if (umsg > parea->apmi[imsg].umsg)
		imsgLo = imsg + 1;
	  else
		imsgHi = imsg - 1;
	}

	ASSERT(imsg >= 0);
	ASSERT(imsg <= imsgLast);

	// Check if we got an exact match and return message number
	// if caller wants an exact one

	if (parea->apmi[imsg].umsg == umsg) {
	  if (fMode & UMSG_EXACT || !fMode)
		return parea->msgnBase + imsg;
	}

	// Check if the caller wants next or previous message

	if (fMode & UMSG_NEXT) {
	  if (parea->apmi[imsg].umsg > umsg)
		return parea->msgnBase + imsg;
	  else
		return imsg < imsgLast ? parea->msgnBase + imsg + 1 : 0;
	} else
	if (fMode & UMSG_PREV) {
	  if (parea->apmi[imsg].umsg < umsg)
		return parea->msgnBase + imsg;
	  else
		return imsg > 0 ? parea->msgnBase + imsg - 1 : 0;
	} 

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message area object
//

  FMAAPI_(BOOL) PktCheckArea(PPAREA parea)
  {
	ASSERT(parea != NULL);

	// Check the message area object magic number

	if (parea->magic != PAREA_MAGIC)
	  return FALSE;

	// Debug build only checks

#ifdef _DEBUG
	if (parea->cmsg) {
	  ASSERT(parea->cpmi > 0 && parea->apmi != NULL);
	} else {
	  ASSERT(parea->msgnHigh == 0);
	}
#endif

	// Check message map

#ifdef _DEBUG
	if (parea->apmi) {
	  LONG lMsgEnd;
	  UINT imsg;

	  for (imsg = 0; imsg < parea->cmsg; imsg++) {

		if (!parea->apmi[imsg].lmh) {
		  ASSERT(imsg == parea->cmsg - 1);
		  ASSERT(parea->apmi[imsg].cbHdr == 0);
		  ASSERT(parea->apmi[imsg].cbMsg == 0);
		} else {

		  ASSERT(parea->apmi[imsg].cbHdr > sizeof(PMH) + 3);    // Three trailing nulls for By,To,Re
		  ASSERT(parea->apmi[imsg].cbHdr < parea->apmi[imsg].cbMsg);

		  if (imsg == 0) {
			ASSERT(parea->apmi[0].lmh == (LONG)sizeof(PBH));
		  } else {
			lMsgEnd = parea->apmi[imsg - 1].lmh + parea->apmi[imsg - 1].cbMsg;
			ASSERT(lMsgEnd == parea->apmi[imsg].lmh);
		  }

		  // The following would fail under memory stress test for good reason,
		  // that is messaeg gets allocated on disk but message map does not
		  // get expanded. Missing this test now should not compromize pkt file
		  // integrity through, we're just not seeing last message ...

#ifndef FMA_STRESS_TEST
		  if (imsg == parea->cmsg - 1) {
			lMsgEnd = parea->apmi[imsg].lmh + parea->apmi[imsg].cbMsg;
			ASSERT(lMsgEnd == FmaGetFileSize(parea->hfilePkt) - (LONG)sizeof(USHORT));
		  }
#endif
		}
	  }
	}
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine rescans area
//

  FMAAPI_(BOOL) PktRescanArea(PPAREA parea)
  {
	ULONG cbRead, cbHdr, cbText;
	CHAR achText[512];
	LONG lmh, lmt;
	PCH pch;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

#ifdef FMA_LOG_PKT_AREA
//Log("PktRescanArea: %s", parea->achPath);
#endif

	// Reset existing message map if any

	parea->cmsg = 0; parea->cpmi = 0; parea->msgnHigh = 0;
	if (parea->apmi) MemFree(parea->apmi); parea->apmi = NULL;

	// Loop scanning packed messages

	for (lmh = sizeof(PBH); PktLoadPmh(parea, NULL, lmh); lmh+= cbHdr + cbText) {

	  // Get packed message text offset and calculate size of 
	  // the packed message header

	  lmt = FmaGetFilePos(parea->hfilePkt);
	  cbHdr = lmt - lmh;

	  // Loop reading message text up to its terminating null

	  for (cbText = 0; ; lmt+= cbRead, cbText+= cbRead) {

		// Read portion of the message text and check if there is
		// a terminating null in this portion

		if (FmaReadFile(parea->hfilePkt, lmt, achText, sizeof(achText), &cbRead)) {
		  if ((pch = memchr(achText, 0, cbRead)) != NULL) {
			cbText+= (ULONG)(pch - achText) + 1;
			break;
		  }
		} else
		  break;
	  }

	  // Add new message map entry

	  if (!PktAddMsg(parea, lmh, cbHdr, cbHdr + cbText))
		return FALSE;
	}

	// Shrink message map allocation

	DoFixMsgMap(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds a message to the end of area message map
//

  FMAAPI_(BOOL) PktAddMsg(PPAREA parea, LONG lmh, ULONG cbHdr, ULONG cbMsg)
  {
	UINT cpmi, cpmiGrow = MSGMAP_GROW;
	PPMI apmi, ppmi;

	// Check if there is no room for a new message and reallocate message map

	if (!(parea->cmsg < parea->cpmi)) {

	  // Tweak allocation grow factor to prevent excessive heap fragmentation

	  if (parea->cpmi > MSGMAP_GROW * MSGMAP_GROWTWEAK) {
		cpmiGrow+= parea->cpmi / MSGMAP_GROWTWEAK;
	  }

	  // Reallocate message map buffer and check if ok

	  cpmi = parea->cpmi + cpmiGrow;
	  if ((apmi = MemRealloc(parea->apmi, cpmi * sizeof(PMI))) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		parea->cpmi = cpmi;
		parea->apmi = apmi;
	  }
	}

	// Add the new message to the message map

	ppmi = &parea->apmi[parea->cmsg];
	ppmi->lmh = lmh;
	ppmi->cbHdr = cbHdr;
	ppmi->cbMsg = cbMsg;
	ppmi->umsg = parea->cmsg ? parea->apmi[parea->cmsg - 1].umsg + 1 : 1;

	// Increment message count

	parea->cmsg++;

#ifdef FMA_LOG_PKT_AREA
//Log("PktAddMsg: cmsg=%lu cpmi=%u umsg= %ld pmi=%ld/%ld/%ld", parea->cmsg, parea->cpmi, ppmi->umsg, ppmi->lmh, ppmi->cbHdr, ppmi->cbMsg);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktArea.c
/////////////////////////////////////////////////////////////////////////////
