/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt helper routines module
//
// PktHelp.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 08/Sep/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
//          18/Feb/2000 kvitekp - Added PktSet/CheckPassword()
//
//          18/Feb/2000 kvitekp - Fixed pbh orig/dest zero point bug
//
//          08/Sep/2000 kvitekp - Fixed dest node bug in PktLoad/SavePmh
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "PktApi.h"
#include "PktApiP.h"

#include "FmaVer.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Following is used for stress testing only

#ifdef FMA_STRESS_TEST
  FMAAPI_(BOOL) FmaQueryDiskAllocFailure(HFILE hfile, ULONG cb);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine sets network address given it's components
//

  static void FMASUB DoSetNetAddr(NETADDR * pnetAddr, USHORT zone, USHORT net, 
								  USHORT node, USHORT point)
  {
	ASSERT(pnetAddr != NULL);

	pnetAddr->zone  = zone; 
	pnetAddr->net   = net;
	pnetAddr->node  = node;
	pnetAddr->point = point;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets dttm given packet header date/time info
//

  static void FMASUB DoPbhToDttm(PPBH ppbh, PDTTM pdttm)
  {
	ASSERT(ppbh != NULL);
	ASSERT(pdttm != NULL);

	pdttm->da = ppbh->day;
	pdttm->mo = ppbh->month + 1;
	pdttm->yr = ppbh->year - 1980;

	pdttm->hh = ppbh->hour;
	pdttm->mm = ppbh->minute;
	pdttm->ss = ppbh->second / 2;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets packet header date/time info given dttm 
//

  static void FMASUB DoDttmToPbh(PPBH ppbh, PDTTM pdttm)
  {
	ASSERT(ppbh != NULL);
	ASSERT(pdttm != NULL);

	ppbh->day    = (USHORT)(pdttm->da);
	ppbh->month  = (USHORT)(pdttm->mo - 1);
	ppbh->year   = (USHORT)(pdttm->yr + 1980);

	ppbh->hour   = (USHORT)(pdttm->hh);
	ppbh->minute = (USHORT)(pdttm->mm);
	ppbh->second = (USHORT)(pdttm->ss * 2);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks capability word validity
//

  static BOOL FMASUB DoCheckCapWord(PPBH ppbh)
  {
	USHORT capWordCopy;

	ASSERT(ppbh != NULL);

	// Get byte swapped capability word copy

	capWordCopy = (USHORT)((ppbh->capWordCopy << 8) | (ppbh->capWordCopy >> 8));

	// Check if matches with the capability word elliminating RFC-822 bit

	return (capWordCopy & ~PCAP_RFC822) == (ppbh->capWord & ~PCAP_RFC822);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets packet info
//

  static void FMASUB DoGetPbhInfo(PPAREA parea)
  {
	PBH pbh = parea->pbh;

	// First treat it as StoneAge packet

	DoSetNetAddr(&parea->netOrig, pbh.origZone0, pbh.origNet, pbh.origNode, pbh.origPoint);
	DoSetNetAddr(&parea->netDest, pbh.destZone0, pbh.destNet, pbh.destNode, pbh.destPoint);

	DoPbhToDttm(&pbh, &parea->dttm);

	// Check capability word and if not valid or capabilities word is zero,
	// or indicates no 2+ support, leave as stone age

	if (!DoCheckCapWord(&pbh) || !pbh.capWord || !(pbh.capWord & PCAP_TYPE2PLUS)) 
	  return;

	// Check if origination network is in auxiliary place and if so, 
	// update origination network address

	if (pbh.origPoint && pbh.origNet == (USHORT)-1)
	  parea->netOrig.net = pbh.auxNet;

	// Pick up the modern zone numbers if any so that they supersede old ones

	if (pbh.origZone) parea->netOrig.zone = pbh.origZone;
	if (pbh.destZone) parea->netDest.zone = pbh.destZone;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets packet info
//

  static void FMASUB DoSetPbhInfo(PPAREA parea)
  {
	PBH pbh = parea->pbh;

	// Set packet type

	pbh.type = PKT_TYPE2;

	// Set capability word and its byte swapped copy

	pbh.capWord = PCAP_TYPE2PLUS;
	pbh.capWordCopy = (USHORT)((pbh.capWord << 8) | (pbh.capWord >> 8));

	// Set product code unless already specified

	if (!pbh.prodCodeLow && !pbh.prodCodeHigh) {
	  pbh.prodCodeLow  = (BYTE)(FMA_PRODCODE & 0x0ff);
	  pbh.prodCodeHigh = (BYTE)(FMA_PRODCODE >> 8);
	}

	// Set product revision unless already specified

	if (!pbh.prodRevMajor && !pbh.prodRevMinor) {
	  pbh.prodRevMajor = rmj;
	  pbh.prodRevMinor = rmm;
	}

	// Set origination address info

	pbh.origZone  = 
	pbh.origZone0 = parea->netOrig.zone;
	pbh.origNet   = parea->netOrig.net;
	pbh.origNode  = parea->netOrig.node;
	pbh.origPoint = parea->netOrig.point;

	// Check if from point and move origination network number
	// to auxiliary location. NOTE: FTS-001.016 requires this 
	// but Squish does not interpret auxNet, thus removed...

//	if (pbh.origPoint) {
	  pbh.auxNet  = pbh.origNet;
//	  pbh.origNet = (USHORT)-1;
//	} 

	// Set destination address info

	pbh.destZone  = 
	pbh.destZone0 = parea->netDest.zone;
	pbh.destNet   = parea->netDest.net;
	pbh.destNode  = parea->netDest.node;
	pbh.destPoint = parea->netDest.point;

	// Set date and time info

	DoDttmToPbh(&pbh, &parea->dttm);

	// Leave baud rate and packet password as is

//  pbh.baud;
//  pbh.achPassword;

	// Update stored packet header

	parea->pbh = pbh;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads Pkt message base header
//

  FMAAPI_(BOOL) PktLoadPbh(PPAREA parea)
  {
	ULONG cbRead;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

	// Seek to the beginning of the packet file and load packet header

	if (!FmaReadFile(parea->hfilePkt, 0, &parea->pbh, sizeof(PBH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(PBH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check the packet header

	if (!PktCheckPbh(parea))
	  return FALSE;

	// Get packet packet info

	DoGetPbhInfo(parea);

#ifdef FMA_LOG_PKT_HELP
Log("PktLoadPbh        : [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves Pkt message base header
//

  FMAAPI_(BOOL) PktSavePbh(PPAREA parea)
  {
	ULONG cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive || !parea->pbh.type);

	// Set packet packet info

	DoSetPbhInfo(parea);

	// Check the packet header

	if (!PktCheckPbh(parea)) {
	  return FALSE;
	}

	// Seek to the beginning of the packet file and save packet header

	if (!FmaWriteFile(parea->hfilePkt, 0, &parea->pbh, sizeof(PBH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(PBH));

#ifdef FMA_LOG_PKT_HELP
Log("PktSavePbh        : [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks Pkt message base header
//

  FMAAPI_(BOOL) PktCheckPbh(PPAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != 1);

	// Check packet type

	if (parea->pbh.type != PKT_TYPE2) {
	  FMAERROR(INVALID_DATA);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets Pkt message base end
//

  FMAAPI_(BOOL) PktSetEnd(PPAREA parea, LONG offs)
  {
	static USHORT zeroWord;
	ULONG cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != 1);
	ASSERT(parea->bExclusive || !parea->pbh.type);

	// Check if offset is not specified and assume current file position

	if (offs == -1) {
	  if ((offs = FmaGetFilePos(parea->hfilePkt)) == -1)
		return FALSE;
	}

	// Seek to the desired offset and write packet end nulls

	if (!FmaWriteFile(parea->hfilePkt, offs, &zeroWord, sizeof(zeroWord), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(zeroWord));

	// Chop the file past the packet end nulls

	if (!FmaSetFileSize(parea->hfilePkt, offs + sizeof(zeroWord)))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine resets the current message cache
//

  FMAAPI_(BOOL) PktResetCurMsgCache(PPAREA parea)
  {
	ASSERT(parea != NULL);

	// Reset current message cache

	parea->msgnCur = 0;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the current message cache
//

  FMAAPI_(BOOL) PktSetCurMsgCache(PPMSG pmsg)
  {
	PPAREA parea;

	ASSERT(pmsg != NULL);

	parea = pmsg->parea;

	// Set current message cache

	parea->msgnCur = pmsg->msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  FMAAPI_(BOOL) PktLoadPmh(PPAREA parea, PFMH pfmh, LONG lmh)
  {
	ULONG cbRead;
	LONG offs;
	UINT cch;
	PMH pmh;
	FMH fmh;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

	ASSERT(lmh >= (LONG)sizeof(PBH));
	ASSERT(lmh <= (LONG)FmaGetFileSize(parea->hfilePkt));

	// Check if positioned at the packed end terminator and 
	// leave without setting error condition if so

	if (lmh == (LONG)(FmaGetFileSize(parea->hfilePkt) - sizeof(USHORT))) {
	  FmaSetFilePos(parea->hfilePkt, lmh);
	  return FALSE;
	}

	// Read the fixed portion of packed message header and check if ok

	if (!FmaReadFile(parea->hfilePkt, lmh, &pmh, sizeof(PMH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(PMH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check packed message header version

	if (pmh.type != PMH_TYPE2) {
	  FMAERROR(INVALID_DATA);
	  return FALSE;
	}

	// Clear target message header

	memset(&fmh, 0, sizeof(fmh));

	// Copy fixed portion of packed message header into taget message header

	fmh.origNode = pmh.origNode;
	fmh.destNode = pmh.destNode;
	fmh.origNet = pmh.origNet;
	fmh.destNet = pmh.destNet;
	fmh.attr = pmh.attr;
	fmh.cost = pmh.cost;

	// Set message written time

	memcpy(fmh.achDate, pmh.achDate, sizeof(fmh.achDate));
	FmaAsciiToDttm(fmh.achDate, &fmh.dttmWritten);

	// Load packed message from name

	offs = lmh + sizeof(PMH);
	if (FmaReadFile(parea->hfilePkt, offs, fmh.achTo, lengof(fmh.achTo), &cbRead)) {
	  cch = strlen(fmh.achTo); memset (fmh.achTo + cch, 0, sizeof(fmh.achTo) - cch);
	} else
	  return FALSE;

	// Load packed message to name

	offs+= cch + 1;
	if (FmaReadFile(parea->hfilePkt, offs, fmh.achBy, lengof(fmh.achBy), &cbRead)) {
	  cch = strlen(fmh.achBy); memset (fmh.achBy + cch, 0, sizeof(fmh.achBy) - cch);
	} else
	  return FALSE;

	// Load packed message subject

	offs+= cch + 1;
	if (FmaReadFile(parea->hfilePkt, offs, fmh.achRe, lengof(fmh.achRe), &cbRead)) {
	  cch = strlen(fmh.achRe); memset (fmh.achRe + cch, 0, sizeof(fmh.achRe) - cch);
	} else
	  return FALSE;

	// Position packet file pointer to the message text beginning

	offs+= cch + 1; FmaSetFilePos(parea->hfilePkt, offs);

	// Update caller's message header

	if (pfmh) *pfmh = fmh;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  FMAAPI_(BOOL) PktSavePmh(PPAREA parea, PFMH pfmh, LONG lmh)
  {
	ULONG cbWritten;
	UINT cb, cch;
	LONG offs;
	PMH pmh;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	ASSERT(lmh >= (LONG)sizeof(PBH));
	ASSERT(lmh <= (LONG)FmaGetFileSize(parea->hfilePkt));

	// Clear packed message header

	memset(&pmh, 0, sizeof(pmh));

	// Initalize packed message header

	pmh.type = PMH_TYPE2;
	pmh.origNode = pfmh->origNode;
	pmh.destNode = pfmh->destNode;
	pmh.origNet = pfmh->origNet;
	pmh.destNet = pfmh->destNet;
	pmh.attr = pfmh->attr;
	pmh.cost = pfmh->cost;

	// Set message written time

	if (pfmh->achDate[0]) {
	  memcpy(pmh.achDate, pfmh->achDate, sizeof(pmh.achDate));
	} else {
	  cch = FmaFormatDttm(pmh.achDate, sizeof(pmh.achDate), &pfmh->dttmWritten, 
						  FDTTM_TIMESECONDS, ' ', ':', "  ");

	  ASSERT(cch == lengof(pfmh->achDate));
	}

	// Write the fixed portion of packed message header and check if ok

	if (!FmaWriteFile(parea->hfilePkt, lmh, &pmh, sizeof(PMH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(PMH));

	// Write packed message from name

	offs = lmh + sizeof(PMH); cb = strlen(pfmh->achTo) + 1;
	if (!FmaWriteFile(parea->hfilePkt, offs, pfmh->achTo, cb, &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == cb);

	// Write packed message to name

	offs+= cb; cb = strlen(pfmh->achBy) + 1;
	if (!FmaWriteFile(parea->hfilePkt, offs, pfmh->achBy, cb, &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == cb);

	// Write packed message subject

	offs+= cb; cb = strlen(pfmh->achRe) + 1;
	if (!FmaWriteFile(parea->hfilePkt, offs, pfmh->achRe, cb, &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == cb);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine shrinks packet file
//

  FMAAPI_(BOOL) PktShrinkPkt(PPAREA parea, LONG offs, ULONG cb)
  {
	ULONG cbFile, cbMove, cbRead = 0, cbWritten = 0;
	UINT cbChunk, imsg;
	LONG lFrom, lTo;
	CHAR ach[512];
	PPMSG pmsg;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	ASSERT(offs >= (LONG)sizeof(PBH));
	ASSERT(offs <= (LONG)FmaGetFileSize(parea->hfilePkt));
	ASSERT(offs + (LONG)cb <= (LONG)FmaGetFileSize(parea->hfilePkt));

	// Check if nothing to do

	if (!cb) return TRUE;

	// Calculate starting offsets and number of bytes to move

	lFrom = offs + cb; lTo = offs;
	cbFile = FmaGetFileSize(parea->hfilePkt);
	cbMove = cbFile - lFrom;

	// Loop forward moving the subsequent data over the deleted one

	while (cbMove > 0) {

	  // Read portion of the packet file and check if ok

	  cbChunk = (UINT)min(cbMove, sizeof(ach));
	  if (!FmaReadFile(parea->hfilePkt, lFrom, ach, cbChunk, &cbRead))
		return FALSE;

	  // Write portion of the packet file and check if ok

	  if (!FmaWriteFile(parea->hfilePkt, lTo, ach, cbRead, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cbRead);

	  // Advance to the next portion

	  lFrom+= cbRead; lTo+= cbRead; cbMove-= cbRead;
	}

	// Chop the packet file size

	FmaSetFileSize(parea->hfilePkt, cbFile - cb);

	// Adjust message map offsets past the deleted range

	lFrom = offs + cb;
	for (imsg = 0; imsg < parea->cmsg; imsg++) {
	  if (parea->apmi[imsg].lmh >= lFrom)
		parea->apmi[imsg].lmh-= cb;
	}

	// Adjust all the open messages stored offsets

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->lmh >= lFrom) pmsg->lmh-= cb;
	  if (pmsg->lmt >= lFrom) pmsg->lmt-= cb;
	  if (pmsg->lmtEnd >= lFrom) pmsg->lmtEnd-= cb;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine expands packet file
//

  FMAAPI_(BOOL) PktExpandPkt(PPAREA parea, LONG offs, ULONG cb)
  {
	ULONG cbRead = 0, cbWritten = 0;
	LONG lFrom, lTo, lEnd;
	UINT cbChunk, imsg;
	CHAR ach[512];
	PPMSG pmsg;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	ASSERT(offs >= (LONG)sizeof(PBH));
	ASSERT(offs <= (LONG)FmaGetFileSize(parea->hfilePkt));

	// Check if nothing to do

	if (!cb) return TRUE;

	// Following is used for stress testing only

#ifdef FMA_STRESS_TEST
	if (FmaQueryDiskAllocFailure(parea->hfilePkt, cb)) return FALSE;
#endif

	// Get starting offset

	lEnd = FmaGetFileSize(parea->hfilePkt);

	// Loop backward moving the subsequent data 

	while (lEnd > offs) {

	  // Calculate size of the portion to move

	  if (lEnd - offs > (LONG)sizeof(ach))
		cbChunk = sizeof(ach);
	  else
		cbChunk = (UINT)(lEnd - offs);

	  // Read portion of the packet file and check if ok

	  lFrom = lEnd - cbChunk;
	  if (!FmaReadFile(parea->hfilePkt, lFrom, ach, cbChunk, &cbRead))
		return FALSE;

	  // Write portion of the packet file and check if ok

	  lTo = lFrom + cb;
	  if (!FmaWriteFile(parea->hfilePkt, lTo, ach, cbRead, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cbRead);

	  // Get back for another portion

	  lEnd-= cbRead;
	}

	// Overwrite expanded porion of the file. Note that this also
	// appends the file if necessary

	memset(ach, 0, sizeof(ach));
	for (lTo = offs, lEnd = offs + cb; lTo < lEnd; lTo+= cbWritten) {
	  cbChunk = (UINT)min(lEnd - lTo, (LONG)sizeof(ach));
	  if (!FmaWriteFile(parea->hfilePkt, lTo, ach, cbChunk, &cbWritten))
		return FALSE;
	}

	// Adjust message map offsets past the deleted range

	for (imsg = 0; imsg < parea->cmsg; imsg++) {
	  if (parea->apmi[imsg].lmh >= offs)
		parea->apmi[imsg].lmh+= cb;
	}

	// Adjust all the open messages stored offsets

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->lmh >= offs) pmsg->lmh+= cb;
	  if (pmsg->lmt >= offs) pmsg->lmt+= cb;
	  if (pmsg->lmtEnd >= offs) pmsg->lmtEnd+= cb;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message at the specified frame offset
//

  FMAAPI_(BOOL) PktDelMsg(PPAREA parea, UINT imsg)
  {
	LONG offs, cb;
	UINT cbMove;
	PPMSG pmsg;
	MSGN msgn;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	ASSERT(parea->apmi != NULL);
	ASSERT(parea->cmsg > 0);
	ASSERT(imsg < parea->cmsg);

	// Get the packed message frame

	offs = parea->apmi[imsg].lmh;
	cb = parea->apmi[imsg].cbMsg;

	// Shrink packet file and check if ok

	if (!PktShrinkPkt(parea, offs, cb))
	  return FALSE;

	// Remove entry from the message map

	cbMove = (int)(parea->cmsg - (imsg + 1)) * sizeof(PMI);
	memmove(&parea->apmi[imsg], &parea->apmi[imsg + 1], cbMove);
	parea->cmsg--;

#ifdef _DEBUG
	ASSERT((ULONG)parea->cpmi >= parea->cmsg);
	cbMove = (int)(parea->cpmi - parea->cmsg) * sizeof(PMI);
	memset(&parea->apmi[parea->cmsg], -1, cbMove);
#endif

	// Adjust stored message numbers in all open messages
	
	msgn = parea->msgnBase + imsg;
	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->msgn > msgn) pmsg->msgn--;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  FMAAPI_(PPMSG) PktGetOpenMsg(PPAREA parea, MSGN msgn)
  {
	PPMSG pmsg;

	ASSERT(parea  != NULL);

	// Scan through the open messages list looking for a match

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->msgn == msgn)
		return pmsg;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets packet origination and destination addresses
//

  FMAAPI_(BOOL) PktSetNetAddr(PPAREA parea, NETADDR * pnetOrig, NETADDR * pnetDest)
  {
	ASSERT(parea != NULL);

	if (pnetOrig) parea->netOrig = *pnetOrig;
	if (pnetDest) parea->netDest = *pnetDest;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets packet origination and destination addresses
//

  FMAAPI_(BOOL) PktGetNetAddr(PPAREA parea, NETADDR * pnetOrig, NETADDR * pnetDest)
  {
	ASSERT(parea != NULL);

	if (pnetOrig) *pnetOrig = parea->netOrig;
	if (pnetDest) *pnetDest = parea->netDest;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets packet password
//

  FMAAPI_(BOOL) PktSetPassword(PPAREA parea, PSZ pszPassword)
  {
	ASSERT(parea != NULL);

	// Reset existing password if any

	memset(parea->pbh.achPassword, 0, sizeof(parea->pbh.achPassword));

	// Set new password if any

	if (pszPassword) {
	  strncpy(parea->pbh.achPassword, pszPassword, sizeof(parea->pbh.achPassword));
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks packet password
//

  FMAAPI_(BOOL) PktCheckPassword(PPAREA parea, PSZ pszPassword)
  {
	UINT ich;
	  
	ASSERT(parea != NULL);

	// Check if no password

	if (!parea->pbh.achPassword[0] && (!pszPassword || !pszPassword[0]))
	  return TRUE;

	// Compare password

	for (ich = 0; ich < sizeof(parea->pbh.achPassword); ich++) {
	  if (parea->pbh.achPassword[ich] != pszPassword[ich])
		return FALSE;
	  if (parea->pbh.achPassword[ich] == '\0')
		return TRUE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktHelp.c
/////////////////////////////////////////////////////////////////////////////
