/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt locking routines module
//
// PktLock.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 08/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////
//
//  PktLockBase()       Locks the message base files so that no other processes
//                      could access it
//
//  PktEnterExclusive() Locks message base files and loads base header in order
//                      to modify base contents. Under no circumstances message 
//                      base files can be updated outside the exclusive state.
//  
//  PktLockArea()       Locks message base files and caches index for speedy 
//                      access. In order to modify message base you still
//                      need to acquire exclusive state.
//  
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine locks or unlocks Pkt message base
//

  FMAAPI_(BOOL) PktLockBase(PPAREA parea, BOOL bLock)
  {
	int nTry;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   >: [%09Fp] bBaseLocked=%d", bLock, parea, parea->bBaseLocked);
#endif

	if (bLock) {

	  // Check if already locked and leave incrementing the lock count

	  if (parea->bBaseLocked) {
		parea->bBaseLocked++;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d ALREADY LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  } 

	  ASSERT(!parea->bBaseLocked);

	  // Loop trying to lock the base

	  for  (nTry = 0; nTry < PBH_LOCK_RETRY; nTry++) {

		// Lock the base and check if ok

		if (FmaLock(parea->hfilePkt, PBH_LOCK_OFFS, PBH_LOCK_LENG, TRUE)) {
		  parea->bBaseLocked++;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		  return TRUE;
		}

		// Sleep for a second a retry

		FmaSleep(1000);
	  }

	  // If we got here the base can't be locked

#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d LOCK FAILED", bLock, parea, parea->bBaseLocked);
#endif

	} else {

	  // Check if base is not locked 

	  if (!parea->bBaseLocked) {
		FMAERROR(NOT_LOCKED);
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d NOT LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return FALSE;
	  }

	  // Check if locked more than once and leave decrementing the lock count

	  if (parea->bBaseLocked > 1) {
		parea->bBaseLocked--;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d STILL LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  } 

	  ASSERT(parea->bBaseLocked == 1);

	  // Unlock the base and check if ok

	  if (FmaLock(parea->hfilePkt, PBH_LOCK_OFFS, PBH_LOCK_LENG, FALSE)) {
		parea->bBaseLocked--;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d UNLOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  }

	  // If we got here the base can't be unlocked

#ifdef FMA_LOG_PKT_LOCK
Log("PktLockBase(%d)   <: [%09Fp] bBaseLocked=%d UNLOCK FAILED", bLock, parea, parea->bBaseLocked);
#endif
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine acquires exclusive access to the message base
//

  FMAAPI_(BOOL) PktEnterExclusive(PPAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive>: [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Check if area is already exclusive and leave incremeting exclusive count

	if (parea->bExclusive) {
	  parea->bExclusive++;
#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d ALREADY EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return TRUE;
	}

	// Lock message base and check if ok

	if (!PktLockBase(parea, TRUE)) {
#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT LOCK", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	ASSERT(!parea->bExclusive);

	// Increment exclusive count early so that 
	// PktLoadPbh() knows we're exclusive

	parea->bExclusive++;

#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive : [%09Fp] bBaseLocked=%d bExclusive=%d NOW EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Load base header and roll back if failed

	if (!PktLoadPbh(parea)) {
	  PktLockBase(parea, FALSE);
	  parea->bExclusive--;
#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT LOAD PBH", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d HAS EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine relinquishes exclusive access to the message base
//

  FMAAPI_(BOOL) PktLeaveExclusive(PPAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);

#ifdef FMA_LOG_PKT_LOCK
Log("PktLeaveExclusive>: [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Check if not exclusive and fail if so

	if (!parea->bExclusive) {
	  FMAERROR(NOT_EXCLUSIVE);
#ifdef FMA_LOG_PKT_LOCK
Log("PktLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d NOT EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Check if exclusive more than once and leave decrementing exclusive count

	if (parea->bExclusive > 1) {
	  parea->bExclusive--;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLeaveExclusive<: [%09Fp] bExclusive=%d STILL EXCLUSIVE", parea, parea->bExclusive);
#endif
	  return TRUE;
	}

	ASSERT(parea->bExclusive == 1);

	// Save base header and check if ok

	if (!PktSavePbh(parea)) {
#ifdef FMA_LOG_PKT_LOCK
Log("PktLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT SAVE PBH", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Unlock message base and check if ok

	if (!PktLockBase(parea, FALSE)) {
#ifdef FMA_LOG_PKT_LOCK
Log("PktEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT UNLOCK", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Decrement exclusive count

	parea->bExclusive--;

#ifdef FMA_LOG_PKT_LOCK
Log("PktLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d NOW NOT EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine locks Pkt area
//

  FMAAPI_(BOOL) PktLockArea(HAREA harea, BOOL bLock)
  {
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_PKT_LOCK
Log("\nPktLockArea(%d)   >: [%09Fp] bAreaLocked=%d", bLock, parea, parea->bAreaLocked);
#endif

	// Check if locking or unlocking

	if (bLock) {

	  // Check if area is already locked and leave incremeting lock count

	  if (parea->bAreaLocked) {
		parea->bAreaLocked++;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d ALREADY LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return TRUE;
	  }

	  // Lock message base and check if ok

	  if (!PktLockBase(parea, TRUE)) {
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT LOCK", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  ASSERT(!parea->bAreaLocked);

	  // Increment area lock count

	  parea->bAreaLocked++;

	} else {

	  // Check if area is not locked and leave

	  if (!parea->bAreaLocked) {
		FMAERROR(NOT_LOCKED);
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d NOT LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Check if area is locked more than once and leave decrementing lock count

	  if (parea->bAreaLocked > 1) {
		parea->bAreaLocked--;
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d STILL LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return TRUE;
	  }

	  ASSERT(parea->bAreaLocked == 1);

	  // Unlock message base and check if ok

	  if (!PktLockBase(parea, FALSE)) {
#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT UNLOCK", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Decrement area lock count

	  parea->bAreaLocked--;
	}

#ifdef FMA_LOG_PKT_LOCK
Log("PktLockArea(%d)   <: [%09Fp] bAreaLocked=%d", bLock, parea, parea->bAreaLocked);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktLock.c
/////////////////////////////////////////////////////////////////////////////
