/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt message routines module
//
// PktMsg.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 08/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates message object
//

  static PPMSG FMASUB DoCreateMsg(PPAREA parea, MSGN msgn, UINT mode, ULONG fl)
  {
	PPMSG pmsg;

	ASSERT(parea != NULL);

	// Create message object and check if ok

	if ((pmsg = MemAlloc(sizeof(*pmsg), MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize message object

	pmsg->magic = PMSG_MAGIC;
	pmsg->parea = parea;
	pmsg->msgn = msgn;
	pmsg->mode = mode;
	pmsg->fl = fl;

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys message object
//

  static BOOL FMASUB DoDestroyMsg(PPMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Free associated message control info

	if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);

	// Reset message object magic number

	pmsg->magic = ~PMSG_MAGIC;

	// Free message object

	MemFree(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine translates special message numbers
//

  static BOOL FMASUB DoXlateMsgn(PPAREA parea, MSGN * pmsgn)
  {
	MSGN msgn = *pmsgn;

	ASSERT(parea != NULL);

	// Select current message

	if (msgn == MSGN_CUR || msgn == MSGN_CURORPREV || msgn == MSGN_CURORNEXT) {

		if (parea->msgnCur < parea->msgnBase || 
			parea->msgnCur > parea->msgnHigh) {
		if (msgn == MSGN_CURORPREV) goto Prev;
		if (msgn == MSGN_CURORNEXT) goto Next;
NotF:   FMAERROR(NOT_FOUND);
		return FALSE;
	  } else {
		msgn = parea->msgnCur;
	  }
	} else

	// Select previous message

	if (msgn == MSGN_PREV) {

Prev: if (parea->msgnCur > parea->msgnHigh)
		msgn = parea->msgnHigh;
	  if (parea->msgnCur > parea->msgnBase) {
		msgn = parea->msgnCur - 1;
	  } else
		goto NotF;
	} else

	// Select next message

	if (msgn == MSGN_NEXT) {

Next: if (parea->msgnCur < parea->msgnBase) {
		msgn = parea->msgnBase;
	  } else 
	  if (parea->msgnCur < parea->msgnHigh) {
		msgn = parea->msgnCur + 1;
	  } else 
		goto NotF;
	}

	// Set message number for caller

	if (pmsgn) *pmsgn = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the new message at the end of the base
//

  static BOOL FMASUB DoFixNewMsg(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;

	ASSERT(parea->bExclusive);

	// Append an entry to the message map and check if ok. Note that we're not
	// shrinking message map allocation since there might be more messages to add.

	if (!PktAddMsg(parea, 0, 0, 0))
	  return FALSE;

	// That's it... there is no actual message yet, it will be allocated 
	// later during the call to PktWriteMsg()

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the old message
//

  static BOOL FMASUB DoFixOldMsg(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;
	UINT imsg;

	ASSERT(parea->bExclusive);

	// Check the message number and if invalid, reset current message cache 
	// if caller wants null message and leave

	if (pmsg->msgn < parea->msgnBase || 
		pmsg->msgn > parea->msgnHigh) {
	  if (!pmsg->msgn) PktResetCurMsgCache(parea);
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Set message info into the message structure

	imsg = pmsg->msgn - parea->msgnBase;
	pmsg->lmh = parea->apmi[imsg].lmh;
	pmsg->cbHdr = parea->apmi[imsg].cbHdr;
	pmsg->cbMsg = parea->apmi[imsg].cbMsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens new message
//

  static BOOL FMASUB DoOpenNewMsg(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;
	BOOL bNew, bOk = TRUE;

	ASSERT(parea->bExclusive);

	// Set up the new message flag

	bNew = !pmsg->msgn || (pmsg->msgn > parea->msgnHigh);

	// Make up the new message number

	if (bNew) pmsg->msgn = parea->msgnHigh + 1;

	// Fix up index data to indicate that 
	// message creation is in process

	if (bOk) bOk = bNew ? DoFixNewMsg(pmsg) : DoFixOldMsg(pmsg);

	// Increment number of messages and highest message number

	if (bOk && bNew) parea->msgnHigh++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens existing message
//

  static BOOL FMASUB DoOpenExistingMsg(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;
	UINT imsg;

	ASSERT(pmsg->mode != OPENMSG_CREATE);

	// Check the message number and if invalid, reset current message cache 
	// if caller wants null message and leave

	if (pmsg->msgn < parea->msgnBase || 
		pmsg->msgn > parea->msgnHigh) {
	  if (!pmsg->msgn) PktResetCurMsgCache(parea);
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Set message info into the message structure

	imsg = pmsg->msgn - parea->msgnBase;
	pmsg->lmh = parea->apmi[imsg].lmh;
	pmsg->cbHdr = parea->apmi[imsg].cbHdr;
	pmsg->cbMsg = parea->apmi[imsg].cbMsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes new message which has not been written
//

  static BOOL FMASUB DoFixStrayNewMsg(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;
	BOOL bOk = TRUE;

	ASSERT(pmsg->mode == OPENMSG_CREATE);

	// heck if write frame has been allocated and remove it

	if ((pmsg->fs & PMF_GOTFRAME) && !(pmsg->fs & PMF_MSGHWRITTEN)) {
	  if (pmsg->lmh && pmsg->cbMsg) {
		PktShrinkPkt(parea, pmsg->lmh, pmsg->cbMsg);
		pmsg->lmh = 0; pmsg->cbMsg = 0;
	  }
	}

	// Adjust area message count. Note that this also voids 
	// associated messge map entry

	parea->cmsg--;

	// Adjust highest message number

	parea->msgnHigh--;

	// Adjust current message number

	if (parea->msgnCur == pmsg->msgn) {
	  parea->msgnCur = 0;
	} else
	if (parea->msgnCur > pmsg->msgn) {
	  parea->msgnCur--;
	}

#ifdef FMA_LOG_PKT_READ
Log("DoFixStrayNewMsg  : [%09Fp] msgn=%ld msgnHigh=%ld", pmsg, pmsg->msgn, parea->msgnHigh);
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes message end
//

  static BOOL FMASUB DoFixMsgEnd(PPMSG pmsg)
  {
	PPAREA parea = pmsg->parea;
	LONG lmtEnd, lMsgEnd;
	PPMI ppmi;
	ULONG cb;

	ASSERT(pmsg->mode != OPENMSG_READ);

	// Check if message has no frame and leave

	if (!pmsg->lmh || !pmsg->lmt) return TRUE;

	// Determine last written offset

	lmtEnd = pmsg->lmtEnd ? pmsg->lmtEnd : pmsg->lmt;

	ASSERT(lmtEnd >= pmsg->lmh + (LONG)pmsg->cbHdr);

	// Check if last written offset fills the allocated 
	// message frame and leave if so, Note one off to 
	// protect text terminator 

	lMsgEnd = pmsg->lmh + (LONG)pmsg->cbMsg - 1;
	if (lmtEnd >= lMsgEnd) return TRUE;

	// Shrink message frame

	cb = lMsgEnd - lmtEnd;
	PktShrinkPkt(parea, lmtEnd, cb);

	// Fix up the message map entry

	ppmi = &parea->apmi[pmsg->msgn - parea->msgnBase];
	ppmi->cbMsg-= cb;
	if (ppmi->cbHdr >= ppmi->cbMsg) ppmi->cbHdr-= cb;

#ifdef FMA_LOG_PKT_READ
Log("DoFixMsgEnd       : [%09Fp] msgn=%ld lmtEnd=%08lx lMsgEnd=%08lx", pmsg, pmsg->msgn, lmtEnd, lMsgEnd);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  static BOOL FMASUB DoGetOpenMsgIndex(PPMSG pmsg, UINT * pimsg)
  {
	PPAREA parea = pmsg->parea;
	PPMSG pmsgThis;
	UINT imsg;

	// Scan through the open messages list looking for a match

	for (pmsgThis = parea->pmsgList, imsg = 0; pmsgThis != NULL; 
		 pmsgThis = pmsgThis->pmsgNext, imsg++) {
	  if (pmsgThis == pmsg) {
		if (pimsg) *pimsg = imsg;
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens message
//

  FMAAPI_(HMSG) PktOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl)
  {
	PPMSG pmsg = NULL;
	PPAREA parea;

	// Validate parameters

	if ((parea = (PPAREA)harea) == NULL || !PktCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_PKT_MSG
Log("PktOpenMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  // Clean up and leave

	  if (pmsg) {
		if (pmsg->fs & PMF_EXCLUSIVE) PktLeaveExclusive(parea);
		DoDestroyMsg(pmsg);
	  }

	  return NULL;
	}

#ifdef FMA_LOG_PKT_MSG
Log("\nPktOpenMsg       >: [%09Fp] msgn=%ld in %s, mode=%d", pmsg, msgn, parea->achPath, mode);
#endif

	// Translate special message numbers and check if ok

	if (!DoXlateMsgn(parea, &msgn))
	  goto Fail;

	// Create message object and check if ok

	if ((pmsg = DoCreateMsg(parea, msgn, mode, fl)) == NULL)
	  goto Fail;

	// Check if we need an exclusive access and acquire it

	if (mode == OPENMSG_CREATE || 
		mode == OPENMSG_WRITE ||
		mode == OPENMSG_READWRITE) {
	  if (PktEnterExclusive(parea)) 
		SETFLAG(pmsg->fs, PMF_EXCLUSIVE);
	  else 
		goto Fail;
	}

	// Create a new message, or open an existing message, as specified

	if (mode == OPENMSG_CREATE) {
	  if (!DoOpenNewMsg(pmsg)) 
		goto Fail;
	} else {
	  if (!DoOpenExistingMsg(pmsg)) 
		goto Fail;
	}

	// Set current message cache to this message

	PktSetCurMsgCache(pmsg);

	// Add message to the top of the list of open messages for this area

	ListAdd(&parea->pmsgList, pmsg, 0);    

	ASSERT(PktCheckMsg(pmsg));

#ifdef FMA_LOG_PKT_MSG
Log("PktOpenMsg       <: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes message
//

  FMAAPI_(BOOL) PktCloseMsg(HMSG hmsg)
  {
	BOOL bOk = TRUE;
	PPAREA parea;
	PPMSG pmsg;
	UINT imsg;

	// Validate parameters

	if ((pmsg = (PPMSG)hmsg) == NULL || !PktCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_PKT_MSG
Log("PktCloseMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_PKT_MSG
Log("\nPktCloseMsg      >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	// Check if this is a new message which has not been written to and
	// update base to forget about it

	if (pmsg->mode == OPENMSG_CREATE && !(pmsg->fs & PMF_MASKWRITTEN)) {
	  if (!DoFixStrayNewMsg(pmsg))
		bOk = FALSE;
	}

	// Check if message has not been written up to cbTotalText and
	// shrink packet file to maintain its integrity

	if (pmsg->mode != OPENMSG_READ) {
	  if (!DoFixMsgEnd(pmsg))
		bOk = FALSE;
	}

	// Check if we entered an exclusive access and relinquish it

	if (pmsg->fs & PMF_EXCLUSIVE) {
	  if (!PktLeaveExclusive(parea))
		bOk = FALSE;
	}

	// Delete message from the list of open messages for this area

	if (DoGetOpenMsgIndex(pmsg, &imsg)) {
	  ListDel(&parea->pmsgList, imsg);
	} else {
	  ASSERT(DoGetOpenMsgIndex(pmsg, &imsg));
	}

	// Destroy message object

	DoDestroyMsg(pmsg);

	// Check if failed and jump to failure return

	if (!bOk) goto Fail;

#ifdef FMA_LOG_PKT_MSG
Log("PktCloseMsg      <: [%09Fp] in %s", pmsg, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message object
//

  FMAAPI_(BOOL) PktCheckMsg(PPMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check the message object magic number

	if (pmsg->magic != PMSG_MAGIC)
	  return FALSE;

	// Check the area object in the debugging build

	ASSERT(pmsg->parea != NULL);
	ASSERT(PktCheckArea(pmsg->parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be read
//

  FMAAPI_(BOOL) PktCheckMsgReadMode(PPMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be read

	if (pmsg->mode != OPENMSG_READ && pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_READ);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be written
//

  FMAAPI_(BOOL) PktCheckMsgWriteMode(PPMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be written

	if (pmsg->mode != OPENMSG_CREATE && pmsg->mode != OPENMSG_WRITE &&
		pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_WRITE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktMsg.c
/////////////////////////////////////////////////////////////////////////////
