/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish area routines module
//
// SquArea.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 22/Feb/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          22/Feb/2000 kvitekp - Fixed SquUidToMsgn range check
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Squish message area files structure

  typedef struct _SFILES {      /* files */
	CHAR    achSqdFile[MAX_PATH];       // Squish data file name
	CHAR    achSqiFile[MAX_PATH];       // Squish index file name
	CHAR    achSqlFile[MAX_PATH];       // Squish lastread file name
  } SFILES, FAR * PSFILES;

  // Forward declarations

  static VOID FMASUB DoCloseAreaFiles(PSAREA parea);

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates area object
//

  static PSAREA FMASUB DoCreateArea(PSZ pszPath, ULONG fl, NETADDR * pnetAddr)
  {
	PSAREA parea;
	PCH pchExt;
	int cch;

	// Get area path length and check if ok

	if (!pszPath || (cch = strlen(pszPath)) == 0) {
BadP: FMAERROR(INVALID_PARAM);
	  return NULL;
	}

	// Check if there is an ".sq?" extension specified and remove it

	if (cch > lengof(SQU_SQD_FILEEXT) &&
	   (pchExt = &pszPath[cch - lengof(SQU_SQD_FILEEXT)])[0] == '.' &&
		TOLOWER(pchExt[1]) == TOLOWER(SQU_SQD_FILEEXT[1]) &&
		TOLOWER(pchExt[2]) == TOLOWER(SQU_SQD_FILEEXT[2])) {
	  cch-= lengof(SQU_SQD_FILEEXT);
	}

	// Get area path length and check if ok

	if (cch >= MAX_PATH - lengof(SQU_SQD_FILEEXT))
	  goto BadP;

	// Create message area object and check if ok

	if ((parea = MemAlloc(sizeof(SAREA) + cch, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize new area object

	parea->magic = SAREA_MAGIC;
	parea->papi = SquGetApi();
	parea->fl = fl;
	if (pnetAddr) parea->netAddr = *pnetAddr;
	memcpy(parea->achPath, pszPath, cch);
	parea->hfileSqd = -1;
	parea->hfileSqi = -1;
	parea->hfileSql = -1;

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys area object
//

  static BOOL FMASUB DoDestroyArea(PSAREA parea)
  {
	// Check if we have area object and leave if not

	if (!parea) return FALSE;

	// Close associated files

	DoCloseAreaFiles(parea);

	// Delete associated index list

	ASSERT(parea->psicList == NULL);

	// Reset message area object magic number

	parea->magic = ~SAREA_MAGIC;

	// Free message area object

	MemFree(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine makes up the area file names
//

  static BOOL FMASUB DoMakeAreaFileNames(PSAREA parea, PSFILES pfiles)
  {
	ASSERT(parea != NULL);

	strcpy(pfiles->achSqdFile, parea->achPath); strcat(pfiles->achSqdFile, SQU_SQD_FILEEXT);
	strcpy(pfiles->achSqiFile, parea->achPath); strcat(pfiles->achSqiFile, SQU_SQI_FILEEXT);
	strcpy(pfiles->achSqlFile, parea->achPath); strcat(pfiles->achSqlFile, SQU_SQL_FILEEXT);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens the area files
//

  static BOOL FMASUB DoOpenAreaFiles(PSAREA parea, PSFILES pfiles, ULONG fl, UINT fOpen)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	// Make up the file open mode flags

	if (fl & FOA_READONLY)
	  fOpen|= FOPENFILE_READONLY;
	else
	  fOpen|= FOPENFILE_READWRITE;

	// Make up the file sharing flags

	if (fl & FOA_DENYALL)
	  fOpen|= FOPENFILE_DENYALL;
	else
	  fOpen|= FOPENFILE_DENYNONE;

	// Open the data file and check if ok

	if (!FmaOpenFile(pfiles->achSqdFile, fOpen, &parea->hfileSqd)) {
Fail: DoCloseAreaFiles(parea);
	  return FALSE;
	}

	// Open the index file and check if ok

	if (!FmaOpenFile(pfiles->achSqiFile, fOpen, &parea->hfileSqi)) {
	  goto Fail;
	}

	// Open the lastread file and check if ok. Note that this file
	// is always created if it does not exist unless in readonly mode

	if (!(fl & FOA_READONLY)) fOpen|= FOPENFILE_CREATE;
	if (!FmaOpenFile(pfiles->achSqlFile, fOpen, &parea->hfileSql)) {
	  goto Fail;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes the area files
//

  static VOID FMASUB DoCloseAreaFiles(PSAREA parea)
  {
	ASSERT(parea != NULL);

	// Close all open message base files

	FmaCloseFile(&parea->hfileSqd);
	FmaCloseFile(&parea->hfileSqi);
	FmaCloseFile(&parea->hfileSql);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes area files
//

  static BOOL FMASUB DoDeleteAreaFiles(PSAREA parea, PSFILES pfiles)
  {
	BOOL bOk = TRUE;

	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	// Make sure the files area closed

	DoCloseAreaFiles(parea);

	// Delete area files and check if ok

	if (!FmaDeleteFile(pfiles->achSqdFile)) bOk = FALSE;
	if (!FmaDeleteFile(pfiles->achSqiFile)) bOk = FALSE;
	if (!FmaDeleteFile(pfiles->achSqlFile)) bOk = FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine initializes squish base header
//

  static VOID FMASUB DoInitSbh(PSBH psbh)
  {
	// Initialize squish base header

	memset(psbh, 0, sizeof(SBH));
	psbh->cb = sizeof(SBH);
	psbh->umsgNext = 1;
	psbh->lfmFrameEnd = sizeof(SBH);
	psbh->cbSfh = sizeof(SFH);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area directory
//

  static BOOL FMASUB DoMakeAreaDir(PSAREA parea)
  {
	CHAR achPath[MAX_PATH];
	PCH pch;

	ASSERT(parea != NULL);

	// Copy path to the local storage and chop the base name

	strcpy(achPath, parea->achPath);
	if ((pch = strrchr(achPath, PATH_SEP)) != NULL) {
	  *pch = '\0';
	} else
	  return TRUE;

	// Check if at root or directory exists and leave if so

	if (!achPath[0] || FmaDirExists(achPath))
	  return TRUE;

	// Create directory and check if ok

	if (!FmaMakeDir(achPath)) {
	  FMAERROR(CANNOT_CREATE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area
//

  static BOOL FMASUB DoCreateNewArea(PSAREA parea, ULONG fl, PSFILES pfiles)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	ASSERT(parea->hfileSqd == -1);
	ASSERT(parea->hfileSqi == -1);

	// Create area directory if necessary

	DoMakeAreaDir(parea);

	// Create area files and check if ok

	if (!DoOpenAreaFiles(parea, pfiles, fl, FOPENFILE_CREATE | FOPENFILE_TRUNCATE))
	  return FALSE;

	// Initialize new area base header

	DoInitSbh(&parea->sbh);

	// Save empty base header and check if ok

	if (!SquSaveSbh(parea)) {
	  DoDeleteAreaFiles(parea, pfiles);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine clears existing area
//

  static BOOL FMASUB DoClearExistingArea(PSAREA parea, ULONG fl, PSFILES pfiles)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	ASSERT(parea->hfileSqd == -1);
	ASSERT(parea->hfileSqi == -1);
	ASSERT(parea->hfileSql == -1);

	// Open and truncate area files and check if ok

	if (!DoOpenAreaFiles(parea, pfiles, fl, FOPENFILE_TRUNCATE))
	  return FALSE;

	// Initialize empty area base header

	DoInitSbh(&parea->sbh);

	// Save empty base header and check if ok

	if (!SquSaveSbh(parea)) {
	  DoCloseAreaFiles(parea);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens area
//

  FMAAPI_(HAREA) SquOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr)
  {
	BOOL bAreaExists = FALSE;
	SFILES files;
	PSAREA parea;

	// Assert known structure sizes to ensure correct structure packing

	ASSERT(sizeof(SBH) == 256);
	ASSERT(sizeof(SFH) == 28);
	ASSERT(sizeof(SMH) == 238);
	ASSERT(sizeof(SQI) == 12);
	ASSERT(sizeof(SLR) == 4);

#ifdef FMA_LOG_SQU_AREA
Log("\nSquOpenArea      >: %s", pszPath);
#endif

	// Create new area object and check if ok

	if ((parea = DoCreateArea(pszPath, fl, pnetAddr)) == NULL) {
Fail: DoDestroyArea(parea);

#ifdef FMA_LOG_SQU_AREA
Log("SquOpenArea      <: %s FAILED, error=%d", pszPath, FmaGetLastError());
#endif

	  return NULL;
	}

	// Make up the base file names

	DoMakeAreaFileNames(parea, &files);

	// Check if all area message base files exist

	if (FmaFileExists(files.achSqdFile)) bAreaExists++;
	if (FmaFileExists(files.achSqiFile)) bAreaExists++;
	if (FmaFileExists(files.achSqlFile)) bAreaExists++;

	// Check if some files are missing

	if (bAreaExists && bAreaExists != 3) {

	  // Check if it's a lastread file that is missing and
	  // don't fail if so, otherwise we're missing some 
	  // critical message base files

	  if (!(bAreaExists == 2 && !FmaFileExists(files.achSqlFile))) {
		FMAERROR(MISSING_FILES);
		goto Fail;
	  }
	}

	// Act upon an opening mode

	switch (fOpenMode) {

	  case OPENAREA_EXISTING:

		// Fail if does not exist, otherwise open

		if (!bAreaExists) {
		  FMAERROR(NOT_FOUND);
		  goto Fail;
		} else 
Open:   if (!DoOpenAreaFiles(parea, &files, fl, 0)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_ALWAYS:

		// Create if does not exist, otherwise open

		if (!bAreaExists) {
		  if (!DoCreateNewArea(parea, fl, &files))
			goto Fail;
		} else
		  goto Open;

		break;

	  case OPENAREA_CREATE_NEW:

		// Fail if exists, otherwise create new

		if (bAreaExists) {
		  FMAERROR(ALREADY_EXISTS);
		  goto Fail;
		} else
		if (!DoCreateNewArea(parea, fl, &files)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_CREATE_ALWAYS:

		// Clear if exists, otherwise create new

		if (bAreaExists) {
		  if (!DoClearExistingArea(parea, fl, &files)) {
			goto Fail;
		  }
		} else 
		if (!DoCreateNewArea(parea, fl, &files)) {
		  goto Fail;
		}

		break;
	}

	ASSERT(parea->hfileSqd != -1);
	ASSERT(parea->hfileSqi != -1);
	ASSERT(parea->hfileSql != -1);

	// Load base header and if failed, remove area files if
	// they were just created for the new area

	if (!SquLoadSbh(parea)) {
	  if (!bAreaExists) DoDeleteAreaFiles(parea, &files);
	  goto Fail;
	}

	// Set the cached frame offsets so that the check against
	// next to current message (which is zero) will do fine

	SquResetCurMsgCache(parea);

	// Increment open area count in a thread safe manner

	FmaInterlockedIncrement(&fmaRtm.nOpenAreas);

	ASSERT(SquCheckArea(parea));

	// Check if caller wants to lock the area upon opening

	if (parea->fl & FOA_LOCKAREA) {
	  if (!SquLockArea(parea, TRUE)) {
		SquCloseArea(parea);
		parea = NULL;
		goto Fail;
	  }
	}

#ifdef FMA_LOG_SQU_AREA
Log("SquOpenArea      <: [%09Fp] cmsg=%ld msgnHigh=%ld umsgNext=%ld in %s", parea, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext, parea->achPath);
#endif

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes area
//

  FMAAPI_(BOOL) SquCloseArea(HAREA harea)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_SQU_AREA
Log("\nSquCloseArea     >: [%09Fp] cmsg=%ld msgnHigh=%ld umsgNext=%ld in %s", parea, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext, parea->achPath);
#endif

	// Check if message base trimming has been delayed and do it now

	if (parea->fl & FOA_DELAYTRIMMING) {

	  // Acquire exclusive access, trim database and
	  // relinquish exclusive access

	  if (SquEnterExclusive(parea)) {
		SquTrimBase(parea, FALSE);
		SquLeaveExclusive(parea);
	  }
	}

	// Check if caller locked the area upon opening and
	// area is still locked and if so, unlock it

	if (parea->fl & FOA_LOCKAREA && parea->bAreaLocked) {
	  SquLockArea(parea, FALSE);
	}

	// Check if there are any open messages

	if (parea->pmsgList) {
	  FMAERROR(STILL_OPEN);
#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea      : [%09Fp] %d msg(s) still open", parea, ListCount(parea->pmsgList)); 
#endif

	  // Close all open messages

	  while (parea->pmsgList) {
		SquCloseMsg(parea->pmsgList);
	  }
	}

	// Check if we have exclusive access

	while (parea->bExclusive) {
#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea      : [%09Fp] bExclusive=%d", parea, parea->bExclusive); 
#endif
	  SquLeaveExclusive(parea);
	}

	// Check if we have area locked

	while (parea->bAreaLocked) {
#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea      : [%09Fp] bAreaLocked=%d", parea, parea->bAreaLocked); 
#endif
	  SquLockArea(parea, FALSE);
	}

	// Check if we have index caching

	while (parea->bSqiCached) {
#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea      : [%09Fp] bSqiCached=%d", parea, parea->bSqiCached); 
#endif
	  SquEndSqiCache(parea);
	}

	// Check if we have base locked

	while (parea->bBaseLocked) {
#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea      : [%09Fp] bBaseLocked=%d", parea, parea->bBaseLocked); 
#endif
	  SquLockBase(parea, FALSE);
	}

	// Destroy message area object

	DoDestroyArea(parea);

	// Decrement open area count in a thread safe manner

	FmaInterlockedDecrement(&fmaRtm.nOpenAreas);

#ifdef FMA_LOG_SQU_AREA
Log("SquCloseArea     <: [%09Fp] CLOSED", parea); 
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message
//

  FMAAPI_(BOOL) SquValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg)
  {
	PSAREA parea;
	SQI sqi;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!msgn || msgn > parea->sbh.msgnHigh) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the index entry for this message number and check if ok

	if (!SquGetSqi(parea, &sqi, msgn))
	  return FALSE;

	// Check if caller needs message index and calculate it

	if (pimsg) *pimsg = msgn - 1;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message map which is array of unique message ids
//

  FMAAPI_(BOOL) SquCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg)
  {
	ULONG cb, cumsg, iumsg = 0;
	PSAREA parea;
	UMSG * aumsg;
	MSGN msgn;
	SQI sqi;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(paumsg != NULL);
	ASSERT(pcumsg != NULL);

	// Check if no messages in this area and calculate message map size

	if ((cumsg = parea->sbh.cmsg) == 0) {
	  *paumsg = NULL; *pcumsg = 0;
	  return TRUE;
	} else
	  cb = cumsg * sizeof(UMSG);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Allocate memory to hold message map and check if ok

	if ((aumsg = MemAlloc((int)cb, 0)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	// Loop through all the messages retrieving umsg ids

	for (msgn = 1; msgn <= parea->sbh.msgnHigh; msgn++) {
	  if (SquGetSqi(parea, &sqi, msgn)) {
		aumsg[iumsg++] = sqi.umsg;
	  }
	}

	ASSERT(iumsg == cumsg);

	// Set caller's pointers

	*paumsg = aumsg; *pcumsg = iumsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates messages
//

  FMAAPI_(BOOL) SquEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl)
  {
	ENUMMSGINFO emi;
	BOOL bOk = TRUE;
	ULONG cumsg, iumsg;
	UMSG * aumsg;
	PSAREA parea;
	MSGN msgn;
	SQI sqi;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea) || !pfnCallback) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if (!parea->sbh.cmsg) return FALSE;

	// Prepare enum message info structure

	memset(&emi, 0, sizeof(emi));
	emi.harea = harea;

	// Check if safe enum which will work if messages area deleted

	if (fl & FENUMMSGS_SAFEENUM) {

	  // Create message map and check if ok

	  if (!SquCreateMsgMap(parea, &aumsg, &cumsg) || !cumsg)
		return FALSE;

	  // Check enumeration direction

	  if (fl & FENUMMSGS_REVERSE) {

		// Enumerate in reversed order

		for (iumsg = cumsg - 1; (LONG) iumsg >= 0; iumsg--) {
		  msgn = SquUidToMsgn(parea, aumsg[iumsg], UMSG_EXACT);
		  if (msgn && SquGetSqi(parea, &sqi, msgn)) {
			emi.msgn = msgn; emi.umsg = sqi.umsg; emi.imsg = msgn - 1;
			emi.offs = sqi.lfm; emi.hash = sqi.hash;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  } else {

		// Enumerate in normal order

		for (iumsg = 0; iumsg < cumsg; iumsg++) {
		  msgn = SquUidToMsgn(parea, aumsg[iumsg], UMSG_EXACT);
		  if (msgn && SquGetSqi(parea, &sqi, msgn)) {
			emi.msgn = msgn; emi.umsg = sqi.umsg; emi.imsg = msgn - 1;
			emi.offs = sqi.lfm; emi.hash = sqi.hash;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  }

	  // Free message map

	  MemFree(aumsg);

	} else {

	  // Check enumeration direction

	  if (fl & FENUMMSGS_REVERSE) {

		// Enumerate in reversed order

		for (msgn = parea->sbh.msgnHigh; msgn > 0; msgn--) {
		  if (SquGetSqi(parea, &sqi, msgn)) {
			emi.msgn = msgn; emi.umsg = sqi.umsg; emi.imsg = msgn - 1;
			emi.offs = sqi.lfm; emi.hash = sqi.hash;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  } else {

		// Enumerate in normal order

		for (msgn = 1; msgn <= parea->sbh.msgnHigh; msgn++) {
		  if (SquGetSqi(parea, &sqi, msgn)) {
			emi.msgn = msgn; emi.umsg = sqi.umsg; emi.imsg = msgn - 1;
			emi.offs = sqi.lfm; emi.hash = sqi.hash;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  }
	}

	// Return whatever last callback returned

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message
//

  FMAAPI_(BOOL) SquDeleteMsg(HAREA harea, MSGN msgn)
  {
	PSAREA parea;
	BOOL bOk;
	LONG lfm;
	SFH sfh;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_SQU_AREA
Log("\nSquDeleteMsg     >: [%09Fp] msgn=%ld cmsg=%ld msgnHigh=%ld umsgNext=%ld", parea, msgn, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext);
#endif

	// Check message number

	if (!msgn || msgn > parea->sbh.msgnHigh) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Check if this message is open and leave if so

	if (SquGetOpenMsg(parea, msgn)) {
	  FMAERROR(STILL_OPEN);
	  return FALSE;
	}

	// Acquire exclusive access and check if ok

	if (!SquEnterExclusive(parea))
	  return FALSE;

	// Load frame header and delete the message 

	if (SquGetFrame(parea, &lfm, msgn) &&
		SquLoadSfh(parea, &sfh, lfm)) {
	  bOk = SquDelMsg(parea, msgn, &sfh, lfm);
	} else
	  bOk = FALSE;

	// Relinquish exclusive access and check if ok

	if (!SquLeaveExclusive(parea))
	  bOk = FALSE;

#ifdef FMA_LOG_SQU_AREA
Log("SquDeleteMsg     <: [%09Fp] msgn=%ld cmsg=%ld msgnHigh=%ld umsgNext=%ld", parea, msgn, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext);
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets current message
//

  FMAAPI_(BOOL) SquSetCurMsg(HAREA harea, MSGN msgn)
  {
	PSAREA parea;
	LONG lfm;
	SFH sfh;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check message number

	if (!msgn || msgn > parea->sbh.msgnHigh) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the frame offset and check if ok

	if (!SquGetFrame(parea, &lfm, msgn))
	  return FALSE;

	// Load the frame header and check if ok

	if (!SquLoadSfh(parea, &sfh, lfm))
	  return FALSE;

	// Set current message cache

	parea->msgnCur = msgn;
	parea->lfmCur  = lfm;
	parea->lfmPrev = sfh.lfmPrev;
	parea->lfmNext = sfh.lfmNext;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets current message
//

  FMAAPI_(MSGN) SquGetCurMsg(HAREA harea)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(parea->msgnCur <= parea->sbh.msgnHigh);

	// Return current message

	return parea->msgnCur;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets first message number
//

  FMAAPI_(MSGN) SquGetFirstMsg(HAREA harea)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return first message number

	return parea->sbh.cmsg ? 1 : 0;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets highest message number
//

  FMAAPI_(MSGN) SquGetHighMsg(HAREA harea)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return highest message number

	return parea->sbh.msgnHigh;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message count
//

  FMAAPI_(LONG) SquGetMsgCount(HAREA harea)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return message count

	return parea->sbh.cmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts message number to unique message id
//

  FMAAPI_(UMSG) SquMsgnToUid(HAREA harea, MSGN msgn)
  {
	PSAREA parea;
	SQI sqi;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!msgn || msgn > parea->sbh.msgnHigh) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the index entry for this message number and check if ok

	if (!SquGetSqi(parea, &sqi, msgn))
	  return FALSE;

	// Return unique message id

	return sqi.umsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts unique message id to message number
//

  FMAAPI_(MSGN) SquUidToMsgn(HAREA harea, UMSG umsg, UINT fMode)
  {
	LONG msgn = 0, msgnLo, msgnHi;      // these have to be signed for bsearch
	SQI sqi = {0, 0, 0};
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message unique id

	if (!umsg) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Run binary search looking for umsgid match

	msgnLo = 1; msgnHi = parea->sbh.msgnHigh;

	while (msgnLo <= msgnHi) {

	  // Split the current range in the middle

	  msgn = (msgnLo + msgnHi) / 2;

	  // Get the index entry for this message number

	  if (!SquGetSqi(parea, &sqi, msgn))
		break;

	  // Check if we got a match and break

	  if (umsg == sqi.umsg)
		break;

	  // Select a new range

	  if (umsg > sqi.umsg)
		msgnLo = msgn + 1;
	  else
		msgnHi = msgn - 1;
	}

	// Check if we got an exact match and return message number
	// if caller wants an exact one

	if (sqi.umsg == umsg) {
	  if (fMode & UMSG_EXACT || !fMode)
		return msgn;
	}

	// Check if the caller wants next or previous message

	if (fMode & UMSG_NEXT) {
	  if (sqi.umsg > umsg)
		return msgn;

	  while (++msgn <= (LONG)parea->sbh.msgnHigh) {
		if (SquGetSqi(parea, &sqi, msgn))
		  return msgn;
	  }
	} else
	if (fMode & UMSG_PREV) {
	  if (sqi.umsg < umsg)
		return msgn;

	  while (--msgn > 0) {
		if (SquGetSqi(parea, &sqi, msgn))
		  return msgn;
	  }
	} 

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message area object
//

  FMAAPI_(BOOL) SquCheckArea(PSAREA parea)
  {
	ASSERT(parea != NULL);

	// Check the message area object magic number

	if (parea->magic != SAREA_MAGIC)
	  return FALSE;

	ASSERT(SquCheckSbh(parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets message base trimming values
//

  FMAAPI_(BOOL) SquSetBaseTrimming(PSAREA parea, ULONG cmsgMax, ULONG cmsgSkip, 
								   USHORT keepDays)
  {
	BOOL bUpdate = FALSE;

	ASSERT(parea != NULL);

	ASSERT(SquCheckSbh(parea));

	// Check if message base limits area changing

	if (cmsgMax  != (ULONG) -1 && parea->sbh.cmsgMax  != cmsgMax)  bUpdate = TRUE;
	if (cmsgSkip != (ULONG) -1 && parea->sbh.cmsgSkip != cmsgSkip) bUpdate = TRUE;
	if (keepDays != (USHORT)-1 && parea->sbh.keepDays != keepDays) bUpdate = TRUE;

	// Update only if message base trimming values are changing

	if (bUpdate) {

	  // Acquire exclusive access and check if ok

	  if (!SquEnterExclusive(parea))
		return FALSE;

	  // Update message base trimming values

	  if (cmsgMax  != (ULONG) -1) parea->sbh.cmsgMax = cmsgMax;
	  if (cmsgSkip != (ULONG) -1) parea->sbh.cmsgSkip = cmsgSkip;
	  if (keepDays != (USHORT)-1) parea->sbh.keepDays = keepDays;

	  // Relinquish exclusive access and check if ok

	  if (!SquLeaveExclusive(parea))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquArea.c
/////////////////////////////////////////////////////////////////////////////
