/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish helper routines module
//
// SquHelp.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 14/Sep/1999
// Updated: 28/Sep/1999
//
// History: 14/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Following is used for stress testing only

#ifdef FMA_STRESS_TEST
  FMAAPI_(BOOL) FmaQueryDiskAllocFailure(HFILE hfile, ULONG cb);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adjusts open message numbers after message has been deleted
//

  static BOOL FMASUB DoFixOpenMsg(PSAREA parea, MSGN msgnDeleted)
  {
	PSMSG pmsg;

	ASSERT(parea  != NULL);

	// Scan through the open messages list adjusting stored message numbers 
	// for those which are above the deleted messages range

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {

	  ASSERT(pmsg->msgn != msgnDeleted);

	  if (pmsg->msgn > msgnDeleted) pmsg->msgn--;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads squish message base header
//

  FMAAPI_(BOOL) SquLoadSbh(PSAREA parea)
  {
	ULONG cbRead;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

	// Seek to the beginning of the data file and load the base header

	if (!FmaReadFile(parea->hfileSqd, 0, &parea->sbh, sizeof(SBH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(SBH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check the squish base header

	if (!SquCheckSbh(parea))
	  return FALSE;

#ifdef FMA_LOG_SQU_HELP
Log("SquLoadSbh        : [%09Fp] bBaseLocked=%d bExclusive=%d cmsg=%ld msgnHigh=%ld umsgNext=%ld", parea, parea->bBaseLocked, parea->bExclusive, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves squish message base header
//

  FMAAPI_(BOOL) SquSaveSbh(PSAREA parea)
  {
	ULONG cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);
	ASSERT(parea->bExclusive || parea->sbh.umsgNext == 1);

	// Check the squish base header

	if (!SquCheckSbh(parea)) {
	  return FALSE;
	}

	// Seek to the beginning of the data file and save the base header

	if (!FmaWriteFile(parea->hfileSqd, 0, &parea->sbh, sizeof(SBH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(SBH));

#ifdef FMA_LOG_SQU_HELP
Log("SquSaveSbh        : [%09Fp] bBaseLocked=%d bExclusive=%d cmsg=%ld msgnHigh=%ld umsgNext=%ld", parea, parea->bBaseLocked, parea->bExclusive, parea->sbh.cmsg, parea->sbh.msgnHigh, parea->sbh.umsgNext);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks squish message base header
//

  FMAAPI_(BOOL) SquCheckSbh(PSAREA parea)
  {
	SBH sbh;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != 1);

	// Pick up the squish base header on stack

	sbh = parea->sbh;

	// Check squish base header size

	if (sbh.cb < sizeof(SBH)) {
Fail: FMAERROR(INVALID_DATA);
	  return FALSE;
	}

	// Check message count and related values

	if (sbh.cmsg != sbh.msgnHigh) goto Fail;
	if (sbh.cmsg >  sbh.umsgNext + 1) goto Fail;
	if (sbh.msgnHigh > sbh.umsgNext + 1) goto Fail;

	// Check frame chain pointers

	if (sbh.lfmFrameEnd < (LONG)sbh.cb) goto Fail;
	if (sbh.lfmFirst    > sbh.lfmFrameEnd) goto Fail;
	if (sbh.lfmLast     > sbh.lfmFrameEnd) goto Fail;
	if (sbh.lfmFree     > sbh.lfmFrameEnd) goto Fail;
	if (sbh.lfmLastFree > sbh.lfmFrameEnd) goto Fail;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine resets the current message cache
//

  FMAAPI_(BOOL) SquResetCurMsgCache(PSAREA parea)
  {
	ASSERT(parea != NULL);

	// Reset current message cache

	parea->msgnCur = 0;
	parea->lfmCur  = 0;
	parea->lfmPrev = 0;
	parea->lfmNext = parea->sbh.lfmFirst;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the current message cache
//

  FMAAPI_(BOOL) SquSetCurMsgCache(PSMSG pmsg)
  {
	PSAREA parea;

	ASSERT(pmsg != NULL);

	parea = pmsg->parea;

	// Set current message cache

	parea->msgnCur = pmsg->msgn;
	parea->lfmCur  = pmsg->lfm;
	parea->lfmPrev = pmsg->sfh.lfmPrev;
	parea->lfmNext = pmsg->sfh.lfmNext;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads squish frame header at the specified offset
//

  FMAAPI_(BOOL) SquLoadSfh(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ULONG cbRead;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);
	ASSERT(psfh != NULL);

	// Check frame offset

	if (lfm < (LONG)parea->sbh.cb || lfm >= parea->sbh.lfmFrameEnd) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Seek to specified frame offset and load the frame header

	if (!FmaReadFile(parea->hfileSqd, lfm, psfh, sizeof(SFH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(SFH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check the frame header magic

	if (psfh->magic != SFH_MAGIC) {
BadF: FMAERROR(INVALID_FRAME);
	  return FALSE;
	}

	// Check if none zero frame links point into the base header

	if ((psfh->lfmPrev && psfh->lfmPrev < (LONG)parea->sbh.cb) || 
		(psfh->lfmNext && psfh->lfmNext < (LONG)parea->sbh.cb))
	  goto BadF;

	// Check if frame links point past the end of base

	if (psfh->lfmPrev > parea->sbh.lfmFrameEnd || 
		psfh->lfmNext > parea->sbh.lfmFrameEnd)
	  goto BadF;

	// Check if frame links point to itself

	if (psfh->lfmPrev == lfm || 
		psfh->lfmNext == lfm)
	  goto BadF;

	// Note: treat other inconsistencies as minor 
	// but do assert them in a debug build

	ASSERT(psfh->cbFrame >= sizeof(SMH));
	ASSERT(!psfh->cbMsg || psfh->cbMsg   >= sizeof(SMH));
	ASSERT(!psfh->cbMsg || psfh->cbMsg   <= psfh->cbFrame);
	ASSERT(!psfh->cbMsg || sizeof(SMH) + psfh->cbCtrl  <= psfh->cbMsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves squish frame header at the specified offset
//

  FMAAPI_(BOOL) SquSaveSfh(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ULONG cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	// Check frame offset

	if (lfm < (LONG)parea->sbh.cb) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(!psfh->lfmPrev || psfh->lfmPrev >= (LONG)parea->sbh.cb);
	ASSERT(!psfh->lfmNext || psfh->lfmNext >= (LONG)parea->sbh.cb);

	ASSERT(psfh->lfmPrev < parea->sbh.lfmFrameEnd);
	ASSERT(psfh->lfmNext < parea->sbh.lfmFrameEnd);

	ASSERT(psfh->lfmPrev != lfm);
	ASSERT(psfh->lfmNext != lfm);

	// Seek to specified frame offset and save the frame header

	if (!FmaWriteFile(parea->hfileSqd, lfm, psfh, sizeof(SFH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(SFH));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the frame of the specified length at the end of file
//

  FMAAPI_(BOOL) SquAddFrame(PSAREA parea, PLONG plfm, ULONG cbFrame)
  {
	char ch = 0;
	long offs;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);
	ASSERT(parea->hfileSqd != -1);

	// Following is used for stress testing only

#ifdef FMA_STRESS_TEST
	if (FmaQueryDiskAllocFailure(parea->hfileSqd, cbFrame)) return FALSE;
#endif

	// Calculate offset of the very last byte of this frame

	offs = parea->sbh.lfmFrameEnd + parea->sbh.cbSfh + cbFrame - 1;

	// Write a byte at this offset to make sure we have enough disk space

	if (!FmaWriteFile(parea->hfileSqd, offs, &ch, 1, NULL))
	  return FALSE;

	// Set the caller frame offset and update the base header

	*plfm = parea->sbh.lfmFrameEnd;
	parea->sbh.lfmFrameEnd = offs + 1;

	// Note that we don't link this space to any 
	// frame chains at this moment

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets specified message frame offset
//

  FMAAPI_(BOOL) SquGetFrame(PSAREA parea, PLONG plfm, MSGN msgn)
  {
	SQI sqi;

	ASSERT(parea != NULL);
	ASSERT(plfm != NULL);
	ASSERT(msgn && msgn <= parea->sbh.msgnHigh);

	// Check if current or next/prev message and 
	// return cached values if so

	if (msgn == parea->msgnCur) {
	  *plfm = parea->lfmCur;
	  return TRUE;
	} else 
	if (msgn == parea->msgnCur - 1) {
	  *plfm = parea->lfmPrev;
	  return TRUE;
	} else
	if (msgn == parea->msgnCur + 1) {
	  *plfm = parea->lfmNext;
	  return TRUE;
	}

	// Load index entry and check if ok

	if (!SquGetSqi(parea, &sqi, msgn))
	  return FALSE;

	// Set caller's frame offset and leave

	*plfm = sqi.lfm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the specified frame next frame offset
//

  FMAAPI_(BOOL) SquSetNextFrame(PSAREA parea, LONG lfm, LONG lfmNext)
  {
	SFH sfh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	// Load this frame header and check if ok

	if (!SquLoadSfh(parea, &sfh, lfm))
	  return FALSE;

	// Update the next frame offset

	sfh.lfmNext = lfmNext;

	// Save this frame header and check if ok

	if (!SquSaveSfh(parea, &sfh, lfm))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the specified frame previous frame offset
//

  FMAAPI_(BOOL) SquSetPrevFrame(PSAREA parea, LONG lfm, LONG lfmPrev)
  {
	SFH sfh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	// Load this frame header and check if ok

	if (!SquLoadSfh(parea, &sfh, lfm))
	  return FALSE;

	// Update the previous frame offset

	sfh.lfmPrev = lfmPrev;

	// Save this frame header and check if ok

	if (!SquSaveSfh(parea, &sfh, lfm))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if the specified frame is in used chain
//

  FMAAPI_(BOOL) SquCheckUsed(PSAREA parea, LONG lfm)
  {
	LONG lfmThis;
	SFH sfh;

	ASSERT(parea != NULL);

	ASSERT(SquCheckSbh(parea));

	// Scan through the used chain looking for a match

	for (lfmThis = parea->sbh.lfmFirst; lfmThis > 0; lfmThis = sfh.lfmNext) {

	  // Check for a match

	  if (lfm == lfmThis) 
		return TRUE;

	  // Load the frame header and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfmThis))
		return FALSE;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the specified frame to the used chain
//

  FMAAPI_(BOOL) SquAddUsed(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	ASSERT(SquCheckSbh(parea));
	ASSERT(!SquCheckUsed(parea, lfm));

	// Change frame type to used

	psfh->type = FRAME_USED;

	// Check if there are no used frames

	if (!parea->sbh.lfmFirst) {

	  // Reset this frame links

	  psfh->lfmPrev = 0;
	  psfh->lfmNext = 0;

	  // Save this frame header and check if ok

	  if (!SquSaveSfh(parea, psfh, lfm))
		return FALSE;

	  // Update the squish base header used chain links

	  parea->sbh.lfmFirst = lfm;
	  parea->sbh.lfmLast = lfm;

	  return TRUE;
	}

	// If we already have used frames link this frame 
	// to the last used frame

	psfh->lfmPrev = parea->sbh.lfmLast;
	psfh->lfmNext = 0;

	// Link last used frame to this frame and check if ok

	if (!SquSetNextFrame(parea, psfh->lfmPrev, lfm))
	  return FALSE;

	// Now save the this frame header and if failed, 
	// unlink previous last used frame from this one

	if (!SquSaveSfh(parea, psfh, lfm)) {
	  SquSetNextFrame(parea, psfh->lfmPrev, 0);
	  return FALSE;
	}

	// Finally update base header last used frame link
	// to point to this frame

	parea->sbh.lfmLast = lfm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes the specified frame from the used chain
//

  FMAAPI_(BOOL) SquDelUsed(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	ASSERT(SquCheckSbh(parea));
	ASSERT(SquCheckUsed(parea, lfm));

	// Check that the frames which pretend to be chain heads/tails are
	// actually what they seem

	if ((!psfh->lfmPrev && lfm != parea->sbh.lfmFirst) ||
		(!psfh->lfmNext && lfm != parea->sbh.lfmLast)) {
	  FMAERROR(INVALID_FRAME);
	  return FALSE;
	}

	// Remove this message from the used frame chain

	if (psfh->lfmPrev) {
	  if (!SquSetNextFrame(parea, psfh->lfmPrev, psfh->lfmNext))
		return FALSE;
	}

	if (psfh->lfmNext) {
	  if (!SquSetPrevFrame(parea, psfh->lfmNext, psfh->lfmPrev))
		return FALSE;
	}

	// Update the head and tail links for the used frame chain

	if (parea->sbh.lfmFirst == lfm) parea->sbh.lfmFirst = psfh->lfmNext;
	if (parea->sbh.lfmLast == lfm) parea->sbh.lfmLast = psfh->lfmPrev;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if the specified frame is in free chain
//

  FMAAPI_(BOOL) SquCheckFree(PSAREA parea, LONG lfm)
  {
	LONG lfmThis;
	SFH sfh;

	ASSERT(parea != NULL);

	ASSERT(SquCheckSbh(parea));

	// Scan through the used chain looking for a match

	for (lfmThis = parea->sbh.lfmFree; lfmThis > 0; lfmThis = sfh.lfmNext) {

	  // Check for a match

	  if (lfm == lfmThis) 
		return TRUE;

	  // Load the frame header and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfmThis))
		return FALSE;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the specified frame to the free chain
//

  FMAAPI_(BOOL) SquAddFree(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	ASSERT(SquCheckSbh(parea));
	ASSERT(!SquCheckFree(parea, lfm));

	// Change frame type to free and reset message info

	psfh->type = FRAME_FREE;
	psfh->cbCtrl = 0;
	psfh->cbMsg = 0;

	// Check if there are no free frames

	if (!parea->sbh.lfmFree) {

	  // Reset this frame links

	  psfh->lfmPrev = 0;
	  psfh->lfmNext = 0;

	  // Save this frame header and check if ok

	  if (!SquSaveSfh(parea, psfh, lfm))
		return FALSE;

	  // Update the squish base header free chain links

	  parea->sbh.lfmFree = lfm;
	  parea->sbh.lfmLastFree = lfm;

	  return TRUE;
	}

	// If we already have free frames link this frame 
	// to the last free frame

	psfh->lfmPrev = parea->sbh.lfmLastFree;
	psfh->lfmNext = 0;

	// Link last free frame to this frame and check if ok

	if (!SquSetNextFrame(parea, psfh->lfmPrev, lfm))
	  return FALSE;

	// Now save the this frame header and if failed, 
	// unlink previous last free frame from this one

	if (!SquSaveSfh(parea, psfh, lfm)) {
	  SquSetNextFrame(parea, psfh->lfmPrev, 0);
	  return FALSE;
	}

	// Finally update base header last free frame link
	// to point to this frame

	parea->sbh.lfmLastFree = lfm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes the specified frame from the free chain
//

  FMAAPI_(BOOL) SquDelFree(PSAREA parea, PSFH psfh, LONG lfm)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	ASSERT(SquCheckSbh(parea));
	ASSERT(SquCheckFree(parea, lfm));

	// Check that the frames which pretend to be chain heads/tails are
	// actually what they seem

	if ((!psfh->lfmPrev && lfm != parea->sbh.lfmFree) ||
		(!psfh->lfmNext && lfm != parea->sbh.lfmLastFree)) {
	  FMAERROR(INVALID_FRAME);
	  return FALSE;
	}

	// Remove this message from the free frame chain

	if (psfh->lfmPrev) {
	  if (!SquSetNextFrame(parea, psfh->lfmPrev, psfh->lfmNext))
		return FALSE;
	}

	if (psfh->lfmNext) {
	  if (!SquSetPrevFrame(parea, psfh->lfmNext, psfh->lfmPrev))
		return FALSE;
	}

	// Update the head and tail links for the free frame chain

	if (parea->sbh.lfmFree == lfm) parea->sbh.lfmFree = psfh->lfmNext;
	if (parea->sbh.lfmLastFree == lfm) parea->sbh.lfmLastFree = psfh->lfmPrev;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine finds free frame with length larger than specified
//

  FMAAPI_(BOOL) SquFindFree(PSAREA parea, PSFH psfh, PLONG plfm, ULONG cbFrame)
  {
	LONG lfmThis, lfmLast;
	SFH sfh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);

	ASSERT(SquCheckSbh(parea));

	// Assume we found nothing

	*plfm = 0;

	// Loop through all of the entries in the free chain

	for (lfmLast = 0, lfmThis = parea->sbh.lfmFree; lfmThis;
		 lfmLast = lfmThis, lfmThis = sfh.lfmNext) {

	  // Load this frame header and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfmThis))
		return FALSE;

	  // Check if this is a free frame and we're not circling

	  if (sfh.type != FRAME_FREE ||
		  sfh.lfmPrev != lfmLast || 
		  sfh.lfmNext == lfmThis) {
		FMAERROR(INVALID_FRAME);
		return FALSE;
	  }

	  // Check if frame is large enough and leave

	  if (sfh.cbFrame >= cbFrame) {
		*plfm = lfmThis;
		*psfh = sfh;
		break;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  FMAAPI_(BOOL) SquLoadSmh(PSAREA parea, PSMH psmh, LONG lfm)
  {
	ULONG cbRead;
	LONG offs;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

	ASSERT(lfm >= (LONG)parea->sbh.cb);
	ASSERT(lfm <= (LONG)parea->sbh.lfmFrameEnd);

	ASSERT(psmh != NULL);

	// Read the message header and check if ok

	offs = lfm + parea->sbh.cbSfh;
	if (!FmaReadFile(parea->hfileSqd, offs, psmh, sizeof(SMH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(SMH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  FMAAPI_(BOOL) SquSaveSmh(PSAREA parea, PSMH psmh, LONG lfm)
  {
	ULONG cbWritten;
	LONG offs;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

	ASSERT(lfm >= (LONG)parea->sbh.cb);
	ASSERT(lfm <= (LONG)parea->sbh.lfmFrameEnd);

	ASSERT(psmh != NULL);

	// Write the message header and check if ok

	offs = lfm + parea->sbh.cbSfh;
	if (!FmaWriteFile(parea->hfileSqd, offs, psmh, sizeof(SMH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(SMH));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns safe message and control info lengths
//

  FMAAPI_(BOOL) SquGetSafeMsgCtrlLen(PSMSG pmsg, PULONG pcbMsg, PULONG pcbCtrl)
  {
	ULONG cbCtrl, cbMsg;

	ASSERT(pmsg != NULL);

	// Check if frame size is valid and return zero lengths if not
	// This is necessary since some msgs out there have these
	// invalid but we dont want to fail reading such messages

	if (pmsg->sfh.cbFrame < sizeof(SMH)) {
	  cbCtrl = 0; cbMsg = 0;
	} else {
	  cbCtrl = pmsg->sfh.cbCtrl; cbMsg = pmsg->sfh.cbMsg;

	  // Check message length and if invalid, assume message 
	  // spans entire frame

	  if (cbMsg < sizeof(SMH) || cbMsg > pmsg->sfh.cbFrame) {
		FMAERROR(INVALID_SIZE);
		cbMsg = pmsg->sfh.cbFrame;
	  }

	  // Check message info length and if invalid, assume there
	  // is no control info so that it will be read as message text

	  if (sizeof(SMH) + cbCtrl > cbMsg) {
		FMAERROR(INVALID_SIZE);
		cbCtrl =  0;
	  }
	}

	// Set the values for caller

	if (pcbCtrl) *pcbCtrl = cbCtrl;
	if (pcbMsg) *pcbMsg = cbMsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  FMAAPI_(PSMSG) SquGetOpenMsg(PSAREA parea, MSGN msgn)
  {
	PSMSG pmsg;

	ASSERT(parea  != NULL);

	// Scan through the open messages list looking for a match

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->msgn == msgn)
		return pmsg;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message at the specified frame offset
//

  FMAAPI_(BOOL) SquDelMsg(PSAREA parea, MSGN msgn, PSFH psfh, LONG lfm)
  {
	SFH sfh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(psfh->magic == SFH_MAGIC);

	// Remove this message from the used frame chain

	if (psfh->lfmPrev) {
	  if (!SquSetNextFrame(parea, psfh->lfmPrev, psfh->lfmNext))
		return FALSE;
	}

	if (psfh->lfmNext) {
	  if (!SquSetPrevFrame(parea, psfh->lfmNext, psfh->lfmPrev))
		return FALSE;
	}

	// Update the head and tail links for the used frame chain

	if (parea->sbh.lfmFirst == lfm) parea->sbh.lfmFirst = psfh->lfmNext;
	if (parea->sbh.lfmLast == lfm) parea->sbh.lfmLast = psfh->lfmPrev;

	// Delete associated entry in the index and check if ok

	if (!SquDelSqi(parea, NULL, msgn))
	  return FALSE;

	// Fix cached prev/next links

	if (msgn == parea->msgnCur + 1) parea->lfmNext = psfh->lfmNext;
	if (msgn == parea->msgnCur - 1) parea->lfmPrev = psfh->lfmPrev;

	// Check if deleted current message

	if (msgn == parea->msgnCur) {

	  // Load the previous message frame header and if ok, adjust 
	  // the cached prev/next links and current message number, 
	  // otherwise goto to the first message

	  if (SquLoadSfh(parea, &sfh, psfh->lfmPrev)) {
		parea->msgnCur--;
		parea->lfmCur  = psfh->lfmPrev;
		parea->lfmPrev = sfh.lfmPrev;
		parea->lfmNext = sfh.lfmNext;
	  } else {
		SquResetCurMsgCache(parea);
	  }
	} else {

	  // If deleted message before the current one, 
	  // decrement the current message number

	  if (msgn < parea->msgnCur) parea->msgnCur--;
	}

	// Adjust message count and highest message number

	parea->sbh.cmsg--; parea->sbh.msgnHigh--;

	// Adjust the high water mark if it's above or equal to deleted message.
	// This looks weird since high water mark is stored as unique message id and 
	// not as message number, however we still do this for compatibility reasons.

	if (parea->sbh.umsgHwtr >= msgn) parea->sbh.umsgHwtr--;

	// Adjust message numbers for all open messages above the deleted one

	DoFixOpenMsg(parea, msgn);

	// Add this frame to the free chain and check if ok

	if (!SquAddFree(parea, psfh, lfm))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine trims base message count
//

  FMAAPI_(BOOL) SquTrimBase(PSAREA parea, BOOL bAdding)
  {
	LONG lfmFirst = 0, lfmFirstPrev = 0;
	LONG lfm, lfmPrev, lfmNext;
	ULONG cmsgMax, cmsgDeleted = 0;
	MSGN msgnDelete;
	BOOL bOk = TRUE;
	SFH sfh;
	SQI sqi;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	// Check if there is no message number limit or
	// message base is empty and leave if so

	if (!parea->sbh.cmsgMax || !parea->sbh.cmsg) return TRUE;

	// Check if there are only messages we have to keep

	if (parea->sbh.cmsg <= parea->sbh.cmsgSkip) return TRUE;

	// Calculate target number of messages

	cmsgMax = !bAdding ? parea->sbh.cmsgMax : parea->sbh.cmsgMax - 1;

	// Check if the message limit is not exceeded and leave

	if (parea->sbh.cmsg <= cmsgMax) return TRUE;

	// Find out the message number to start deleting at

	msgnDelete = 1 + parea->sbh.cmsgSkip;

	// Delete enough messages to stay within limit not touching
	// first sbh.cmsgSkip messages

	while (parea->sbh.cmsg > cmsgMax) {

	  // Check if this message is open and leave early.

	  if (SquGetOpenMsg(parea, msgnDelete)) {
		FMAERROR(STILL_OPEN);
		bOk = FALSE;
		break;
	  }

	  // Get message index entry and load its frame header

	  if (!SquGetSqi(parea, &sqi, msgnDelete) ||
		  !SquLoadSfh(parea, &sfh, sqi.lfm)) {
		bOk = FALSE;
		break;
	  }

	  // Add this frame to the free chain saving its 
	  // prev/next links since the call would change them

	  lfmPrev = sfh.lfmPrev; 
	  lfmNext = sfh.lfmNext;

	  if (!SquAddFree(parea, &sfh, sqi.lfm)) {
		bOk = FALSE;
		break;
	  }

	  // Save the new first frame and its previous frame offset if
	  // we don't have it already

	  lfmFirst = lfmNext;
	  if (!lfmFirstPrev) lfmFirstPrev = lfmPrev;

	  // Now delete associated index entry

	  if (!SquDelSqi(parea, &sqi, msgnDelete)) {
		bOk = FALSE;
		break;
	  }

	  // Adjust message numbers for all open messages above the deleted one

	  DoFixOpenMsg(parea, msgnDelete);

	  // Adjust message count and highest message number and
	  // increment number of messages deleted

	  parea->sbh.cmsg--; parea->sbh.msgnHigh--;
	  cmsgDeleted++;
	}

	// Adjust the first frame offset and links

	if (!parea->sbh.cmsgSkip) {
	  parea->sbh.lfmFirst = lfmFirst;
	  if (!SquSetPrevFrame(parea, lfmFirst, 0))
		bOk = FALSE;
	} else {
	  if (!SquSetNextFrame(parea, lfmFirstPrev, lfmFirst) ||
		  !SquSetPrevFrame(parea, lfmFirst, lfmFirstPrev))
		bOk = FALSE;
	}

	// Check if current message is right after the deleted range and 
	// adjust cached previous frame offset

	if (parea->msgnCur == 1 + parea->sbh.cmsgSkip + cmsgDeleted) {
	  if (!parea->sbh.cmsgSkip) {
		parea->lfmPrev = 0;
	  } else {
		if (!SquGetFrame(parea, &parea->lfmPrev, parea->sbh.cmsgSkip))
		  bOk = FALSE;
	  }
	} else

	// Check if current message is right before the deleted range and 
	// adjust cached next frame offset

	if (parea->msgnCur == parea->sbh.cmsgSkip) {
	  parea->lfmPrev = lfmFirst;
	}

	// Check if current message was above sbh.cmsgSkip and 
	// adjust cached frame offsets

	if (parea->msgnCur > parea->sbh.cmsgSkip) {

	  // Check if current message was inside the deleted range and if so,
	  // move current message to the first message above that range

	  if (parea->msgnCur <= parea->sbh.cmsgSkip + cmsgDeleted) {
		if (parea->sbh.cmsgSkip && 
			SquGetFrame(parea, &lfm, parea->sbh.cmsgSkip) &&
			SquLoadSfh(parea, &sfh, lfm)) {
		  parea->msgnCur = parea->sbh.cmsgSkip;
		  parea->lfmCur  = lfm;
		  parea->lfmPrev = sfh.lfmPrev;
		  parea->lfmNext = sfh.lfmNext;
		} else {
		  SquResetCurMsgCache(parea);
		}
	  } else {

		// Current message was above the deleted range so decrement it

		parea->msgnCur-= cmsgDeleted;
	  }
	}

	// Adjust the high water mark if necessary. This looks weird since 
	// high water mark is stored as unique message id and not as message number, 
	// however we still do this for compatibility reasons.

	if (parea->sbh.umsgHwtr >= parea->sbh.cmsgSkip) {
	  LONG hwtr = parea->sbh.umsgHwtr - cmsgDeleted;
	  if (hwtr < (LONG)parea->sbh.cmsgSkip)
		parea->sbh.umsgHwtr = parea->sbh.cmsgSkip;
	  else
		parea->sbh.umsgHwtr = hwtr;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine calculates squish hash for the given string
//

  FMAAPI_(ULONG) SquGetHash(PCH pchText, UINT cchText)
  {
	ULONG fix, hash = 0l;
	UINT cch;
	PCH pch;

	// Check if nothing to do and return original crc

	if (!pchText || !cchText) return hash;

	// Check if asciiz string and get its length

	if (cchText == (UINT)-1) cchText = strlen(pchText);

	// Calculate hash value for the given string

	for (pch = pchText, cch = 0; cch < cchText; pch++, cch++) {
	  hash = (hash << 4) + (ULONG)TOLOWER(*pch);
	  if ((fix = (hash & 0xf0000000l)) != 0l) {
		hash|= fix >> 24;
		hash|= fix;
	  }
	}

	// Strip off high bit

	return hash & 0x7FFFFFFFlu;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquHelp.c
/////////////////////////////////////////////////////////////////////////////
