/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish index routines module
//
// SquIndex.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 14/Sep/1999
// Updated: 23/Mar/2000
//
// History: 14/Sep/1999 kvitekp - Initially created.
//
//          23/Mar/2000 kvitekp - Relaxed empty SQI check in DoCheckSqiCache
//
/////////////////////////////////////////////////////////////////////////////
//
// Remarks: This code implements virtual squish index which can be file based
//          or memory based (i.e. cached). The caching is controlled by calls
//          to SquBeginSqiCache()/SquEndSqiCache() which can be nested. When 
//          not caching, all index manipulation apis update the index file.
//          When caching, the same apis update cached index in memory and
//          the index file is updated during a call to SquEndSqiCache()...
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Index cache allocation constants

#define SIC_SQI_MAX  ((1*1024)/sizeof(SQI)) // Max number of index entries on page
#define SIC_SQI_GROW SIC_SQI_MAX            // Index cache grow factor

  // Macro to check for empty index entry

#define SQI_EMPTY(psqi) ((psqi)->lfm == 0)

  // Module static data

  static SQI sqiEmpty = SQI_INIT_EMPTY;

  // Function prototypes

#ifndef NDEBUG  
  static BOOL FMASUB DoDumpSqiCache(PSAREA parea);
  static BOOL FMASUB DoCheckSqiCache(PSAREA parea);
  static BOOL FMASUB DoCompareSqiCache(PSAREA parea);
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds squish index cache page
//

  static PSIC FMASUB DoAddSicPage(PSAREA parea, UINT csqi)
  {
	PSIC psic;
	UINT cb;

	ASSERT(parea != NULL);
	ASSERT(csqi > 0);

	// Allocate new cache page and check if ok

	cb = sizeof(SIC) + (csqi - 1) * sizeof(SQI);
	if ((psic = MemAlloc(cb, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} else
	  psic->csqi = csqi;

	// Link up the new cache page

	ListAdd(&parea->psicList, psic, LIST_END);

#ifdef FMA_LOG_SQU_INDEX
//Log("DoAddSicPage      : [%09Fp] csqi=%d", parea, psic->csqi);
#endif

	return psic;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes squish index cache page
//

  static BOOL FMASUB DoDelSicPage(PSAREA parea, PSIC psicDel)
  {
	PSIC psic;
	UINT isic;

	ASSERT(parea != NULL);

	// Scan index cache pages looking for the specified page and if found,
	// unlink and destroy it

	for (isic = 0, psic = parea->psicList; psic != NULL; psic = psic->psicNext, isic++) {
	  if (psic == psicDel) {
		psic = ListDel(&parea->psicList, isic);
		ASSERT(psic != NULL);
		MemFree(psic);

#ifdef FMA_LOG_SQU_INDEX
//Log("DoDelSicPage      : [%09Fp]", parea);
#endif
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes squish index cache
//

  static VOID FMASUB DoDelSic(PSAREA parea)
  {
	PSIC psic;

	ASSERT(parea != NULL);

	// Delete all index cache pages

	while (parea->psicList) {
	  psic = ListDel(&parea->psicList, 0);
	  ASSERT(psic != NULL);
	  MemFree(psic);
	}

	// Reset index cache dirty pointers

	parea->msgnLoDirty = 0;
	parea->msgnHiDirty = 0;

#ifdef FMA_LOG_SQU_INDEX
//Log("DoDelSic          : [%09Fp]", parea);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads specified number of squish index entries
//

  static LONG FMASUB DoReadSqi(PSAREA parea, ULONG isqi, SQI asqi[], ULONG csqi)
  {
	ULONG cb, cbRead;
	LONG offs;

	ASSERT(asqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Find out specified index entry offset and number of bytes to read

	offs = isqi * sizeof(SQI);
	cb   = csqi * sizeof(SQI);

	// Read index entries and check if ok

	if (!FmaReadFile(parea->hfileSqi, offs, asqi, cb, &cbRead))
	  return -1;

	// Return number of index entries read

	return cbRead / sizeof(SQI);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes specified number of squish index entries
//

  static LONG FMASUB DoWriteSqi(PSAREA parea, ULONG isqi, SQI asqi[], ULONG csqi)
  {
	ULONG cb, cbWritten;
	LONG offs;

	ASSERT(asqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Find out specified index entry offset and number of bytes to write

	offs = isqi * sizeof(SQI);
	cb   = csqi * sizeof(SQI);

	// Write index entries and check if ok

	if (!FmaWriteFile(parea->hfileSqi, offs, asqi, cb, &cbWritten))
	  return -1;

	ASSERT(cbWritten == cb);

	// Return number of index entries written

	return cbWritten / sizeof(SQI);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads sqi cache
//

  static BOOL FMASUB DoLoadSic(PSAREA parea)
  {
	UINT isqi, csqi;
	ULONG msgi;
	PSIC psic;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Delete existing squish index cache if any

	DoDelSic(parea);

	// Find out total number of sqi entries to load and
	// loop reading index entries

	for (csqi = (UINT)(FmaGetFileSize(parea->hfileSqi) / sizeof(SQI)), msgi = 0; 
		 csqi > 0; csqi-= isqi, msgi+= isqi) {

	  // Calculate number if index entries to load into cache page

	  isqi = min(csqi, SIC_SQI_MAX);

	  // Allocate cache page  and check if ok

	  if ((psic = DoAddSicPage(parea, isqi)) == NULL) {
Fail:   DoDelSic(parea);
		return FALSE;
	  }

	  // Read cache page worth portion of the index file

	  if (DoReadSqi(parea, msgi, psic->asqi, isqi) != (LONG)isqi) {
		goto Fail;
	  }

	  // Set number of used index entries on this cache page

	  psic->isqi = isqi;
	}

#ifdef FMA_LOG_SQU_INDEX
Log("DoLoadSic         : [%09Fp] csqi=%ld", parea, FmaGetFileSize(parea->hfileSqi) / sizeof(SQI));
#endif

	ASSERT(DoCompareSqiCache(parea));
	ASSERT(DoCheckSqiCache(parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves sqi cache
//

  static BOOL FMASUB DoSaveSic(PSAREA parea)
  {
	MSGN msgnLoDirty, msgnHiDirty;
	MSGN msgnFrom, msgnTo, msgn;
	PSQI psqi;
	PSIC psic;
	int csqi;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Pick up dirty boundary on stack

	msgnLoDirty = parea->msgnLoDirty;
	msgnHiDirty = parea->msgnHiDirty;

	ASSERT(msgnLoDirty <= msgnHiDirty);

	// Check if index cache is dirty

	if (msgnLoDirty || msgnHiDirty) {

	  // Loop through all the index cache pages

	  for (msgn = 1, psic = parea->psicList; psic != NULL; 
		   msgn+= psic->isqi, psic = psic->psicNext) {

		// Check if this cache page does not contain index entries

		if (!psic->isqi) continue;

		// Check if this cache page index entries are entirely 
		// below the low dirty boundary and skip it

		if (msgn + psic->isqi - 1 < msgnLoDirty) continue;

		// Check if this cache page index entries are entirely 
		// above the hight dirty boundary and leave loop

		if (msgn > msgnHiDirty) break;

		// Find out the region of index entries to save on this page

		msgnFrom = max(msgn, msgnLoDirty);
		msgnTo = min(msgn + psic->isqi - 1, msgnHiDirty);

		ASSERT(msgnFrom <= msgnTo);

		// Calulate pointer to index entry and number of index entries to save

		psqi = &psic->asqi[msgnFrom - msgn];
		csqi = (int)(msgnTo - msgnFrom + 1);

		// Write out index entries and check if ok

		if (DoWriteSqi(parea, msgnFrom - 1, psqi, csqi) != (LONG)csqi) {
		  return FALSE;
		} 

#ifdef FMA_LOG_SQU_INDEX
Log("DoSaveSic         : [%09Fp] msgnLoDirty=%ld msgnHiDirty=%ld msgnFrom=%ld msgnTo=%ld", parea, parea->msgnLoDirty, parea->msgnHiDirty, msgnFrom, msgnTo);
#endif
	  }
	}

	ASSERT(DoCheckSqiCache(parea));
	ASSERT(DoCompareSqiCache(parea));

	// Delete squish index cache

	DoDelSic(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds specified range to index cache dirty range
//

  static BOOL FMASUB DoAddDirty(PSAREA parea, MSGN msgnFrom, MSGN msgnTo)
  {
	BOOL bChanged = FALSE;

	ASSERT(parea != NULL);
	ASSERT(msgnFrom != 0);
	ASSERT(msgnTo != 0);
	ASSERT(msgnFrom <= msgnTo);

	// Adjust low dirty boundary

	if (!parea->msgnLoDirty || msgnFrom < parea->msgnLoDirty) {
	  parea->msgnLoDirty = msgnFrom;
	  bChanged = TRUE;
	}

	// Adjust high dirty boundary

	if (!parea->msgnHiDirty || msgnTo > parea->msgnHiDirty) {
	  parea->msgnHiDirty = msgnTo;
	  bChanged = TRUE;
	}

#ifdef FMA_LOG_SQU_INDEX
Log("DoAddDirty        : [%09Fp] msgnLoDirty=%ld msgnHiDirty=%ld msgnFrom=%ld msgnTo=%ld", parea, parea->msgnLoDirty, parea->msgnHiDirty, msgnFrom, msgnTo);
#endif

	return bChanged;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine dumps squish index cache
//
#ifndef NDEBUG
  static BOOL FMASUB DoDumpSqiCache(PSAREA parea)
  {
	MSGN msgn = 1;
	PSIC psic;
	UINT isqi;

	ASSERT(parea != NULL);

	Log("DoDumpSqiCache   >:");

	for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
	  Log(" psic=[%09Fp] csqi=%u, isqi=%u", psic, psic->csqi, psic->isqi);
	  for (isqi = 0; isqi < psic->isqi; isqi++, msgn++) {
		Log("  %4lu: %08lx %ld %08lx", msgn, psic->asqi[isqi].lfm, psic->asqi[isqi].umsg, psic->asqi[isqi].hash);
	  }
	}

	Log("DoDumpSqiCache   <:");

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine checks squish index cache
//
#ifndef NDEBUG
  static BOOL FMASUB DoCheckSqiCache(PSAREA parea)
  {
	BOOL bIsEmpty, bInEmpty = FALSE;
	MSGN msgn = 1;
	UMSG umsg = 0;
	PSIC psic;
	UINT isqi;

	ASSERT(parea != NULL);

	for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {

	  // Check if page index entry counters are valid

	  if (!psic->isqi || !psic->csqi || psic->isqi > psic->csqi) {
		Log("DoCheckSqiCache   : csqi=%d isqi=%d (BAD PAGE)", psic->csqi, psic->isqi);
Fail:   DoDumpSqiCache(parea);
		return FALSE;
	  }

	  // Check this page index entries 

	  for (isqi = 0; isqi < psic->isqi; isqi++, msgn++) {
		bIsEmpty = SQI_EMPTY(&psic->asqi[isqi]);

		// Check if not in empty entries and if hit empty, switch to empty,
		// otherwise check the non empty index entry

		if (!bInEmpty) {
		  if (bIsEmpty) {
			bInEmpty = TRUE;
		  } else {

			// Check if it has valid frame offset

			if (psic->asqi[isqi].lfm < (LONG)parea->sbh.cb || 
				psic->asqi[isqi].lfm > (LONG)parea->sbh.lfmFrameEnd) {
			  Log("DoCheckSqiCache   : msgn=%ld -- %08lx %ld %08lx (BAD FRAME OFFSET)", msgn,
				   psic->asqi[isqi].lfm, psic->asqi[isqi].umsg, psic->asqi[isqi].hash);
			  goto Fail;
			}

			// Check if it has ascending umsg

			if (psic->asqi[isqi].umsg > umsg) {
			  umsg = psic->asqi[isqi].umsg;
			} else {
			  Log("DoCheckSqiCache   : msgn=%ld -- %08lx %ld %08lx (UMSG NOT ASCENDING)", msgn, 
				   psic->asqi[isqi].lfm, psic->asqi[isqi].umsg, psic->asqi[isqi].hash);
			  goto Fail;
			}
		  }
		} else {

		  // If in empty block, check if got non empty and complain

		  if (!bIsEmpty) {
			  Log("DoCheckSqiCache   : msgn=%ld -- %08lx %ld %08lx (NON EMPTY AFTER EMPTY)", msgn,
				   psic->asqi[isqi].lfm, psic->asqi[isqi].umsg, psic->asqi[isqi].hash);
			  goto Fail;
		  }
		}
	  }
	}

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// This routine compares squish index cache to an index file
//
#ifndef NDEBUG
  static BOOL FMASUB DoCompareSqiCache(PSAREA parea)
  {
	MSGN msgn = 1;
	PSIC psic;
	UINT isqi;
	SQI sqi;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Scan through all cached index entries

	for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
	  for (isqi = 0; isqi < psic->isqi; isqi++, msgn++) {

		// Read index entry and check if ok

		if (DoReadSqi(parea, msgn - 1, &sqi, 1) != 1) {
		  Log("DoCompareSqiCache : can't read index file");
		  return FALSE;
		}

		// Compare cached and file index entry

		if (memcmp(&psic->asqi[isqi], &sqi, sizeof(SQI))) {
		  Log("DoCompareSqiCache : msgn=%ld, cache/disk: %08lx/%08lx %ld/%ld %08lx/%08lx",
			   psic->asqi[isqi].lfm, sqi.lfm,
			   psic->asqi[isqi].umsg, sqi.umsg,
			   psic->asqi[isqi].hash, sqi.hash);
		  return FALSE;
		}
	  }
	}

	return TRUE;
  }
#endif
/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine begins squish message base index caching
//

  FMAAPI_(BOOL) SquBeginSqiCache(PSAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

#ifdef FMA_LOG_SQU_INDEX
Log("SquBeginSqiCache >: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif

	// Check if already caching and leave incrementing cache count

	if (parea->bSqiCached) {
	  parea->bSqiCached++;
#ifdef FMA_LOG_SQU_INDEX
Log("SquBeginSqiCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d ALREADY CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif
	  return TRUE;
	}

	// Load squish index cache and check if ok

	if (!DoLoadSic(parea)) {
#ifdef FMA_LOG_SQU_INDEX
Log("SquBeginSqiCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d CANNOT LOAD CACHE", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif
	  return FALSE;
	}

	// Increment cache count

	parea->bSqiCached++;

#ifdef FMA_LOG_SQU_INDEX
Log("SquBeginSqiCache <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d NOW CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine ends squish message base index caching
//

  FMAAPI_(BOOL) SquEndSqiCache(PSAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

#ifdef FMA_LOG_SQU_INDEX
Log("SquEndSqiCache   >: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif

	// Check if not caching and fail

	if (!parea->bSqiCached) {
	  FMAERROR(NOT_CACHING);
#ifdef FMA_LOG_SQU_INDEX
Log("SquEndSqiCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d NOT CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif
	  return FALSE;
	}

	// Check if caching more than once and leave decrementing cache count

	if (parea->bSqiCached > 1) {
	  parea->bSqiCached--;
#ifdef FMA_LOG_SQU_INDEX
Log("SquEndSqiCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d STILL CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif
	  return TRUE;
	}

	ASSERT(parea->bSqiCached == 1);

	// Save squish index cache and check if ok

	if (!DoSaveSic(parea)) {
#ifdef FMA_LOG_SQU_INDEX
Log("SquEndSqiCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d CANNOT SAVE CACHE", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif
	  return FALSE;
	}

	// Decrement cache count

	parea->bSqiCached--;

#ifdef FMA_LOG_SQU_INDEX
Log("SquEndSqiCache   <: [%09Fp] bBaseLocked=%d bExclusive=%d bSqiCached=%d NOW NOT CACHING", parea, parea->bBaseLocked, parea->bExclusive, parea->bSqiCached);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns number of entries in squish index
//

  FMAAPI_(ULONG) SquGetSqiCount(PSAREA parea)
  {
	ULONG csqi = 0;
	PSIC psic;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  // Scan index cache pages adding up the used entries counts

	  for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
		csqi+= psic->isqi;
	  }

	} else {

	  // If squish index is not cached just calculate the number 
	  // of index entries using the current idnex file size

	  csqi = FmaGetFileSize(parea->hfileSqi) / sizeof(SQI);
	}

	return csqi;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets squish index entry
//

  FMAAPI_(BOOL) SquGetSqi(PSAREA parea, PSQI psqi, MSGN msgn)
  {
	ULONG isqi;
	PSIC psic;

	ASSERT(psqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check for valid message number

	if (!msgn) return FALSE;

	// Calculate requested message index entry index

	isqi = msgn - 1;

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, copy index entry for the caller

	  for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
		if (psic->isqi && isqi < psic->isqi) {
		  *psqi = psic->asqi[isqi];
		  return TRUE;
		} else 
		  isqi-= psic->isqi;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // read the index entry and check if ok

	  if (isqi < SquGetSqiCount(parea)) {
		if (DoReadSqi(parea, isqi, psqi, 1) == 1) 
		  return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets squish index entry
//

  FMAAPI_(BOOL) SquSetSqi(PSAREA parea, PSQI psqi, MSGN msgn)
  {
	ULONG isqi;
	PSIC psic;

	ASSERT(psqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check for valid message number

	if (!msgn) return FALSE;

	// Calculate requested message index entry index

	isqi = msgn - 1;

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, update index entry and dirty range

	  for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
		if (psic->isqi && isqi < psic->isqi) {
		  DoAddDirty(parea, msgn, msgn);
		  psic->asqi[isqi] = *psqi;
		  return TRUE;
		} else 
		  isqi-= psic->isqi;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // write the index entry and check if ok

	  if (isqi < SquGetSqiCount(parea)) {
		if (DoWriteSqi(parea, isqi, psqi, 1) == 1) 
		  return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes squish index entry
//

  FMAAPI_(BOOL) SquDelSqi(PSAREA parea, PSQI psqi, MSGN msgn)
  {
	SQI asqi[512 / sizeof(SQI)];
	ULONG isqi, csqi;
	PSIC psic;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check for valid message number

	if (!msgn) return FALSE;

	// Calculate requested message index entry index

	isqi = msgn - 1;

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  if (msgn == 77) {
		msgn = msgn;
	  }

	  // Scan index cache looking for the page with specified message 
	  // index entry and if found, delete index entry and update dirty range

	  for (psic = parea->psicList; psic != NULL; psic = psic->psicNext) {
		if (psic->isqi && isqi < psic->isqi) {
		  if (psqi) *psqi = psic->asqi[isqi];
		  DoAddDirty(parea, msgn, SquGetSqiCount(parea));

		  // Check if this is the only used entry on this page and delete it, 
		  // otherwise move used index entries over the deleted one

		  if (psic->isqi == 1) {
			DoDelSicPage(parea, psic);
		  } else {
			memmove(&psic->asqi[isqi], &psic->asqi[isqi + 1], 
				   (int)(psic->isqi - isqi) * sizeof(SQI));
			psic->isqi--;
		  }

		  // Append an empty entry to keep index size the same and if failed,
		  // chop idnex file size to prevent leftover index entries

		  if (!SquAppSqi(parea, &sqiEmpty, NULL)) {
			if (!FmaSetFileSize(parea->hfileSqi, SquGetSqiCount(parea) * sizeof(SQI))) {
			  FMAERROR(CANNOT_DELETE);
			  return FALSE;
			}
		  }

		  ASSERT(DoCheckSqiCache(parea));
		  ASSERT(DoDumpSqiCache(parea));

		  return TRUE;
		} else 
		  isqi-= psic->isqi;
	  }

	} else {

	  // Check if not past the end of index file and if so,
	  // move the idnex entries over the deleted one

	  if (isqi < SquGetSqiCount(parea)) {

		// Check if caller needs the deleted index entry and read it

		if (psqi) {
		  if (DoReadSqi(parea, isqi, psqi, 1) != 1) 
			return FALSE;
		}

		// Loop moving the index entries over the deleted one

		loop {

		  // Read portion of index file and check if ok

		  if ((csqi = DoReadSqi(parea, isqi + 1, asqi, numbof(asqi))) == (ULONG)-1) 
			return FALSE;

		  // Check if read nothing and leave loop

		  if (!csqi) break;

		  // Write portion of index file and check if ok

		  if (DoWriteSqi(parea, isqi, asqi, csqi) != (LONG)csqi) 
			return FALSE;

		  // Advance to the next chunk and loop

		  isqi+= csqi;
		}

		// Overwrite the leftover entry with an empty one

		if (DoWriteSqi(parea, isqi, &sqiEmpty, 1) == 1) 
		  return TRUE;
	  }
	}

	// If we got here the specified message index entry could not be found

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine appends specified squish index entry to the end of index file
//

  FMAAPI_(BOOL) SquAppSqi(PSAREA parea, PSQI psqi, MSGN * pmsgn)
  {
	ULONG isqi;
	PSIC psic;
	MSGN msgn;

	ASSERT(psqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  // Scan index cache pages looking for the last one

	  for (msgn = 1, psic = parea->psicList; psic != NULL; 
		   msgn+= psic->isqi, psic = psic->psicNext) {
		if (!psic->psicNext) break;
	  }

	  // Check if there is no last cache page or it's all used and
	  // add another cache page if so

	  if (!psic || psic->isqi == psic->csqi) {
		if (psic) msgn+= psic->isqi;
		if ((psic = DoAddSicPage(parea, SIC_SQI_GROW)) == NULL) 
		  return FALSE;
	  } 

	  ASSERT(psic->isqi < psic->csqi);

	  // Add  the specified index entry on the last page

	  msgn+= psic->isqi;
	  psic->asqi[psic->isqi++] = *psqi;
	  DoAddDirty(parea, msgn, msgn);
	  if (pmsgn) *pmsgn = msgn;

	} else {

	  // Write index entry at the end of index file and check if ok

	  isqi = SquGetSqiCount(parea);
	  if (DoWriteSqi(parea, isqi, psqi, 1) != 1) 
		return FALSE;

	  // Check if caller wants message number and set it in

	  if (pmsgn) *pmsgn = isqi + 1;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds index entry by reusing an empty entry or appending a new
//

  FMAAPI_(BOOL) SquAddSqi(PSAREA parea, PSQI psqi, MSGN * pmsgn)
  {
	SQI asqi[512 / sizeof(SQI)];
	UINT isqi, csqi, msgi = 0;
	PSIC psic;
	MSGN msgn;

	ASSERT(psqi != NULL);
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqi != -1);

	// Check if squish index is cached

	if (parea->bSqiCached) {

	  // Scan index cache pages looking for the empty entry and
	  // overwrite it if found

	  for (msgn = 1, psic = parea->psicList; psic != NULL; 
		   msgn+= psic->isqi, psic = psic->psicNext) {
		for (isqi = 0; isqi < psic->isqi; isqi++) {
		  if (SQI_EMPTY(&psic->asqi[isqi])) {
			psic->asqi[isqi] = *psqi;
			msgn+= isqi;
			DoAddDirty(parea, msgn, msgn);
			if (pmsgn) *pmsgn = msgn;
			return TRUE;
		  }
		}
	  }

	} else {

	  // Loop looking for the empty index entry

	  loop {

		// Read portion of index file and check if ok

		if ((csqi = (UINT)DoReadSqi(parea, msgi, asqi, numbof(asqi))) == (UINT)-1) 
		  return FALSE;

		// Check if read nothing and leave loop

		if (!csqi) break;

		// Scan portion of the index file looking for the empty entry

		for (isqi = 0; isqi < csqi; isqi++) {
		  if (SQI_EMPTY(&asqi[isqi])) {

			// Overwrite the empty index entry

			msgi+= isqi;
			if (DoWriteSqi(parea, msgi, psqi, 1) != 1) 
			  return FALSE;

			// Set message number for callers

			if (pmsgn) *pmsgn = 1 + msgi;

			return TRUE;
		  }
		}

		// Advance to the next chunk and loop

		msgi+= csqi;
	  }
	}

	// If we got here the empty index entry has not been found,
	// so append a new entry

	return SquAppSqi(parea, psqi, pmsgn);
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquIndex.c
/////////////////////////////////////////////////////////////////////////////
