/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish locking routines module
//
// SquLock.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 14/Sep/1999
// Updated: 14/Sep/1999
//
// History: 14/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////
//
//  SquLockBase()       Locks the message base files so that no other processes
//                      could access it
//
//  SquEnterExclusive() Locks message base files and loads base header in order
//                      to modify base header, message frame chains or index. 
//                      Under no circumstances these should be updated outside 
//                      the exclusive state.
//  
//  SquLockArea()       Locks message base files and caches index for speedy 
//                      access. In order to modify message base you still
//                      need to acquire exclusive state.
//  
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine locks or unlocks squish message base
//

  FMAAPI_(BOOL) SquLockBase(PSAREA parea, BOOL bLock)
  {
	int nTry;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   >: [%09Fp] bBaseLocked=%d", bLock, parea, parea->bBaseLocked);
#endif

	if (bLock) {

	  // Check if already locked and leave incrementing the lock count

	  if (parea->bBaseLocked) {
		parea->bBaseLocked++;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d ALREADY LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  } 

	  ASSERT(!parea->bBaseLocked);

	  // Loop trying to lock the base

	  for  (nTry = 0; nTry < SBH_LOCK_RETRY; nTry++) {

		// Lock the base and check if ok

		if (FmaLock(parea->hfileSqd, SBH_LOCK_OFFS, SBH_LOCK_LENG, TRUE)) {
		  parea->bBaseLocked++;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		  return TRUE;
		}

		// Sleep for a second a retry

		FmaSleep(1000);
	  }

	  // If we got here the base can't be locked

#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d LOCK FAILED", bLock, parea, parea->bBaseLocked);
#endif

	} else {

	  // Check if base is not locked 

	  if (!parea->bBaseLocked) {
		FMAERROR(NOT_LOCKED);
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d NOT LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return FALSE;
	  }

	  // Check if locked more than once and leave decrementing the lock count

	  if (parea->bBaseLocked > 1) {
		parea->bBaseLocked--;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d STILL LOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  } 

	  ASSERT(parea->bBaseLocked == 1);

	  // Unlock the base and check if ok

	  if (FmaLock(parea->hfileSqd, SBH_LOCK_OFFS, SBH_LOCK_LENG, FALSE)) {
		parea->bBaseLocked--;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d UNLOCKED", bLock, parea, parea->bBaseLocked);
#endif
		return TRUE;
	  }

	  // If we got here the base can't be unlocked

#ifdef FMA_LOG_SQU_LOCK
Log("SquLockBase(%d)   <: [%09Fp] bBaseLocked=%d UNLOCK FAILED", bLock, parea, parea->bBaseLocked);
#endif
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine acquires exclusive access to the message base
//

  FMAAPI_(BOOL) SquEnterExclusive(PSAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive>: [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Check if area is already exclusive and leave incremeting exclusive count

	if (parea->bExclusive) {
	  parea->bExclusive++;
#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d ALREADY EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return TRUE;
	}

	// Lock message base and check if ok

	if (!SquLockBase(parea, TRUE)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT LOCK", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	ASSERT(!parea->bExclusive);

	// Increment exclusive count early so that 
	// SquLoadSbh() knows we're exclusive

	parea->bExclusive++;

#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive : [%09Fp] bBaseLocked=%d bExclusive=%d NOW EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Load base header and roll back if failed

	if (!SquLoadSbh(parea)) {
	  SquLockBase(parea, FALSE);
	  parea->bExclusive--;
#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT LOAD JBH", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d HAS EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine relinquishes exclusive access to the message base
//

  FMAAPI_(BOOL) SquLeaveExclusive(PSAREA parea)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileSqd != -1);

#ifdef FMA_LOG_SQU_LOCK
Log("SquLeaveExclusive>: [%09Fp] bBaseLocked=%d bExclusive=%d", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	// Check if not exclusive and fail if so

	if (!parea->bExclusive) {
	  FMAERROR(NOT_EXCLUSIVE);
#ifdef FMA_LOG_SQU_LOCK
Log("SquLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d NOT EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Check if exclusive more than once and leave decrementing exclusive count

	if (parea->bExclusive > 1) {
	  parea->bExclusive--;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLeaveExclusive<: [%09Fp] bExclusive=%d STILL EXCLUSIVE", parea, parea->bExclusive);
#endif
	  return TRUE;
	}

	ASSERT(parea->bExclusive == 1);

	// Save base header and check if ok

	if (!SquSaveSbh(parea)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT SAVE JBH", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Unlock message base and check if ok

	if (!SquLockBase(parea, FALSE)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquEnterExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d CANNOT UNLOCK", parea, parea->bBaseLocked, parea->bExclusive);
#endif
	  return FALSE;
	}

	// Decrement exclusive count

	parea->bExclusive--;

#ifdef FMA_LOG_SQU_LOCK
Log("SquLeaveExclusive<: [%09Fp] bBaseLocked=%d bExclusive=%d NOW NOT EXCLUSIVE", parea, parea->bBaseLocked, parea->bExclusive);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine locks squish area
//

  FMAAPI_(BOOL) SquLockArea(HAREA harea, BOOL bLock)
  {
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_SQU_LOCK
Log("\nSquLockArea(%d)   >: [%09Fp] bAreaLocked=%d", bLock, parea, parea->bAreaLocked);
#endif

	// Check if locking or unlocking

	if (bLock) {

	  // Check if area is already locked and leave incremeting lock count

	  if (parea->bAreaLocked) {
		parea->bAreaLocked++;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d ALREADY LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return TRUE;
	  }

	  // Lock message base and check if ok

	  if (!SquLockBase(parea, TRUE)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT LOCK", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Begin index caching and check if ok

	  if (!SquBeginSqiCache(parea)) {
		SquLockBase(parea, FALSE);
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT BEGIN INDEX CACHING", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  ASSERT(!parea->bAreaLocked);

	  // Increment area lock count

	  parea->bAreaLocked++;

	} else {

	  // Check if area is not locked and leave

	  if (!parea->bAreaLocked) {
		FMAERROR(NOT_LOCKED);
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d NOT LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Check if area is locked more than once and leave decrementing lock count

	  if (parea->bAreaLocked > 1) {
		parea->bAreaLocked--;
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d STILL LOCKED", bLock, parea, parea->bAreaLocked);
#endif
		return TRUE;
	  }

	  ASSERT(parea->bAreaLocked == 1);

	  // End index caching and check if ok

	  if (!SquEndSqiCache(parea)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT END INDEX CACHING", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Unlock message base and check if ok

	  if (!SquLockBase(parea, FALSE)) {
#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d CANNOT UNLOCK", bLock, parea, parea->bAreaLocked);
#endif
		return FALSE;
	  }

	  // Decrement area lock count

	  parea->bAreaLocked--;
	}

#ifdef FMA_LOG_SQU_LOCK
Log("SquLockArea(%d)   <: [%09Fp] bAreaLocked=%d", bLock, parea, parea->bAreaLocked);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquLock.c
/////////////////////////////////////////////////////////////////////////////
