/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish message routines module
//
// SquMsg.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 24/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

#define GOODMSGN(msgn) ((msgn) > 0 && (msgn) <= parea->sbh.msgnHigh)

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates message object
//

  static PSMSG FMASUB DoCreateMsg(PSAREA parea, MSGN msgn, UINT mode, ULONG fl)
  {
	PSMSG pmsg;

	ASSERT(parea != NULL);

	// Create message object and check if ok

	if ((pmsg = MemAlloc(sizeof(SMSG), MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize message object

	pmsg->magic = SMSG_MAGIC;
	pmsg->parea = parea;
	pmsg->msgn = msgn;
	pmsg->mode = mode;
	pmsg->fl = fl;

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys message object
//

  static BOOL FMASUB DoDestroyMsg(PSMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Free associated message control info

	if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);

	// Reset message object magic number

	pmsg->magic = ~SMSG_MAGIC;

	// Free message object

	MemFree(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine translates special message numbers
//

  static BOOL FMASUB DoXlateMsgn(PSAREA parea, MSGN * pmsgn)
  {
	MSGN msgn = *pmsgn;

	ASSERT(parea != NULL);

	// Select current message

	if (msgn == MSGN_CUR || msgn == MSGN_CURORPREV || msgn == MSGN_CURORNEXT) {
	  if (!GOODMSGN(parea->msgnCur)) {
		if (msgn == MSGN_CURORPREV) goto Prev;
		if (msgn == MSGN_CURORNEXT) goto Next;
NotF:   FMAERROR(NOT_FOUND);
		return FALSE;
	  } else {
		msgn = parea->msgnCur;
	  }
	} else

	// Select previous message

	if (msgn == MSGN_PREV) {

Prev: if (parea->msgnCur > 1) {
		msgn = parea->msgnCur - 1;
		if (msgn > parea->sbh.msgnHigh)
		  msgn = parea->sbh.msgnHigh;
	  } else
		goto NotF;
	} else

	// Select next message

	if (msgn == MSGN_NEXT) {

Next: if (parea->msgnCur < parea->sbh.msgnHigh) {
		msgn = parea->msgnCur + 1;
	  } else 
		goto NotF;
	} 

	// Set message number for caller

	if (pmsgn) *pmsgn = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the new message at the end of the base
//

  static BOOL FMASUB DoFixNewMsg(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	MSGN msgn;
	SQI sqi;

	ASSERT(parea->bExclusive);

	// Initialize new message index entry while incrementing
	// the next unique message id in the base header

	sqi.lfm = 0;
	sqi.umsg = parea->sbh.umsgNext++;
	sqi.hash = (ULONG)-1;

	// Save message unique id in the message structure

	pmsg->umsg = sqi.umsg;

	// Add message index entry and check if ok

	if (!SquAddSqi(parea, &sqi, &msgn))
	  return FALSE;

	// Check if index is in sync and fail if not

	if (msgn != pmsg->msgn) {
	  FMAERROR(INVALID_INDEX);
	  SquDelSqi(parea, &sqi, msgn);
	  return FALSE;
	}

	// That's it... there is no actual frame yet, it will be allocated 
	// later during the call to SquWriteMsg()

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the old message
//

  static BOOL FMASUB DoFixOldMsg(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	SQI sqi;

	ASSERT(parea->bExclusive);

	// Get the message index entry and check if ok

	if (!SquGetSqi(parea, &sqi, pmsg->msgn))
	  return FALSE;

	// Check if the frame offset is valid and 
	// save it in message handle

	if (!sqi.lfm) {
	  FMAERROR(INVALID_INDEX);
	  return FALSE;
	} else
	  pmsg->lfm = sqi.lfm;

	// Load the frame header and check if ok

	if (!SquLoadSfh(parea, &pmsg->sfh, pmsg->lfm))
	  return FALSE;

	// Check if another process is updating this frame

	if (pmsg->sfh.type == FRAME_UPDATE) {
	  FMAERROR(FRAME_IN_USE);
	  return FALSE;
	}

	// Invalidate index entry leaving umsgid intact

	sqi.lfm = 0;
	sqi.hash = (ULONG)-1;

	// Save message unique id in the message structure

	pmsg->umsg = sqi.umsg;

	// Update the index entry and check if ok

	if (!SquSetSqi(parea, &sqi, pmsg->msgn))
	  return FALSE;

	// Change the frame state to updating and
	// save the frame header then check if ok

	pmsg->sfh.type = FRAME_UPDATE;
	if (!SquSaveSfh(parea, &pmsg->sfh, pmsg->lfm))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens new message
//

  static BOOL FMASUB DoOpenNewMsg(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	BOOL bNew, bOk = TRUE;

	// Acquire exclusive access and check if ok

	if (!SquEnterExclusive(parea))
	  return FALSE;

	// Set up the new message flag

	bNew = !pmsg->msgn || (pmsg->msgn > parea->sbh.msgnHigh);

	// If creating a new message trim message base
	// Note: this could be supressed to speed up massive message creation

	if (bNew && !(parea->fl & FOA_DELAYTRIMMING)) SquTrimBase(parea, TRUE);

	// Make up the new message number

	if (bNew) pmsg->msgn = parea->sbh.msgnHigh + 1;

	// Fix up the index and data files to indicate that 
	// message creation is in process

	if (bOk) bOk = bNew ? DoFixNewMsg(pmsg) : DoFixOldMsg(pmsg);

	// Increment number of messages and highest message number

	if (bOk && bNew) {
	  parea->sbh.msgnHigh++;
	  parea->sbh.cmsg++;
	}

	// Relinquish exclusive access and check if ok

	if (!SquLeaveExclusive(parea))
	  bOk = FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens existing message
//

  static BOOL FMASUB DoOpenExistingMsg(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;

	ASSERT(pmsg->mode != OPENMSG_CREATE);

	// Check the message number and if invalid, reset current message cache 
	// if caller wants null message and leave

	if (!GOODMSGN(pmsg->msgn)) {
	  if (!pmsg->msgn) SquResetCurMsgCache(parea);
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	} 

	// Get the frame offset and check if ok

	if (!SquGetFrame(parea, &pmsg->lfm, pmsg->msgn))
	  return FALSE;

	// Load the frame header and check if ok

	if (!SquLoadSfh(parea, &pmsg->sfh, pmsg->lfm))
	  return FALSE;

	// Check the frame type

	if (pmsg->sfh.type == FRAME_USED) {
	  ;
	} else
	if (pmsg->sfh.type == FRAME_UPDATE) {
	  FMAERROR(FRAME_IN_USE);
	  return FALSE;
	} else {
	  FMAERROR(INVALID_FRAME);
	  return FALSE;
	}

	// Set current message cache to this message

	SquSetCurMsgCache(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes new message which has not been written
//

  static BOOL FMASUB DoFixStrayNewMsg(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	SQI sqi = SQI_INIT_EMPTY;

	ASSERT(pmsg->mode == OPENMSG_CREATE);

	// Acquire exclusive access and check if ok

	if (!SquEnterExclusive(parea))
	  return FALSE;

	// Delete associated index entry and adjust message counts

	if (SquDelSqi(parea, &sqi, pmsg->msgn)) {
	  parea->sbh.msgnHigh--;
	  parea->sbh.cmsg--;
	}

	// Relinquish exclusive access and check if ok

	if (!SquLeaveExclusive(parea))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  static BOOL FMASUB DoGetOpenMsgIndex(PSMSG pmsg, UINT * pimsg)
  {
	PSAREA parea = pmsg->parea;
	PSMSG pmsgThis;
	UINT imsg;

	// Scan through the open messages list looking for a match

	for (pmsgThis = parea->pmsgList, imsg = 0; pmsgThis != NULL; 
		 pmsgThis = pmsgThis->pmsgNext, imsg++) {
	  if (pmsgThis == pmsg) {
		if (pimsg) *pimsg = imsg;
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens message
//

  FMAAPI_(HMSG) SquOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl)
  {
	PSMSG pmsg = NULL;
	PSAREA parea;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_SQU_MSG
Log("SquOpenMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  // Clean up and leave

	  if (pmsg) {
		if (pmsg->fs & SMF_EXCLUSIVE) SquLeaveExclusive(parea);
		DoDestroyMsg(pmsg);
	  }

	  return NULL;
	}

#ifdef FMA_LOG_SQU_MSG
Log("\nSquOpenMsg       >: [%09Fp] msgn=%ld in %s, mode=%d", pmsg, msgn, parea->achPath, mode);
#endif

	// Translate special message numbers and check if ok

	if (!DoXlateMsgn(parea, &msgn))
	  goto Fail;

	// Create message object and check if ok

	if ((pmsg = DoCreateMsg(parea, msgn, mode, fl)) == NULL)
	  goto Fail;

	// Check if base is locked and acquire exclusive access early
	// to prevent unnecessary base header reads and writes

	if (parea->bBaseLocked) {
	  if (mode == OPENMSG_CREATE || 
		  mode == OPENMSG_WRITE ||
		  mode == OPENMSG_READWRITE) {
		if (SquEnterExclusive(parea)) 
		  SETFLAG(pmsg->fs, SMF_EXCLUSIVE);
		else 
		  goto Fail;
	  }
	}

	// Create a new message, or open an existing message, as specified

	if (mode == OPENMSG_CREATE) {
	  if (!DoOpenNewMsg(pmsg))
		goto Fail;
	} else {
	  if (!DoOpenExistingMsg(pmsg))
		goto Fail;
	}

	// Add message to the top of the list of open messages for this area

	ListAdd(&parea->pmsgList, pmsg, 0);    

	ASSERT(SquCheckMsg(pmsg));

#ifdef FMA_LOG_SQU_MSG
Log("SquOpenMsg       <: [%09Fp] msgn=%ld umsg=%ld lfm=%08lx in %s", pmsg, pmsg->msgn, pmsg->umsg, pmsg->lfm, parea->achPath);
#endif

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes message
//

  FMAAPI_(BOOL) SquCloseMsg(HMSG hmsg)
  {
	BOOL bOk = TRUE;
	PSAREA parea;
	PSMSG pmsg;
	UINT imsg;

	// Validate parameters

	if ((pmsg = (PSMSG)hmsg) == NULL || !SquCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_SQU_MSG
Log("SquCloseMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_SQU_MSG
Log("\nSquCloseMsg      >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	// Check if this is a new message which has not been written to and
	// update base to forget about it

	if (pmsg->mode == OPENMSG_CREATE && !(pmsg->fs & SMF_MASKWRITTEN)) {
	  if (!DoFixStrayNewMsg(pmsg))
		bOk = FALSE;
	}

	// Check if we entered an exclusive access and relinquish it

	if (pmsg->fs & SMF_EXCLUSIVE) {
	  if (!SquLeaveExclusive(parea))
		bOk = FALSE;
	}

	// Delete message from the list of open messages for this area

	if (DoGetOpenMsgIndex(pmsg, &imsg)) {
	  ListDel(&parea->pmsgList, imsg);
	} else {
	  ASSERT(DoGetOpenMsgIndex(pmsg, &imsg));
	}

	// Destroy message object

	DoDestroyMsg(pmsg);

	// Check if failed and jump to failure return

	if (!bOk) goto Fail;

#ifdef FMA_LOG_SQU_MSG
Log("SquCloseMsg      <: [%09Fp] in %s", pmsg, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message object
//

  FMAAPI_(BOOL) SquCheckMsg(PSMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check the message object magic number

	if (pmsg->magic != SMSG_MAGIC)
	  return FALSE;

	// Check the area object in the debugging build

	ASSERT(pmsg->parea != NULL);
	ASSERT(SquCheckArea(pmsg->parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be read
//

  FMAAPI_(BOOL) SquCheckMsgReadMode(PSMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be read

	if (pmsg->mode != OPENMSG_READ && pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_READ);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be written
//

  FMAAPI_(BOOL) SquCheckMsgWriteMode(PSMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be written

	if (pmsg->mode != OPENMSG_CREATE && pmsg->mode != OPENMSG_WRITE &&
		pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_WRITE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquMsg.c
/////////////////////////////////////////////////////////////////////////////
