/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish read routines module
//
// SquRead.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 23/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message header
//

  static BOOL FMASUB DoReadSmh(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	BOOL bKeepUmsg;

	ASSERT(parea->hfileSqd != -1);

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	// Load the message header and check if ok

	if (!SquLoadSmh(parea, &pmsg->smh, pmsg->lfm))
	  return FALSE;

	// Check if we need to keep stored umsgid in smh.umsg

	bKeepUmsg = (parea->fl & FOA_KEEPUMSG) || (pmsg->fl & FOM_KEEPUMSG);

	// Check if there is a umsgid stored in this message and 
	// blank it out so that application will not see it while
	// saving it in the message structure

	if ((pmsg->smh.attr & SMA_UMSG) && !bKeepUmsg) {
	  RESETFLAG(pmsg->smh.attr, SMA_UMSG);
	  pmsg->umsg = pmsg->smh.umsg;
	  pmsg->smh.umsg = 0;
	}

	// Set flags to prevent further header reading

	SETFLAG(pmsg->fs, SMF_GOTMSGH);

#ifdef FMA_LOG_SQU_READ
Log("DoReadSmh         : [%09Fp] %s to %s \"%s\"", pmsg, pmsg->smh.achBy, pmsg->smh.achTo, pmsg->smh.achRe);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads message control info
//

  static BOOL FMASUB DoReadCtrl(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	ULONG cbCtrl, cbRead;
	LONG offs;

	ASSERT(parea->hfileSqd != -1);

	// Reset associated message control info if any

	if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl); 
	pmsg->pszCtrl = NULL; pmsg->cchCtrl = 0;

	// Get safe message control info length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, NULL, &cbCtrl))
	  return FALSE;

	// Check if there is control info 

	if (cbCtrl) {

	  // Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	  if (!CHECK_DATA_SIZE_LIMIT(cbCtrl)) {
		FMAERROR(DATA_SIZE_LIMIT);
		return FALSE;
	  }
#endif
	  // Allocate memory to hold control info and check if ok

	  if ((pmsg->pszCtrl = MemAlloc((int)cbCtrl + 1, 0)) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  } else {
		pmsg->pszCtrl[cbCtrl] = '\0';
		pmsg->cchCtrl = (UINT)cbCtrl;
	  }

	  ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	  ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	  // Read the message control info and check if ok

	  offs = pmsg->lfm + parea->sbh.cbSfh + sizeof(SMH);
	  if (!FmaReadFile(parea->hfileSqd, offs, pmsg->pszCtrl, cbCtrl, &cbRead)) {
Fail:   MemFree(pmsg->pszCtrl); pmsg->pszCtrl = NULL; pmsg->cchCtrl = 0;
		return FALSE;
	  } else
		if (cbRead != cbCtrl) {
		  FMAERROR(INVALID_DATA);
		  goto Fail;
		}
	}

	// Set flags to prevent further control info reading

	SETFLAG(pmsg->fs, SMF_GOTCTRL);

#ifdef FMA_LOG_SQU_READ
Log("DoReadCtrl        : [%09Fp] cchCtrl=%d \"%s\"", pmsg, pmsg->cchCtrl, pmsg->pszCtrl);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads message text
//

  static BOOL FMASUB DoReadText(PSMSG pmsg, PCH * ppchText, ULONG * pcchText, 
								ULONG * poffText)
  {
	PSAREA parea = pmsg->parea;
	ULONG cbMsg, cbCtrl, cbRead;
	ULONG cch, cchAlloc;
	BOOL bAlloc = FALSE;
	LONG offs, offsEnd;

	ASSERT(parea->hfileSqd != -1);

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	// Get safe message and control info length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, &cbMsg, &cbCtrl))
	  return FALSE;

	// Calculate the message text to read and message end offsets

	offs = pmsg->lfm + parea->sbh.cbSfh + sizeof(SMH) + cbCtrl;
	if (poffText) offs+= *poffText;
	offsEnd = pmsg->lfm + parea->sbh.cbSfh + cbMsg;

	// Check if offset is past the end of the message text and fail,
	// otherwise calculate size of the message text we can read 

	if (offs > offsEnd) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	} else 
	  cch = offsEnd - offs;

	// Check if caller does not need all the message text and
	// adjust the read count if it's greater than requested 

	if (*pcchText != SRM_GETALLTEXT) {
	  if (cch > *pcchText) cch = *pcchText;
	}

	// Check if there is anything to read 

	if (cch > 0) {

	  // Check if we need to allocate text buffer for caller

	  if (*ppchText == NULL) {
		cchAlloc = (*pcchText != SRM_GETALLTEXT) ? *pcchText : cch;

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchAlloc)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate text buffer and check if ok 

		if ((*ppchText = MemAlloc((int)cchAlloc + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  (*ppchText)[cchAlloc] = '\0';
		  bAlloc = TRUE;
		}
	  } 

	  // Read the message text and check if ok

	  if (!FmaReadFile(parea->hfileSqd, offs, *ppchText, cch, &cbRead)) {
Fail:   if (bAlloc) { MemFree(*ppchText); *ppchText = NULL; }
		return FALSE;
	  } else
		if (cbRead != cch) {
		  FMAERROR(INVALID_DATA);
		  goto Fail;
		}
	}

	// Update caller's pointers

	*pcchText = cch;
	if (poffText) *poffText+= cch;

#ifdef FMA_LOG_SQU_READ
Log("DoReadText        : [%09Fp] cchText=%ld", pmsg, cch);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message
//

  FMAAPI_(BOOL) SquReadMsg(HMSG hmsg, PSMH psmh, PCH * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText)
  {
	PSAREA parea;
	PSMSG pmsg;
	PCH pszCtrl;

	// Validate parameters

	if ((pmsg = (PSMSG)hmsg) == NULL || !SquCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_SQU_READ
Log("SquReadMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_SQU_READ
Log("\nSquReadMsg       >: [%09Fp] msgn=%ld umsg=%ld lfm=%08lx in %s", pmsg, pmsg->msgn, pmsg->umsg, pmsg->lfm, parea->achPath);
#endif

	// Check if message is allowed to be read

	if (!SquCheckMsgReadMode(pmsg))
	  goto Fail;

	// Read the message header unless it's cached and check if ok

	if (!(pmsg->fs & SMF_GOTMSGH)) {
	  if (!DoReadSmh(pmsg))
		goto Fail;
	}

	// Read the message control info unless it's cached and check if ok

	if (!(pmsg->fs & SMF_GOTCTRL)) {
	  if (!DoReadCtrl(pmsg))
		goto Fail;
	}

	// Check if caller needs message header

	if (psmh) {
	  *psmh = pmsg->smh;
	}

	// Check if caller needs message control info 

	if (ppchCtrl) {
	  if (!pmsg->pszCtrl || !pmsg->cchCtrl) {
		*ppchCtrl = NULL; *pcchCtrl = 0;
	  } else {

		// Check if caller needs a readonly pointer and set it,
		// otherwise allocate a copy of message control info

		if (*pcchCtrl == SRM_GETCTRLPTR) {
		  *ppchCtrl = pmsg->pszCtrl; *pcchCtrl = pmsg->cchCtrl;
		} else 
		if ((pszCtrl = MemAlloc(pmsg->cchCtrl + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  goto Fail;
		} else {
		  memcpy(pszCtrl, pmsg->pszCtrl, pmsg->cchCtrl + 1);
		  *ppchCtrl = pszCtrl; *pcchCtrl = pmsg->cchCtrl;
		}
	  }
	}

	// Check if caller needs message text.

	if (ppchText) {
	  if (!DoReadText(pmsg, ppchText, pcchText, poffText))
		goto Fail;
	}

#ifdef FMA_LOG_SQU_READ
Log("SquReadMsg       <: [%09Fp] msgn=%ld umsg=%ld lfm=%08lx in %s", pmsg, pmsg->msgn, pmsg->umsg, pmsg->lfm, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message control info length
//

  FMAAPI_(BOOL) SquGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl)
  {
	ULONG cbCtrl;
	PSMSG pmsg;

	// Validate parameters

	if ((pmsg = (PSMSG)hmsg) == NULL || !SquCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	// Check if message is allowed to be read

	if (!SquCheckMsgReadMode(pmsg))
	  goto Fail;

	// Get safe message control info length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, NULL, &cbCtrl))
	  goto Fail;

	// Set message control info length for caller

	if (pcchCtrl) *pcchCtrl = cbCtrl;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message text length
//

  FMAAPI_(BOOL) SquGetMsgTextLen(HMSG hmsg, ULONG * pcchText)
  {
	ULONG cbCtrl, cbMsg;
	PSMSG pmsg;

	// Validate parameters

	if ((pmsg = (PSMSG)hmsg) == NULL || !SquCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	// Check if message is allowed to be read

	if (!SquCheckMsgReadMode(pmsg))
	  goto Fail;

	// Get safe message and control info length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, &cbMsg, &cbCtrl))
	  goto Fail;

	ASSERT(cbMsg >= sizeof(SMH) + cbCtrl);

	// Set message text length for caller

	if (pcchText) *pcchText = cbMsg - (sizeof(SMH) + cbCtrl);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquRead.c
/////////////////////////////////////////////////////////////////////////////
