/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish write routines module
//
// SquWrite.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 23/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets free or allocates new frame to write the message to
//

  static BOOL FMASUB DoGetFrame(PSAREA parea, PSFH psfh, PLONG plfm, ULONG cbFrame)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(psfh != NULL);
	ASSERT(plfm != NULL);

	// Reset target frame info

	*plfm = 0; memset(psfh, 0, sizeof(SFH));

	// Find the frame big enough to hold this message and check for failure
	// Note: this could be supressed to speed up massive message creation
	// by eliminating free frame chain look up

	if (!(parea->fl & FOA_CREATEATEOF))
	  if (!SquFindFree(parea, psfh, plfm, cbFrame))
		return FALSE;

	// Check if we found the free frame and remove it from the free chain,
	// otherwise create a new frame at the end of base

	if (*plfm) {
	  if (!SquDelFree(parea, psfh, *plfm))
		return FALSE;
	} else {
	  if (!SquAddFrame(parea, plfm, cbFrame))
		return FALSE;
	}

	ASSERT(*plfm != 0);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the frame to write the message to
//

  static BOOL FMASUB DoGetWriteFrame(PSMSG pmsg, ULONG cchCtrl, ULONG cchTextTotal)
  {
	PSAREA parea = pmsg->parea;
	LONG lfmPrev = 0, lfmNext = 0;
	ULONG cbFrame;
	LONG lfm;
	SFH sfh;

	ASSERT(parea != NULL);
	ASSERT(parea->bExclusive);

	ASSERT(pmsg->mode == OPENMSG_CREATE);

	// Calculate target frame size

	cbFrame = sizeof(SMH) + cchCtrl + cchTextTotal;

	// Check if we're replacing a message and if so, save its prev/next links and 
	// remove it from the used chain then add it to the free chain

	if (pmsg->lfm) {

	  lfmPrev = pmsg->sfh.lfmPrev;
	  lfmNext = pmsg->sfh.lfmNext;

	  if (!SquDelUsed(parea, &pmsg->sfh, pmsg->lfm))
		return FALSE;

	  if (!SquAddFree(parea, &pmsg->sfh, pmsg->lfm))
		return FALSE;
	}

	// Find the frame big enough to hold this message or 
	// allocate a new frame at the end of base

	if (!DoGetFrame(parea, &sfh, &lfm, cbFrame)) 
	  return FALSE;

#ifdef FMA_LOG_SQU_WRITE
//Log("DoGetWriteFrame   : [%09Fp] lfm=%08lx sfh.cbFrame=%ld sfh.type=%d", pmsg, lfm, sfh.cbFrame, sfh.type);
#endif

	ASSERT(lfm != 0);
	ASSERT(!SquCheckUsed(parea, lfm));
	ASSERT(!SquCheckFree(parea, lfm));

	// Fix up the target frame

	sfh.type = FRAME_USED;
	sfh.cbCtrl = 0;
	sfh.cbMsg = 0;

	// Check if this is the new frame at the end of the base and
	// do additional initialization

	if (sfh.magic != SFH_MAGIC) {
	  sfh.magic = SFH_MAGIC;
	  sfh.cbFrame = cbFrame;
	  sfh.rsvd = 0;
	} 

	// Check if we're not replacing a message or there are no used frames and if so, 
	// add the target frame to the end of the used chain

	if (!pmsg->lfm || !parea->sbh.lfmFirst) {

	  if (!SquAddUsed(parea, &sfh, lfm))
		return FALSE;

	} else {

	  // When we're replacing a message we need to insert target message 
	  // in the used chain at the same position where the original message was,
	  // so set this frame links accordingly

	  sfh.lfmPrev = lfmPrev;
	  sfh.lfmNext = lfmNext;

	  // Link old previous frame to this frame

	  if (lfmPrev) {
		if (!SquSetNextFrame(parea, lfmPrev, lfm))
		  return FALSE;
	  }

	  // Link old next frame to this frame

	  if (lfmNext) {
		if (!SquSetPrevFrame(parea, lfmNext, lfm))
		  return FALSE;
	  }

	  // Fix base header used chain first.last links

	  if (!sfh.lfmPrev) parea->sbh.lfmFirst = lfm;
	  if (!sfh.lfmNext) parea->sbh.lfmLast = lfm;
	}

	ASSERT(sfh.cbFrame >= cbFrame);
	ASSERT(sfh.cbCtrl == 0);
	ASSERT(sfh.cbMsg == 0);

	// Associate target frame with the message

	pmsg->sfh = sfh;
	pmsg->lfm = lfm;

	// Save the frame header and check if ok

	if (!SquSaveSfh(parea, &pmsg->sfh, pmsg->lfm))
	  return FALSE;

	// Set flag that we've got the writing frame

	SETFLAG(pmsg->fs, SMF_GOTFRAME);

	// Update current message cache

	SquSetCurMsgCache(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message header
//

  static BOOL FMASUB DoWriteSmh(PSMSG pmsg, PSMH psmh)
  {
	PSAREA parea = pmsg->parea;
	BOOL bKeepUmsg;
	SQI sqi;

	ASSERT(parea->hfileSqd != -1);

	// Check if message header is not specified and 
	// assume the cached one

	if (!psmh) psmh = &pmsg->smh;

	// Check if don't have umsg and get it from the index file

	if (!pmsg->umsg) {
	  if (SquGetSqi(parea, &sqi, pmsg->msgn))
		pmsg->umsg = sqi.umsg;
	}

	// Check if we need to keep stored umsgid in smh.umsg

	bKeepUmsg = (parea->fl & FOA_KEEPUMSG) || (pmsg->fl & FOM_KEEPUMSG);

	// Check if we have umsgid and set it into the message header,
	// otherwise make sure it's reset

	if (pmsg->umsg) {
	  SETFLAG(psmh->attr, SMA_UMSG);
	  psmh->umsg = pmsg->umsg;
	} else {
	  RESETFLAG(psmh->attr, SMA_UMSG);
	  psmh->umsg = 0;
	}

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);
	ASSERT(pmsg->sfh.cbFrame >= sizeof(SMH));

	// Write the message header and check if ok

	if (!SquSaveSmh(parea, psmh, pmsg->lfm)) {
	  if (!bKeepUmsg) { RESETFLAG(psmh->attr, SMA_UMSG); psmh->umsg = 0; }
	  return FALSE;
	}

	// Reset stored umsgid so that application will not see it unless
	// it specifically requested to keep it

	if (!bKeepUmsg) { RESETFLAG(psmh->attr, SMA_UMSG); psmh->umsg = 0; }

	// Check if wrote a new message header and 
	// update the cached one

	if (psmh != &pmsg->smh) {
	  SETFLAG(pmsg->fs, SMF_GOTMSGH);
	  pmsg->smh = *psmh;
	}

	// Check message length and update if necessary

	if (pmsg->sfh.cbMsg < sizeof(SMH)) {
	  pmsg->sfh.cbMsg = sizeof(SMH);
	  pmsg->sfh.cbCtrl = 0;
	  SETFLAG(pmsg->fs, SMF_SAVESFH);
	}

#ifdef FMA_LOG_SQU_WRITE
Log("DoWriteSmh        : [%09Fp] \"%s\" to \"%s\" re \"%s\"", pmsg, pmsg->smh.achBy, pmsg->smh.achTo, pmsg->smh.achRe);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message control info
//

  static BOOL FMASUB DoWriteCtrl(PSMSG pmsg, PCH pchCtrl, ULONG cchCtrl)
  {
	PSAREA parea = pmsg->parea;
	ULONG cbMsg, cbWritten;
	LONG offs, offsEnd;
	BOOL bNew = FALSE;
	PCH pch;

	ASSERT(parea->hfileSqd != -1);

	// Check if message contol info is not specified and 
	// assume the cached one, otherwise allocate a copy of
	// the new message control info unless it's empty

	if (!pchCtrl) {
	  pchCtrl = pmsg->pszCtrl;
	  cchCtrl = pmsg->cchCtrl;
	} else {
	  bNew = TRUE;
	  if (!cchCtrl) {
		pchCtrl = NULL;
	  } else {

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchCtrl)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate memory to hold message control info

		if ((pch = MemAlloc((int)cchCtrl + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  memcpy(pch, pchCtrl, (int)cchCtrl); pchCtrl = pch; 
		  pchCtrl[cchCtrl] = '\0';
		}
	  }
	}

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	// Calculate the message control info and frame end offsets

	offs = pmsg->lfm + parea->sbh.cbSfh + sizeof(SMH);
	offsEnd = pmsg->lfm + parea->sbh.cbSfh + pmsg->sfh.cbFrame;

	// Check if message control info we're going to write will 
	// run past the end of the frame

	if (offs + (LONG)cchCtrl > offsEnd) {
	  if (bNew && pchCtrl) MemFree(pchCtrl);
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message control info and check if ok

	if (pchCtrl && cchCtrl) {

	  if (!FmaWriteFile(parea->hfileSqd, offs, pchCtrl, cchCtrl, &cbWritten)) {
		if (bNew && pchCtrl) MemFree(pchCtrl);
		return FALSE;
	  } 

	  ASSERT(cbWritten == cchCtrl);
	}

	// Check if wrote a new message control info and update the cached one

	if (bNew) {
	  if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);
	  SETFLAG(pmsg->fs, SMF_GOTCTRL);
	  pmsg->pszCtrl = pchCtrl;
	  pmsg->cchCtrl = (UINT)cchCtrl;
	}

	// Get safe message length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, &cbMsg, NULL))
	  return FALSE;

	// Check if message length appears to be invalid and reset it
	// so it will get corrected when setting control info length

	if (pmsg->sfh.cbMsg != cbMsg) pmsg->sfh.cbMsg = sizeof(SMH);

	// Set message control info length in frame header and check if it runs past
	// the current message length and if so, update message length as well

	pmsg->sfh.cbCtrl = cchCtrl;
	if (sizeof(SMH) + cchCtrl > pmsg->sfh.cbMsg) pmsg->sfh.cbMsg = sizeof(SMH) + cchCtrl;
	SETFLAG(pmsg->fs, SMF_SAVESFH);

#ifdef FMA_LOG_SQU_WRITE
Log("DoWriteCtrl       : [%09Fp] cchCtrl=%d \"%s\"", pmsg, pmsg->cchCtrl, pmsg->pszCtrl);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message text
//

  static BOOL FMASUB DoWriteText(PSMSG pmsg, PCH pchText, ULONG cchText, 
								 ULONG * poffText)
  {
	PSAREA parea = pmsg->parea;
	ULONG cbCtrl, cbWritten;
	LONG offs, offsEnd;

	ASSERT(parea->hfileSqd != -1);

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	// Get safe message control info length and check if ok

	if (!SquGetSafeMsgCtrlLen(pmsg, NULL, &cbCtrl))
	  return FALSE;

	// Calculate the message text and frame end offsets

	offs = pmsg->lfm + parea->sbh.cbSfh + sizeof(SMH) + cbCtrl;
	if (poffText) offs+= *poffText;
	offsEnd = pmsg->lfm + parea->sbh.cbSfh + pmsg->sfh.cbFrame;

	// Check if message text we're going to write will 
	// run past the end of the frame

	if (offs + (LONG)cchText > offsEnd) {
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message text and check if ok

	if (pchText && cchText) {

	  if (!FmaWriteFile(parea->hfileSqd, offs, pchText, cchText, &cbWritten)) 
		return FALSE;

	  ASSERT(cbWritten == cchText);
	}

	// Set message text length in frame header

	pmsg->sfh.cbMsg = offs  + cchText - (pmsg->lfm + parea->sbh.cbSfh);
	SETFLAG(pmsg->fs, SMF_SAVESFH);

	// Update caller's text offset

	if (poffText) *poffText+= cchText;

#ifdef FMA_LOG_SQU_WRITE
Log("DoWriteText       : [%09Fp] cchText=%d", pmsg, cchText);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates the index entry associated with the message
//

  static BOOL FMASUB DoUpdateSqi(PSMSG pmsg)
  {
	PSAREA parea = pmsg->parea;
	SQI sqi;

	ASSERT(parea->hfileSqi != -1);

	// Load index entry and check if ok

	if (!SquGetSqi(parea, &sqi, pmsg->msgn))
	  return FALSE;

	ASSERT(pmsg->lfm != 0);
	ASSERT(pmsg->fs & SMF_GOTMSGH);

	// Set frame offset and hash values

	sqi.lfm = pmsg->lfm;
	sqi.hash = SquGetHash(pmsg->smh.achTo, (UINT)-1);

	// Check if message has been read and set the hash flag

	if (pmsg->smh.attr & SMA_READ) SETFLAG(sqi.hash, SQI_HASH_READ);

	// Save index entry and check if ok

	if (!SquSetSqi(parea, &sqi, pmsg->msgn))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine writes message
//

  FMAAPI_(BOOL) SquWriteMsg(HMSG hmsg, PSMH psmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal)
  {
	BOOL bOk = TRUE;
	PSAREA parea;
	PSMSG pmsg;

	// Validate parameters

	if ((pmsg = (PSMSG)hmsg) == NULL || !SquCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_SQU_WRITE
Log("SquWriteMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_SQU_WRITE
Log("\nSquWriteMsg      >: [%09Fp] msgn=%ld umsg=%ld lfm=%08lx in %s", pmsg, pmsg->msgn, pmsg->umsg, pmsg->lfm, parea->achPath);
#endif

	// Check if message is allowed to be written

	if (!SquCheckMsgWriteMode(pmsg))
	  goto Fail;

	// Check if there is no frame for this messsage

	if (pmsg->mode == OPENMSG_CREATE && !(pmsg->fs & SMF_GOTFRAME)) {

	  // We need a message header and control info 
	  // on the first write to any message

	  if (!psmh || (cchCtrl && !pchCtrl)) {
		FMAERROR(INVALID_PARAM);
		goto Fail;
	  }

	  // Acquire exclusive access and check if ok

	  if (!SquEnterExclusive(parea))
		goto Fail;

	  // Get the frame to write and save result

	  bOk = DoGetWriteFrame(pmsg, cchCtrl, cchTextTotal);

	  // Relinquish exclusive access and check if ok

	  if (!SquLeaveExclusive(parea) || !bOk)
		goto Fail;
	}

	ASSERT(pmsg->lfm >= (LONG)parea->sbh.cb);
	ASSERT(pmsg->lfm <= (LONG)parea->sbh.lfmFrameEnd);

	// Check if caller wants to write message header and write it now

	if (psmh) {
	  if (DoWriteSmh(pmsg, psmh)) {
		SETFLAG(pmsg->fs, SMF_MSGHWRITTEN);
	  } else 
		goto Fail;

	  // Update index entry so that the message will be valid as soon
	  // as the message header is written

	  if (!DoUpdateSqi(pmsg))
		goto Fail;
	}

	// Check if caller wants to write message control info and write it now. 
	// If failed, do not exit right away since we still want to update frame

	if (pchCtrl) {
	  if (DoWriteCtrl(pmsg, pchCtrl, cchCtrl)) {
		SETFLAG(pmsg->fs, SMF_CTRLWRITTEN);
	  } else 
		bOk = FALSE;
	}

	// Check if caller wants to write message text and there is no
	// error condition set when writing control info and write message text.
	// If failed, do not exit right away since we still want to update frame

	if (pchText && bOk) {
	  if (DoWriteText(pmsg, pchText, cchText, poffText)) {
		SETFLAG(pmsg->fs, SMF_TEXTWRITTEN);
	  } else 
		bOk = FALSE;
	}

	ASSERT(pmsg->sfh.cbMsg <= pmsg->sfh.cbFrame);
	ASSERT(pmsg->sfh.cbCtrl <= pmsg->sfh.cbFrame - sizeof(SMH));
	ASSERT(pmsg->sfh.cbCtrl <= pmsg->sfh.cbMsg);

	// Update frame header if necessary

	if (pmsg->fs & SMF_SAVESFH) {
	  if (!SquSaveSfh(parea, &pmsg->sfh, pmsg->lfm))
		return FALSE;
	  else
		RESETFLAG(pmsg->fs, SMF_SAVESFH);
	}

	// Check for error condition and fail now after updating the frame.
	// This would make us more robust against application errors.

	if (!bOk) goto Fail;

#ifdef FMA_LOG_SQU_WRITE
Log("SquWriteMsg      <: [%09Fp] msgn=%ld umsg=%ld lfm=%08lx in %s", pmsg, pmsg->msgn, pmsg->umsg, pmsg->lfm, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquWrite.c
/////////////////////////////////////////////////////////////////////////////
